			The "show graph" command
			========================

Note: User documentation is in the file "graph.txt" in the same directory
where "graph.e" is.


o The command:

<'
define <show_graph'command> "sh[ow] graph[<full'opt>full]\
[<opt>module[s]<eq><non_blank>]\
[<opt>size<eq><size'num>[*<width'num>[*<depth'num>]]]\
[<ps'opt>ps<eq><pswidth'num>*<psheight'num>]\
[<opt>def<eq><def'name>][ <names'non_blank>]" as {
    sn_dot.show_graph("<names'non_blank>", <depth'num|5>, 
	<width'num|5>, <size'num|60>, "<non_blank>", 
	TRUE, <pswidth'num|1>, <psheight'num|1>, "<full'opt>" != "",
	"<def'name>");
};
'>


o Defining sn_dot_parent (for snorb):

See more details in "Java handling" at end

<'
struct sn_dot_parent {

    -- Create a graph for java. Return the name of the .doj file
    create_graph(specifier: string): string is undefined;

    -- A callback routine, to be called when the user clicked on a node
    clicked_on(node: string, sub: string) is undefined;
};
'>



o Basic sn_dot utilities:

<'
extend global {
    sn_dot;

    init() is also {sn_dot = new};
};

struct sn_dot like sn_dot_parent {

    -- Create temp file, run dot and display program
    -- Note: the java_mode flag makes it do something different

    -- Modify this string to execute a postscript viewer other than the default.
    psviewer_cmd: string;

    display_graph(l: list of string, ps: bool) is {
	var exe:= 
	   "$SPECMAN_HOME/ce_tools/class_browser/`sn_arch.sh`/dot_executable";
	l = l.apply(str_replace(it, "\\\\", "\\"));
	files.write_string_list("/tmp/tmp.dot", l);
	
	if java_mode then {
            out(output_from(
		"/user/yoav/dot/DEMO/formatDemo < /tmp/tmp.dot > /tmp/tmp.doj"));
	} else if ps then {
            out(output_from_check(append(exe, " -Tps -o /tmp/tmp.ps /tmp/tmp.dot")));
	    out("\nOutput in /tmp/tmp.ps");
	    if (psviewer_cmd == "") then {
		out(output_from_check("ghostview -landscape /tmp/tmp.ps&"));
	    } else {
		out(output_from_check(append(psviewer_cmd, " /tmp/tmp.ps&")));
	    };
	} else {
            out(output_from_check(append(exe, " -Tgif -o /tmp/tmp.gif /tmp/tmp.dot")));
	    out("\nOutput in /tmp/tmp.gif");
            out(output_from_check("xv /tmp/tmp.gif&"));
	};
    };
};
'>


o graph_types:

Note: in initial version ignores parameter.

<'
extend sn_dot {

    g: list of string;  -- Will contain the list.
    module_map: list of bool;  -- Size of symtab.modules
    initial_sds: list (key: name) of struct_descriptor;
			-- The initial list, before name filtering
    sds: list (key: name) of struct_descriptor;
    depth_num: int;
    width_num: int;
    size_num: int;
    depth_map: list of int;  -- Same size as sds;
    the_inst: base_struct;
    full: bool;

    -- Clear all global structs
    clear_all() is {
	sds.clear();
	initial_sds.clear();
	depth_map.clear();
    };

    -- Return the name of the struct_descriptor with "____" for each blank or <>
    unblank(name: string): string is {
	return str_replace(name, "/[ '<>]/", "____");
    };

    -- Return TRU if this source item is in the module map
    ok_module(si: source_item): bool is {
	return module_map[si.source_module().module_num];
    };

    -- Returns the string representation of this number divided by 10
    div10(num: int): string is {
	return str_replace(dec(num), "/(.)$/", ".\\1");
    };

    -- Returns TRUE if sm is defined (directly) in this sd
    defined_here(sm: struct_member, sd: struct_descriptor): bool is {
	return sm.Context_struct == sd;  
		-- ## A very partial solution - Ronen needs to fix a bug
    };


    -- Returns ">" for sometimes gen-on-the-fly, else "" for sometimes 
    -- generated, else "!"
    gen_string(f: field): string is {
	if f in symtab.gotf_reachable_fields then {
	    return "\>";
	} else if f in symtab.reachable_fields then {
	    return "";
	} else {
	    return "!";
	};
    };
 

    -- Returns "[>][!][%]<short_name>"
    field_details(f: field): string is {
	return append(gen_string(f), (f.is_physical() ? "%" : ""), 
			quote_for_dot(f.short_name));
    };

    -- Quotes this string by putting a backslash before any of "{}<>"
    quote_for_dot(s: string): string is {
	return str_replace(s, "/([{}<>])/", "\\\\\1");
    };

    -- Set the module map to correspond to "modules"
    set_module_map(modules: string) is {
	if modules == "" then {
	    module_map = symtab.modules.apply(
                it.in_good_module() and .name != "graph"
                and .name != "graph2");
	} else {
	    module_map = symtab.modules.apply(.name ~ modules);
	};
	if not module_map.has(it == TRUE) then {
	    error("No modules match the '-modules=", modules, "'");
	};
    };

    -- Create the file and show it
    show_graph(names: string, depth: int, width: int, size: int, 
		modules: string, ps: bool, ps_width: int, ps_height: int, 
		full: bool, def: string) is {

	var title:= dec("show graph ", (full ? "-full " : ""),
		(modules != "" ? append("-modules=", modules, " ") : ""),
		"-size=", size, "*", width, "*", depth, " ",
		(def != "" ? append("-def=", def, " ") : ""),names);

	-- Deal with def stuff
	if def != "" then {
	    if names == "" then {names = "sys"};
	    specman("sn_dot.the_inst=", names);
	    if not the_inst.type().is_a_struct() then {
		error("'", names, "' is not a struct");
	    };
	    names = "...";
	} else {
	    if names == "" then {names = "..."};
	};

	depth_num = depth;
	width_num = width;
	size_num = size;

	-- Create basic lists
	clear_all();
	me.full = full;

	g = {"digraph G{"; "rankdir=LR;"};

	if ps then {
	    g.add("margin=\".2,.2\"; rotate=90; orientation=landscape;");
	    g.add("ratio=fill; page=\"8.5,11\";");
  	    var w:= div10(ps_width * 100);
	    var h:= div10(ps_height * 80);
	    g.add(dec("size=\"", w, ",", h, "\"; label=\"\";"));

	} else {
	    g.add("ratio=compress;");
	};

	g.add(
	  "node [shape=record, font=helvetica, height=\".1\", width=\".1\"];"); 

	-- Define the modules
	set_module_map(modules);

	-- Get the requested structs
	var sds1:= all_types.as_a(list of struct_descriptor)
		   .all(.name !~ "<*" and
			.name !~ "list ..." and .name !~ "/__/");

	sds1 = sds1.all(
		(ok_module(it) and .name !~ "/__/" and .is_a_struct()) or 
		(.fields.has(ok_module(it) and .name !~ "/__/" and 
			(full or .is_a_struct()))));

	-- Normalize sds1
	for each (sd) in sds1 do {
	    var n1:= normalize_name(sd.name);
	    if not initial_sds.key_exists(n1) then {
		initial_sds.add(sd);
	    };
	};
	sds1 = initial_sds.as_a(list of struct_descriptor);
	initial_sds.clear();

	if names ~ "!..." then {
	    var x:= $1;
	    sds1 = sds1.all(.name !~ x);
	    names = "...";
	};

	-- Handle the names and width issues
	var wanted:= sds1.all(.name ~ names);
	if wanted.is_empty() then {
	    error("No structs in specified modules match '", names, "'");
	};

	-- If all wanted, just put them in sds
	var ws:= wanted.size();
	if ws == sds1.size() and ws <= size_num then {
	    for each in sds1 {sds.add(it); depth_map.add(UNDEF)};

	-- Else must mark all to proper depth
	} else {
	    sds1 = {wanted; sds1.all(.name !~ names)};
            for each in sds1 {initial_sds.add(it)};

	    for each (sd) in wanted do {
		mark_to_depth(sd, depth_num);
	    };
	};		

	if  def != "" then {
	    scan_def(def);
	} else {
	    scan_types();
	};

	-- Close and display
	add_legend(def, title, ws - sds.size());
	g.add("}");
	display_graph(g, ps);
    };


    -- Scan all types matching sds
    scan_types() is {

	-- Output the list to the screen
	out("Graphing ", dec(sds.size()), " structs:");
	if sds.size() <= 20 then {
	    out(str_join(sds.name, "  "));
	} else {
	    out(str_join({sds[0..19].name; "..."}, "  "));
	};

	-- Create the graph
	for each (sd) in sds {

	    var dm:= depth_map[index];
	    var head:= append("<top>[", quote_for_dot(sd.name), "]");
	    var parts:= {head};
	    var count: int;

	    for each (f) in sd.fields {
		var is_list: bool;
		var n1:= f.base_type.name;
		if n1 ~ "/^list (.* )?of (\w+)$/" {
		    is_list = TRUE;
		    n1 = $2;
		};
		n1 = normalize_name(n1);
		if f.name ~ "/__/" or not ok_module(f) or 
				n1 ~ "<*" then {
		    continue;
		};
		
		-- See if defined here
		if not defined_here(f, sd) then {
		    continue;
	        };
	
		-- Handle struct fields
		if sds.key_exists(n1) and dm != 0 then {
		    var s:= dec("<", count, "> ", field_details(f), 
				(is_list ? "[]" : ""));
		    if sd.name == n1 then {
			s = append(s, "\\ \\ -\\>same");
		    } else {
		        g.add(dec(unblank(sd.name), ":", count, "->", 
			     unblank(n1), ":top",
			     (f.is_instance() ? "[style=bold, weight=3]" : ""),
			     (f.is_gotf() ? 
			     "[stye=bold, color=cyan1, weight=3]" : ""),
			     ";"));
		    };

		    parts.add(s);
		    count+= 1;

		-- Handle other fields
		} else if full then {
		    var s:= append(field_details(f), (is_list ? "[]" : ""), 
								":\\ ", n1);
		    parts.add(s);
		};
	    };
	
            -- Deal with parent hierarchy
	    var best_par:= sd;
	    var par:= sd;
	    while par.parent != NULL {
		par = par.parent.as_a(struct_descriptor);
		if sds.key_exists(par.name) then {
		    best_par = par;
		};
	    };
	    if best_par != sd then {  -- Has parent in group
		add_to_cluster(best_par.name, sd.name);

		if sd.name !~ "/ /" and sd.parent != NULL then {  
					-- Add parent except for when
		    parts[0] = append(parts[0], "\n(like ", sd.parent.name, 
									")");
		};
	    };


	    g.add(append(unblank(sd.name), "[label=", 
		quote(str_join(parts, "|")), 
		(sd.unit_info != NULL ? ", style=bold" : ""), "]"));

	};

	-- Add the code needed for all clusters
	add_all_clusters();
    };


    -- Mark this item and pointers from it to the specified depth. If
    -- its own depth is already >= depth, it is done
    mark_to_depth(sd: struct_descriptor, depth: int) is {

	if sds.size() >= size_num then {
	    return;
	};

	var i:= sds.key_index(sd.name);
	if i == UNDEF then {   -- Add it
	    sds.add(sd);
	    depth_map.add(depth);
	} else {
	    var d1:= depth_map[i];
	    if d1 >= depth {
		return;  -- It was done at the right depth already
	    };
	};
	if depth == 0 {  -- Nobody to add
	    return;
	};


	-- Do the recursive calls 
        for each (f) in sd.fields {
	    var n1:= f.base_type.name;
	    if n1 ~ "/^list .*of (\w+)$/" {
		n1 = $1;
	    };
	    n1 = normalize_name(n1);
	    if f.name ~ "/__/" or not ok_module(f) or n1 ~ "<*" then {
		continue;
	    };

	    if not defined_here(f, sd) then {
		continue;
	    };
	
	    var sd2:= initial_sds.key(n1);
	    if sd2 == NULL then {
		continue;
	    };

	    mark_to_depth(sd2, depth-1);
	};
    };
};
'>


o Checking that fields are generateable:

<'
extend field {
    // retuirns TRUE if the 'me' field might be generated on the fly 
    // (from a gen action)
    is_gotf(): bool is {
	for each (gi) in pgen3.gotf_infos do {
	    //if gi != NULL && !gi.gen_path.fields.is_empty()
	    if gi != NULL && gi.gen_path.has_fields()
		//&& (me == gi.gen_path.fields.top().field) then {
		&& (me == gi.gen_path.get_last_field()) then {
		return TRUE;
	    };
	};
    };
};
'>


o Creating the legend:

<'
extend sn_dot {
    -- Add the legend box
    add_legend(def: string, title: string, left_out: int) is {
	var top_module:= symtab.modules.all(.name !~"/^graph(\d)?$/")
								.top().name;
	var l:= str_split(date_time(), "");
	var date:= append(l[2], "-", l[1], "-", l[4]);
	title = str_replace(str_expand_dots(title), "/ +$/", "");
	if str_len(title) > 30 then {
	      title = str_replace(title, "/^(..............................)/",
			"\1\l");
	};

	var s:= "";
	if def == "" then {
	    def = "Class";
	    s = append("Legend| Bold box: Unit\l",
		" Bold arrow: Unit instance\l Cyan arrow: Gen on the fly\l",
		" Green border: Inheritance family\l []: List\l",
		" \>: Gen on the fly\l %: Physical\l",
		" !: Do not generate\l\l|");
	};
	s = append(s, def, " diagram created by|", (java_mode ? "\n" : ""),
		title, "\lTop module: ",
		top_module, "\lDate: ", date, "\l");
	if def == "Class" and left_out > 0 then {
	    s = dec(s, "... ", left_out, " structs left out\l");
	};

	g.add(append("_legend_box_[shape=record, color=brown, label=\"", s,
		"\"];"));
    };
};
'>



o Showing clusters:

The "top parent" of a struct in the shown graph is defined
to be the struct in the graph who is a parent of this
struct and whose own parent is not in the graph.

We group all structs who have the same "top parent" 
together in a cluster, shown with a green border 
(note "clusters of one item" are not shown with a border).

Here is how it is done:

  - Whenever we meet a struct which has a top parent different from
    it, we call:

	add_to_cluster(top_parent_name, this_struct_name)

  - At the end, we call:

	add_all_clusters()	

Finally, there is a list under sn_dot:

	clusters: list (key: name) of sn_dot_cluster;

<'
-- Such a struct is created for each cluster
struct sn_dot_cluster {
    name: string;  -- Name of "top parent"
    children: list of string;  -- Names of all shown children
};

extend sn_dot {
    clusters: list (key: name) of sn_dot_cluster;

    -- Clear all lists on "show graph"
    clear_all() is also {
	clusters.clear();
    };

    -- Add to existing cluster or create, as needed
    add_to_cluster(top_parent: string, child: string) is {

	var c:= clusters.key(top_parent);
	if c == NULL then {
	    c = new;
	    c.name = top_parent;
	    clusters.add(c);
	};

	c.children.add(child);
    };


    -- Add all existing clusters to the "g" list
    add_all_clusters() is {
	for each (c) in clusters {
	    g.add(append("subgraph cluster_", unblank(c.name), "{"));

	    for each (ch) in c.children {
		g.add(append(unblank(ch), "->", unblank(c.name), "[", 
		   (java_mode ? "color=white" : "style=invis"), 
		   ", weight=1,minlen=0];"));
		    -- This is voodoo dot programming to request
		    -- parent-to-child ordering within cluster if possible
	    };

            g.add(append("label=\"\";color=forestgreen;\n}"));
	};
    };
};
'>


o How name normalization works:

The problem is that for subtypes, one can encounter both
"green tx packet" and "tx green packet". Thus, the decision
is to always used the "normalized" form, which will be the
first such form encountered in a particular "show graph" run.

Note that the normalized name can be different from the original
name only if the original name has two or more prefixes.

Normalization is implemented by the method:

   sn_dot.normalize_name(name: string): string;

which can only be given names of non-list types.

It works by maintining two lists:

 sorted_names: list (key: it) of string;
 offical_names: list of string;

It works as follows:

 - For names that do not have two or more prefixes, it returns the original.

 - Otherwise, it sorts the prefixes alphabetically and looks in
   sorted_names.

 - If found, it returns the corresponding official_name.

 - Otherwise, it addes sorted_name and name to the corresponding
   lists, and returns name.

<'
extend sn_dot {

    -- Used by normalize_name (see description above)
    sorted_names: list (key: it) of string;
    official_names: list of string;

    -- Return the normalized name (see description above)
    normalize_name(name: string): string is {
	if name !~ "/ .+ /" or name ~ "/[[(]/" then {
	    return name;
	};

	var parts:= str_split(name, "");
	var last:= parts.pop();
	var sorted:= str_join({parts.sort(it); last}, " ");
	var ind:= sorted_names.key_index(sorted);
	if ind != UNDEF then {
	    return official_names[ind];
	} else {
	    sorted_names.add(sorted);
	    official_names.add(name);
	    return name;
	};
    };
	

    -- Clear all lists for "show graph"
    clear_all() is also {
	sorted_names.clear();
	official_names.clear();
	try {
	    symtab.gen_static_analysis();  -- To get gen info
	} else {
	    out("Warning: could not finish static analysis.\n",
		"Some of the generation info may be incomplete.\n",
		"Error message follows:");
	    write_the_error();
	};
    };
};
'>


Java handling
-------------

This package can also be called from within Java via the 
snorb. The two callable routines are create_graph() and clicked_on().

<'
extend sn_dot {

    java_mode: bool;

    -- Create a graph for java. Return the name of the .doj file
    create_graph(specifier: string): string is {
	java_mode = TRUE;
	try {
	    specman("show graph ", specifier);
	    java_mode = FALSE;
	} else {
	    java_mode = FALSE;
	    fail();
	};
	return "/tmp/tmp.doj";
    };

    

    -- A callback routine, to be called when the user clicked on a node
    clicked_on(node: string, sub: string) is {
	specman("source ", node, ".", sub);
    };
};
'>


Using -def
----------


User-define visitor:

<'
struct sn_graph_scanner {
    scan(inst: base_struct, text: *list of string, 
                children: *list of base_struct,
		labels: *list of string) is undefined;

    edit_full_list(l: list of string) is empty;
};


extend sn_dot {

    instances: list (key: it) of base_struct;

    -- Scan all instances starting at the_inst
    scan_def(def: string) is {
	def = append("sn_graph_", def);
	instances.clear();
	var sd:= type_called(def).as_a(struct_descriptor);
	if sd == NULL then {
	    error("No such struct type '", def, "'");
	};
	var s1:= no_save.make_new_from_descriptor(sd, NULL, NULL);
	var scanner: sn_graph_scanner = s1.as_a(sn_graph_scanner);
	if scanner == NULL then {
	    error("'", def, "' is not a child of sn_graph_scanner");
	};

	instances.add(the_inst);
	var next:= 0;
	var too_much: bool;
	var too_wide: bool;

	-- The following is done here as a kludge to avoid the old names
	sorted_names.clear();
	official_names.clear();

	while instances.exists(next) do {

	    -- Prepare
	    var now:= instances[next];

	    var text: list of string;
	    var children: list of base_struct;
	    var labels: list of string;

	    -- Call and decide
	    scanner.scan(now, text, children, labels);
	    if text.is_empty() then {
		return;  -- Do not display this guy
	    };
	    if children.size() != labels.size() and labels.size() != 0 then {
		error("children.size() != labels.size() in '", def, "'");
	    };
	    var parts:= {quote_for_dot(dec(str_join(text, "\n")))};

	    -- add the children
	    for each in children do {
		if it == NULL then {continue};
		if not instances.key_exists(it) then {
		    if index >= width_num then {
			if not too_wide then {
			    out("More then ", dec(width_num),
				" edges from one node - to see e.g. to width ",
				"30 use:\n 'show graph -size=60*30 ...'");
			    too_wide = TRUE;
			};
			break;
		    };
		
		    if instances.size() >= size_num then {
			if not too_much then {
			    out("More then ", dec(size_num), 
				" instances - to see more e.g. up to 100",
  				" instances use:\n 'show graph -size=100 ...'");
			    too_much = TRUE;
			};
			break;
		    };
		    
		    instances.add(it);
		};
	    };

	    -- Go through the labels
	    var has_labels:= labels.size() != 0;
	    for each (child) in children do {
		var child_index: int;

		if child != NULL then {
		    child_index = instances.key_index(child);
		    if child_index == UNDEF then {
		        continue;
		    };
		    if index >= width_num then {
		        break;
		    };
		};

	        if has_labels then {
		    parts.add(dec("<", index, "> ", 
		 		    quote_for_dot(labels[index])));

		    if child != NULL then {
		        g.add(dec("n", next, ":", index, "->n", child_index, 
			   (next == child_index ? dec(":", index) : ""), ";"));
		    };
		} else {
		    g.add(dec("n", next, "->n", child_index, 
			(next == child_index ? dec(":", index) : ""), ";"));
		};
	    };

	    if children.size() > width_num then {
		    parts.add(dec("...\\ (", children.size(), ")"));
	    };
	    g.add(dec("n", next, "[label=", quote(str_join(parts, "|")), "]"));
	
	    next += 1;
	};

	-- A way for the user to change the global list
	scanner.edit_full_list(g);
    };
};
'>



The -def=child option
---------------------

o Example:

   show graph -def=child type(packet)

will show all true children of type packet
(pointing directly only at true children, not all siblings).

<'
struct sn_graph_child like sn_graph_scanner {
    c1: list (key: it) of type_descriptor;

    scan(inst: base_struct, text: *list of string, 
                children: *list of base_struct,
		labels: *list of string) is {

	var this := inst.as_a(struct_descriptor);
	if this == NULL then {
	    error(inst, " is not a struct descriptor");
	};

	text = {this.name};
	var ch:= this.children.all(.parent == this);
	c1.clear();
	for each in ch do {
	    var n:= type_called(sn_dot.normalize_name(it.name));
	    if not c1.key_exists(n) then {
	        c1.add(n);
	    };
	};
	children = c1.as_a(list of base_struct);
	c1.clear();
    };
};
'>


The "show structs" command
--------------------------

o Syntax:

    show structs [-modules=<any>] [-gen_only] [-nosub] <struct-type-wildcard>

Displays all info on that struct type.


o Implementation:


<'
define <show_summary'command> "sh[o[w]] st[r[uct[s]]][<opt>modules=<non_blank>]\
[<gen'opt>gen[_only]][<internal'opt>internal][<nochild'opt>nochild][ <any>]" 
									as {
    sn_dot.show_structs_1("<non_blank>", "<gen'opt>" != "", 
	"<nochild'opt>" == "", "<internal'opt>" != "", 
	str_expand_dots("<any|...>"));
};

extend sn_dot {

    -- This is just an envelope for show_structs(), used for the "display" stuff
    show_structs_1(modules: string, gen_only: bool, show_children: bool,
		 internal: bool, names: string) is {
	if not display.on then {
	    show_structs(modules, gen_only, show_children, internal, names);
	} else {
	    display a source_list_display_window with {
	        .title = append("Show struct ", names);
	        .recipe = append("sn_dot.show_structs(", quote(modules), ", ",
			gen_only, ", ", show_children, ", ", internal, ", ",
			quote(names), ")");
	    };
	};
    };


    internal_flag: bool;

    -- The real "show structs"
    show_structs(modules: string, gen_only: bool, show_children: bool,
		 internal: bool, names: string) is {

	-- Initial stuff
	clear_all();
	set_module_map(modules);
	current_struct = NULL;
	internal_flag = internal;

	-- Create a list of all struct descriptors which "may" be interseting
	-- (i.e. have fields, methods or events in the specified modules)
	var sd_map:= all_types.apply(FALSE);

	-- Mark by method
	for each (m) in symtab.get_all_layers() {
	    if ok_module(m) then {
		sd_map[m.Context_struct.d_index] = TRUE;
	    };
	};

	-- Mark by event
	for each (e) in scheduler.event_map {
	    if e.parent_type == NULL or not ok_module(e) or 
			e.event_name == "quit" {
		continue;
	    };
	    sd_map[e.parent_type.d_index] = TRUE;
	};

	-- Mark by field
	for each struct_descriptor (sd) in all_types do {
	    if sd.fields.has(ok_module(it)) then {
		sd_map[sd.d_index] = TRUE;
	    };
	    if sd_map[sd.d_index] and (
		   sd.name ~ "/__/" or not sd.is_a_struct()) then {
		sd_map[sd.d_index] = FALSE;
	    };
	};
		    
	var sds1:= all_types.all(sd_map[index] and .name ~ names)
					.as_a(list of struct_descriptor);

	if sds1.is_empty() then {
	    error("No struct in the specified modules matches '", names, "'");
	};
	
	-- Add children if requested
	if show_children then {	
	    var sds2: list of struct_descriptor;
	    for each in sds1 do {
		sds2.add(it); 
		sds2.add(.children);
	    };
	    sds1 = sds2;
	};

        -- Normalize sds1
        for each (sd) in sds1 do {
            var n1:= normalize_name(sd.name);
            if not initial_sds.key_exists(n1) then {
                initial_sds.add(sd);
            };
        };
	sds1 = initial_sds.as_a(list of struct_descriptor);
        initial_sds.clear();

	-- Finally, show all the structs
        for each in sds1 do {
	    show_struct_details(it, gen_only);
	};
    };
};
'>


o Showing struct details:

<'
extend sn_dot {

    -- Updated as needed by add_struct_member()
    current_struct: struct_descriptor;  
    current_header: string;

    -- Create the list of strings (in g), then finally write it
    show_struct_details(sd: struct_descriptor, gen_only: bool) is {
	current_header = "";
	 
	-- Go through fields
	if (TRUE) {  -- They want fields (this is here for future filtering)
	    var head:= (gen_only ? "Generated fields:" : "Fields:");
	    for each (f) in sd.fields {
		var is_list: bool;
		var n1:= f.base_type.name;
		if n1 ~ "/^list (.* )?of (\w+)$/" {
		    is_list = TRUE;
		    n1 = $2;
		};
		n1 = normalize_name(n1);
		if f.name ~ "/__/" or not ok_module(f) or 
				n1 ~ "<*" then {
		    continue;
		};
		
		-- See if defined here
		if not defined_here(f, sd) then {
		    continue;
	        };

		-- See if gen_only
		if gen_only and f not in symtab.reachable_fields and f not
          		       in symtab.gotf_reachable_fields then {
		    continue;
		};

		var s:= append((f in symtab.gotf_reachable_fields ? ">" 
			: (f in symtab.reachable_fields ? "" : "!")),
		        (f.is_physical() ? "%" : ""), f.short_name, ":  ",
			f.base_type.name);
		add_struct_member(s, f, head, sd);
	    };
	};


	-- Go through methods. We create a single list of methods (and a
	-- corresponding list of strings), then sort them into methods and TCMs

	var methods: list of method;
	var m_strings: list of string;
	var TCMs: list of tcm;

	if (not gen_only) {

	    -- Create the lists
	    for each (m1) in sd.methods {
		var m:= m1;

		-- Go up chain, looking for a good one or NULL
		-- When one is found (the last one in the requested methods),
		-- we want to continue up the chain until we find the first
		-- one (mold) which was not in the requested modules. 
		-- Then we add "(was defined)" if mold was defined, 
		var ok_m: method;
		var was_defined: bool;
		while m != NULL {
		    if ok_module(m) and defined_here(m, sd) and 
			 (m.fst_action_srcref != 0 or m.kind == undefined or
			  m.result_type != NULL or internal_flag) then {
			ok_m = m;
		    } else if ok_m != NULL and not ok_module(m) then {
			was_defined = TRUE;
		    };
		    m = m.get_previous_method();
		};
		if ok_m != NULL then {
		    var s:= append(ok_m.get_name(), "()");
		    var l: list of string;
		    if was_defined then {
			l.add("extend");
		    };
		    if ok_m.kind == undefined or ok_m.kind == empty then {
			l.add(append("is ", ok_m.kind));
		    };
		    if not l.is_empty() then {
			s = append(s, "   (", str_join(l, ", "), ")");
	            };
		    methods.add(ok_m);
		    m_strings.add(s);
		};
	    };

	    -- Show them by type
	    for each (m) in methods {
		if m is not a tcm then {
		    add_struct_member(m_strings[index], m, "Methods:", sd);
		};
	    };
	    for each (m) in methods {
		if m is a tcm then {
		    add_struct_member(m_strings[index], m, "TCMs:", sd);
		};
	    };
	};

	-- Go through events
	if (not gen_only) {
	    var head:= "Events:";
	    for each (e) in scheduler.event_map {
		if e.parent_type == NULL or not ok_module(e) or 
			e.event_name == "quit" or
			(e.event_name ~ "/'/" and not internal_flag) or
			normalize_name(e.parent_type.name) != sd.name then {
		    continue;  -- For some reason this is a good indication
		};
		var l: list of string;
		if e.kind == empty then {
		    l.add("is empty");
		};
		if e.cover_descriptor != NULL then {
		    l.add("has cover");
		};
		var s:= e.event_name;
		if not l.is_empty() then {
		    s = append(s, "   (", str_join(l, ", "), ")");
		};
		add_struct_member(s, e, head, sd);
	    };
	};




    };

  
    -- Add this to the list, with "head" if needed
    add_struct_member(s: string, si: source_item, head: string, 
		      sd: struct_descriptor) is {

	-- Head struct header if needed
	if sd != current_struct then {
	    current_struct = sd;
	    var l: list of string;
	    if not ok_module(sd) then {
		l.add("extend");
	    };
	    if sd.name !~ "/ /" and sd.parent != type(any_struct) and
						sd.parent != NULL then {
		l.add(append("like ", sd.parent.name));
	    };
	    var s1:= l.is_empty() ? "" : append("  (", str_join(l, ", "), ")");
	    s1 = append("\n--------- ", sd.name, s1);
	    out(str_exactly(s1, 48), 
			str_replace(sd.source_full(), "/^at/", ""));
	};

	-- Add head if needed
	if head != current_header then {
	    current_header = head;
	    out("\n", head, "\n");
	};

	-- Add the item itself
	out("  ", str_exactly(s, 45), 
			str_replace(si.source_full(), "/^at/", ""));
    };
};
'>
