#include "sn_esi.h"
#include "c_simulator.h"

static void set_handles();
static void pre_start();
static void set_sp_info(SN_TYPE(simple_port) sport);
static void set_ep_info(SN_TYPE(event_port) eport);
static void set_bp_info(SN_TYPE(buffer_port) bport);
static void elab_ports(void);
static char* my_sprintf(char* format, ...);
static void set_buffer_port_handle(SN_TYPE(buffer_port) bport,
				   SN_TYPE(rf_port) rfp);
static void set_event_port_handle(SN_TYPE(event_port) eport,
				  SN_TYPE(rf_port) rfp);

static void c_buffer_get_packed(SN_TYPE(buffer_port) port, 
			 SN_TYPE(untyped) *value);
static void c_buffer_put_packed(SN_TYPE(buffer_port) port, 
			   SN_TYPE(untyped) value);

/* used to save adapter id received from Specman */
static int adapter_id;

/* used to give private indices to simple ports */
static global_index = 0;
static int OUTEPA[100];
static int ep_index = 0;


void esi_initialize_adapter(void) {
 SN_LIST(string) c_simulator_id_names = SN_LIST_NEW(string);
 
  SN_LIST_ADD(c_simulator_id_names,"c_simulator");

 /* now register adapter in Specman */
  adapter_id = sn_esi_register_adapter("C simulator",c_simulator_id_names);
  global_index = 0;
  ep_index = 0;
}

void esi_advance() {
  /* cause all relevant puts to be called */
  /* advance simulator */
  c_simulator_advance();
}

void esi_inform_state(SN_TYPE(esi_state) state) {

  switch (state) {
    case SN_ENUM(esi_state, pre_start):
      pre_start();
      break;

    case SN_ENUM(esi_state, run):
      /* now that we know that we have a design
      /* and call it */
      c_simulator_init();

      break;

    case SN_ENUM(esi_state, pre_extract):
      break;

    case SN_ENUM(esi_state, finalize):
      break;
  }
}

int esi_write_stubs(char* base_name, char* extension) {
  SN_LIST(any_port) port_list = sn_esi_get_adapter_ports(adapter_id);
  int i;
  SN_TYPE(any_port) p;
  SN_TYPE(simple_port) sp;
  SN_TYPE(rf_port) rfp;
  SN_LIST(string) extern_strings = SN_LIST_NEW(string);
  SN_LIST(string) get_strings = SN_LIST_NEW(string);
  SN_LIST(string) put_strings = SN_LIST_NEW(string);
  SN_LIST(string) emit_strings = SN_LIST_NEW(string);
  char* filename;
  FILE* fp;

  if (extension[0] != '\0' && strcmp(extension, ".c")) {
    sn_esi_error(adapter_id, SN_ENUM(sn_esi_error_severity, esi_error),
                  "Illegal extension for stub file - %s", extension);
    return 0;
  }

  elab_ports();
  for (i=0; i<SN_LIST_SIZE(port_list); i++) {
    p = SN_LIST_GET(port_list, i, any_port);
    rfp = sn_rf_manager_get_port_by_instance(p);
    switch (sn_rf_port_get_kind(rfp)) {

    case SN_ENUM(rf_port_kind, rf_simple_port):
    {
      SN_TYPE(rf_type) rft = sn_rf_port_get_element_type(rfp);
      SN_TYPE(rf_scalar) rfs;

      switch (sn_rf_type_get_kind(rft)) {
        case SN_ENUM(rf_type_kind, rf_scalar_type):
          rfs = SN_AS_A_CHILD_OF(rft, rf_scalar);
          switch (sn_rf_scalar_get_size_in_bits(rfs)) {
            case 8:
              SN_LIST_ADD(extern_strings, my_sprintf("extern char %s;", sn_esi_full_path(p)));
              break;

            case 16:
              SN_LIST_ADD(extern_strings, my_sprintf("extern short %s;", sn_esi_full_path(p)));
              break;
            
            case 32:
              SN_LIST_ADD(extern_strings, my_sprintf("extern int %s;", sn_esi_full_path(p)));
              break;
            
            default:
              sn_esi_error(adapter_id, SN_ENUM(sn_esi_error_severity, esi_error),
                           "Illegal size of scalar port %s - %d", sn_esi_e_path(p), sn_rf_scalar_get_size_in_bits(rfs));
              break;
          }
          SN_LIST_ADD(get_strings,
                      my_sprintf("    case %d:\n      return %s;",
                                 ((simple_port_info_t*)p->adapter_specific_info)->index,
                                 sn_esi_full_path(p)));
          SN_LIST_ADD(put_strings,
                      my_sprintf("    case %d:\n      %s = _value;\n      break;",
                                 ((simple_port_info_t*)p->adapter_specific_info)->index,
                                 sn_esi_full_path(p)));
          break;
        
        default:
              sn_esi_error(adapter_id, SN_ENUM(sn_esi_error_severity, esi_warn),
                           "Currently only simple ports of scalar are supported - %s", sn_esi_e_path(p));
              break;
      }
    }
    break;
    case SN_ENUM(rf_port_kind, rf_event_port):
      break;
    default:
      sn_esi_error(adapter_id, SN_ENUM(sn_esi_error_severity, esi_warn),
                   "Currently only simple ports of scalar are supported - %s", sn_esi_e_path(p));
      break;
    }
  }

  if (extension[0] != '\0') {
    filename = my_sprintf("%s.c", base_name);
  }
  else {
    filename = my_sprintf("%s_csim.c", base_name);
  }
  if ((fp = fopen(filename, "w")) == NULL) {
    sn_esi_error(adapter_id, SN_ENUM(sn_esi_error_severity, esi_error),
                 "Could not open stub file '%s' for writing", filename);
    return 0;
  }
  else {
    fputs(sn_str_join(extern_strings, "\n"), fp);
    fputs("\nint c_simulator_get(int index) {\n  switch (index) {\n", fp);
    fputs(sn_str_join(get_strings, "\n"), fp);
    fputs("    default:\n      return 0;\n  }\n}", fp);
    fputs("\nvoid c_simulator_put(int index, int _value) {\n  switch (index) {\n", fp);
    fputs(sn_str_join(put_strings, "\n"), fp);
    fputs("    default:\n      break;\n  }\n}", fp);
    fclose(fp);
  }
  return 1;
}

void esi_tick_start(void) {
}

void esi_tick_end(void) {
}

void esi_simple_get_int(SN_TYPE(simple_port) port, int* value) {
  *value = c_simulator_get(((simple_port_info_t*)port->adapter_specific_info)->index);
}

void esi_simple_put_int(SN_TYPE(simple_port) port, int value) {
  c_simulator_put(((simple_port_info_t*)port->adapter_specific_info)->index, value);
}


static void elab_ports() {
  SN_LIST(any_port) port_list = sn_esi_get_adapter_ports(adapter_id);
  int i;
  SN_TYPE(any_port) port;
  SN_TYPE(rf_port) rfp;

  for (i=0; i<SN_LIST_SIZE(port_list); i++) {
    port = SN_LIST_GET(port_list, i, any_port);
    rfp = sn_rf_manager_get_port_by_instance(port);
    switch (sn_rf_port_get_kind(rfp)) {

    case SN_ENUM(rf_port_kind, rf_simple_port):
      set_sp_info((SN_TYPE(simple_port)) port);
      break;
      
    case SN_ENUM(rf_port_kind, rf_buffer_port):
      set_bp_info((SN_TYPE(buffer_port)) port);
      break;
    case SN_ENUM(rf_port_kind, rf_event_port):
      set_ep_info((SN_TYPE(event_port)) port);
      break;
    default:
      break;
    }
  }
}


static void set_sp_info(SN_TYPE(simple_port) sport) {
  if (sport->adapter_specific_info == NULL) {
    simple_port_info_t* spi = (simple_port_info_t*) malloc(sizeof(simple_port_info_t));
    spi->index = global_index++;
    sport->adapter_specific_info = spi;
  }
}


static void set_ep_info(SN_TYPE(event_port) eport) {
  if (eport->adapter_specific_info == NULL) {
    eport->adapter_specific_info = (void*)strdup(sn_esi_full_path((SN_TYPE(any_port))eport) ? 
						 sn_esi_full_path((SN_TYPE(any_port))eport) : "<empty-hdl-path>");
  }
}

static void set_bp_info(SN_TYPE(buffer_port) bport) {
}


static void pre_start() {
  elab_ports();
  set_handles();
}


static void set_handles() {
  
  SN_LIST(any_port) port_list = sn_esi_get_adapter_ports(adapter_id);
  int i;
  SN_TYPE(any_port) port;
  SN_TYPE(rf_port) rfp;
  
  for (i=0; i<SN_LIST_SIZE(port_list); i++) {
    port = SN_LIST_GET(port_list, i, any_port);
    rfp = sn_rf_manager_get_port_by_instance(port);
    switch (sn_rf_port_get_kind(rfp)) {
      
    case SN_ENUM(rf_port_kind, rf_simple_port):
      break;
    case SN_ENUM(rf_port_kind, rf_buffer_port):
      set_buffer_port_handle((SN_TYPE(buffer_port))port,rfp);
      break;
    case SN_ENUM(rf_port_kind, rf_event_port):
      set_event_port_handle((SN_TYPE(event_port)) port,rfp);
      break;
    default:
      break;
    }
  }
}

/* -------------------------------------------------------- */
/* Buffer port routines                                     */
/* -------------------------------------------------------- */

void esi_buffer_get_int(SN_TYPE(buffer_port) port, SN_TYPE(int) *value) {
  BP* bp = (BP*)(port->adapter_specific_info);
  *value = (SN_TYPE(int))get_BP_element(bp);
}

void esi_buffer_get_list_of_int(SN_TYPE(buffer_port) port, 
				SN_LIST(int) value) {
  c_buffer_get_packed(port,(SN_TYPE(untyped)*) &value);
}

void esi_buffer_get_long(SN_TYPE(buffer_port) port, 
			 SN_TYPE(sn_vec) value) {
  BP* bp = (BP*)(port->adapter_specific_info);
  memcpy(SN_LONG_DATA_POINTER(value), get_BP_element(bp), SN_LONG_SIZE_BYTES(value));
}

void esi_buffer_get_list_of_long(SN_TYPE(buffer_port) port, 
				 SN_LIST(sn_vec) value) {
  c_buffer_get_packed(port,(SN_TYPE(untyped)*) &value);
}

void esi_buffer_get_string(SN_TYPE(buffer_port) port, 
			   SN_TYPE(string) *value) {
  BP* bp = (BP*)(port->adapter_specific_info);
  *value = (SN_TYPE(string))get_BP_element(bp);
  
}

void esi_buffer_get_list_of_string(SN_TYPE(buffer_port) port, 
				   SN_LIST(string) value) {
  c_buffer_get_packed(port,(SN_TYPE(untyped)*) &value);
}

void esi_buffer_get_struct(SN_TYPE(buffer_port) port, 
			   SN_TYPE(any_struct) value) {

  c_buffer_get_packed(port,(SN_TYPE(untyped)*) &value);
}


void esi_buffer_get_list_of_struct(SN_TYPE(buffer_port) port, 
				   SN_LIST(any_struct) value) {
  c_buffer_get_packed(port,(SN_TYPE(untyped)*) &value);

}


static void c_buffer_get_packed(SN_TYPE(buffer_port) port, 
				SN_TYPE(untyped) *value) {
  void  *packed_value;
  
  BP* bp = (BP*)(port->adapter_specific_info);
  
  packed_value = get_BP_element(bp);
  sn_esi_unpack((SN_TYPE(any_port)) port,
			     (SN_LIST(bit))packed_value,
			     value);
}


SN_TYPE(int) esi_get_number_of_elements(SN_TYPE(buffer_port) port) {
  return get_BP_num_of_elements((BP*)(port->adapter_specific_info));
}



void esi_buffer_put_int(SN_TYPE(buffer_port) port, SN_TYPE(int) value) {
   put_BP_element((BP*)(port->adapter_specific_info),(void*)value);
}

void esi_buffer_put_list_of_int(SN_TYPE(buffer_port) port, 
				SN_LIST(int) value) {
  c_buffer_put_packed(port,(SN_TYPE(untyped)) value);
}

void esi_buffer_put_long(SN_TYPE(buffer_port) port, 
			 SN_TYPE(sn_vec) value) {
   put_BP_element((BP*)port->adapter_specific_info, SN_LONG_DATA_POINTER(value));
}

void esi_buffer_put_list_of_long(SN_TYPE(buffer_port) port, 
				 SN_LIST(sn_vec) value) {
  c_buffer_put_packed(port,(SN_TYPE(untyped)) value);
}

void esi_buffer_put_string(SN_TYPE(buffer_port) port, 
			   SN_TYPE(string) value) {
 
  put_BP_element((BP*)(port->adapter_specific_info),(void*)strdup(value));
}

void esi_buffer_put_list_of_string(SN_TYPE(buffer_port) port, 
				   SN_LIST(string) value) {
  c_buffer_put_packed(port,(SN_TYPE(untyped)) value);
}


void esi_buffer_put_struct(SN_TYPE(buffer_port) port, SN_TYPE(any_struct) value) {
  c_buffer_put_packed(port,(SN_TYPE(untyped)) value);
}

void esi_buffer_put_list_of_struct(SN_TYPE(buffer_port) port, 
				   SN_LIST(any_struct) value) {
  c_buffer_put_packed(port,(SN_TYPE(untyped)) value);

}


static void c_buffer_put_packed(SN_TYPE(buffer_port) port, 
			   SN_TYPE(untyped) value) {
  SN_LIST(bit) packed_value = 
    sn_esi_pack((SN_TYPE(any_port)) port,value);
  put_BP_element((BP*)(port->adapter_specific_info),(void*)packed_value);
}

void esi_notify_number_of_elements(SN_TYPE(buffer_port) port, SN_TYPE(int) num) {
  notify_BP_num_of_elements(((BP*)(port->adapter_specific_info)),num);
}


SN_TYPE(int) esi_get_number_of_free_elements(SN_TYPE(buffer_port) port) {
  return  get_BP_num_of_free_elements((BP*)(port->adapter_specific_info));
}


static void set_buffer_port_handle(SN_TYPE(buffer_port) bport,
				   SN_TYPE(rf_port) rfp) {
  BP *bp;
  short input = 0;
  SN_TYPE(rf_type) rft;

  if (bport->adapter_specific_info) {
    return;
  }
  
  switch (sn_rf_port_get_direction(rfp)) {
  case SN_ENUM(sn_port_direction,in):
    input = 0;
    break;
  case SN_ENUM(sn_port_direction,out):
    input = 1;
    break;
  default:
    //Error
  }
  
/*
  if (sn_rf_type_get_kind(rft = sn_rf_port_get_element_type(rfp)) == SN_ENUM(rf_type_kind, rf_scalar_type)) {
    bport->adapter_specific_info = 
      create_BP((sn_rf_scalar_get_size_in_bits((SN_TYPE(rf_scalar)) rft)-1)/8 + 1,
                       (int)sn_esi_get_port_attribute((SN_TYPE(any_port)) bport,"buffer_size"),
                       input);
    
    // Just for the fun
    sn_esi_get_packed_size((SN_TYPE(any_port))bport,);
  }
  else {
*/
    bport->adapter_specific_info = 
      // create_BP(sn_esi_get_packed_size((SN_TYPE(any_port))bport);
      //                   sn_esi_get_port_attributes((SN_TYPE(any_port)) bport,"size"), input) 
      create_BP(sizeof(SN_LIST(bit)),
                       (int)sn_esi_get_port_attribute((SN_TYPE(any_port)) bport,"buffer_size"),
                       input);
    
    // Just for the fun
    sn_esi_get_packed_size((SN_TYPE(any_port))bport);
  //}
}



/* -------------------------------------------------------- */
/* Event port routines                                     */
/* -------------------------------------------------------- */
void esi_emit(SN_TYPE(event_port) eport) {
  c_simulator_emit((char*)eport->adapter_specific_info);
}



void handle_event_cb() {
  int i;
  for(i=0;i<ep_index;i++) {
    if (OUTEPA[i] != -1 ) {
      sn_esi_change_occurred(adapter_id,OUTEPA[i]);
    }
  }
}
static void set_event_port_handle(SN_TYPE(event_port) eport,
				  SN_TYPE(rf_port) rfp) {

 
  switch (sn_rf_port_get_direction(rfp)) {
  case SN_ENUM(sn_port_direction,in): 
    c_simulator_set_cb(handle_event_cb);
    OUTEPA[ep_index] = eport->port_id;
    break;
  case SN_ENUM(sn_port_direction,out):
  case SN_ENUM(sn_port_direction,inout):
    OUTEPA[ep_index] = -1;
    break;
  default:
    //Error
  }
  ep_index++;
}






#include <stdarg.h>

static char buffer[200];
static char* my_sprintf(char* format, ...) {
  va_list ap;
  char* result;

  va_start(ap, format);
  vsprintf(buffer, format, ap);
  result = (char*) malloc(strlen(buffer)+1);
  strcpy(result, buffer);
  return result;
}
