C simulator adaptor

<'

struct c_simulator_port_info like esi_adaptor_port_info {
  !index: uint;
};


struct c_simulator_adaptor like esi_adaptor {

  current_index: uint;

  get_index(c: simple_port): uint is {
    result = current_index;
    current_index += 1;
  };


  -- used to force opening of adaptor C library
  !dummy_call: C agent_call();

  -- Do elaboration for this adaptor. Called before both write_stubs()
  -- and start_test(), for all relevent adaptors, with same parameters.
  elaborate(units: list of any_unit, ports: list of any_port) is {
    for each simple_port (sp) in ports {
       var info:= new c_simulator_port_info;
       info.index = get_index(sp);
       sp.adaptor_info = info;
    };
    dummy_call = new;
    dummy_call.lib_name = "c_simulator_adaptor.so";
    dummy_call.call_name = "c_simulator_create_list_bits";
    dummy_call.lazy_mode = now;
    dummy_call.scope_mode = global;
    dummy_call.resolve();
  };


  !get_bits_by_index: C agent_call (i: int): list of bit;
  
  get_bits(c: simple_channel): list of bit is {
      result = get_bits_by_index.call(c.ports[0].adaptor_info.as_a(c_simulator_port_info).index);
  };

  !put_bits_by_index: C agent_call (i: int, l: list of bit);
  
  put_bits(l: list of bit, c: simple_channel) is {
      put_bits_by_index.call(c.ports[0].adaptor_info.as_a(c_simulator_port_info).index, l);
  };

  write_stubs(units: list of any_unit, ports: list of any_port) is {
    var externs: list of string;
    var gets: list of string;
    var puts: list of string;
    var full: list of string;

    for each simple_port (sp) in ports {
      externs.add(appendf("extern int %s;", sp.path));
      gets.add({appendf("    case %d:", sp.adaptor_info.as_a(c_simulator_port_info).index) ;
                appendf("      return c_simulator_create_list_bits(&%s);", sp.path) ; "      break;"});
      puts.add({appendf("    case %d:", sp.adaptor_info.as_a(c_simulator_port_info).index) ;
                appendf("      return c_simulator_store_list_bits(&%s, l);", sp.path) ; "      break;"});
    };

    full.add("#include \"specman_.h\"");
    full.add("");
    full.add("extern SN_LIST(bit) c_simulator_create_list_bits(int*);");
    full.add("extern void c_simulator_store_list_bits(int*, SN_LIST(bit));");
    full.add("");
    full.add(externs);
    full.add("");
    full.add("SN_LIST(bit) c_simulator_get_bits(int i) {");
    full.add("  switch (i) {");
    full.add(gets);
    full.add("    default:");
    full.add("      return NULL;");
    full.add("  }");
    full.add("}");
    full.add("");
    full.add("void c_simulator_put_bits(int i, SN_LIST(bit) l) {");
    full.add("  switch (i) {");
    full.add(puts);
    full.add("    default:");
    full.add("      return;");
    full.add("  }");
    full.add("}");
    esi_manager.append_stub("c_simulator_stub.c", full);
  };

  init() is also {
      name = "c_simulator";
      get_bits_by_index = new;
      get_bits_by_index.lib_name = "c_simulator_stub.so";
      get_bits_by_index.call_name = "c_simulator_get_bits";
      put_bits_by_index = new;
      put_bits_by_index.lib_name = "c_simulator_stub.so";
      put_bits_by_index.call_name = "c_simulator_put_bits";
  };
};

extend esi_manager {
  init() is also {
      add_adaptor(new c_simulator_adaptor);
  };
};

'>
