Seamless adaptor
----------------

<'
extend simple_port {
  cpu: string;
  keep soft cpu == "";

  address_space: string;
  keep soft address_space == "";

  big_endian: bool;
  keep soft big_endian == FALSE;
};

struct seamless_port_info like esi_adaptor_port_info {
  !connection: seamless_connection;
};


struct seamless_adaptor like esi_adaptor {

  connections: list (key:cpu) of seamless_connection;
  name_index: int;


  -- Do elaboration for this adaptor. Called before both write_stubs()
  -- and start_test(), for all relevent adaptors, with same parameters.
  elaborate(units: list of any_unit, ports: list of any_port) is {
    for each simple_port (sp) in ports {
       var info:= new seamless_port_info;
       info.connection = get_connection(sp);
       sp.adaptor_info = info;
    };
  };


  get_connection(p: simple_port): seamless_connection is {
    result = connections.key(p.cpu);
    if (result == NULL or result.addr_space != p.address_space) {
      name_index += 1;
      result = seamless.get_connection(append("con_", name_index), p.cpu, p.address_space);
      result.set_as_big_endian(p.big_endian);
      connections.add(result);
    }
  };

  get_bits(c: simple_channel): list of bit is {
      if (c.ports[0].full_path ~ "/~?\/(.*)/") {
        c.ports[0].full_path = $1;
      };
      result = seamless.get_list(c.ports[0].adaptor_info.as_a(seamless_port_info).connection.tag, c.ports[0].full_path);
  };

  put_bits(l: list of bit, c: simple_channel) is {
      if (c.ports[0].full_path ~ "/~?\/(.*)/") {
        c.ports[0].full_path = $1;
      };
      seamless.set_list(c.ports[0].adaptor_info.as_a(seamless_port_info).connection.tag, c.ports[0].full_path , l);
  };

  init() is also {
      name = "seamless";
  };
};

extend esi_manager {
  init() is also {
      add_adaptor(new seamless_adaptor);
  };
};
'>

