Verilog adaptor
---------------

o Restrictions in stub creation:

  - ## Support for only "drive" and "drive_hold".

  - ## No support for when

  - ## No checking of bad params

  - ## No support for computed names

  - ## No source reference in stub

<'

extend sn_agent_manager {
//##VY Agreed with Yoav to rename agent's in internal identifiers to adaptors - later
    new_agent(actual_agent_name : string, agent_name : string,
        adaptor_lib : string) : sn_simulator_agent is only {

        result = new_adaptor(actual_agent_name);
        
        // A real new agent
        result.name = agent_name;            // Non-normalized name
        result.aliases.add(actual_agent_name); // Must be the first to be added
        result.adaptor_lib = adaptor_lib;
        result.sim_product = sim_product(actual_agent_name);
        
        // Update the master agent (should be done before sn_simulator_exists()
        // is called)
        master_agent = (master_agent == NULL) ? result : master_agent;
       
        // Init the agent (C agent calls etc.)
        result.init_agent(not sn_simulator_exists());
        
        // Check that the agent library external pointer is unique
        // Otherwise this is just a new alias
        result.set_id();
        
        var prev_agent : sn_simulator_agent = agents.first(it.id == result.id);
        
        // Have we found a previous agent which has the same ID?
        // Distinguish between null simulator and statically linked
        // simulator
        // We assume here that SN_SIM will be found by it's name
        // in previous checks
        if ((prev_agent != NULL) and (prev_agent.name != SN_SIM)
            and (agent_name != SN_SIM))then {
            // No need to create new agent
            // Update aliases list
            prev_agent.aliases.add(actual_agent_name);
            return prev_agent;
        }
        else {
            // Register the agent in cross-restore list
            result.save_registered_agent();
            // Add the agent to the agent list
            agents.add(result);
        };
		
        // Set this for compatibility with the old code
        simulator.exists = sn_simulator_exists();
        simulator.sim_product = master_agent.sim_product;
        simulator.product = get_old_product(simulator.sim_product);
        simulator.kind = master_agent.is_verilog() ? verilog : 
                         master_agent.is_vhdl() ? vhdl : unknown;
    };
    
    new_adaptor(name: string): sn_simulator_agent is {
        if result == NULL {
            if is_verilog_name(name) {
                result = new verilog_adaptor;
            } else if is_vhdl_name(name) {
                result = new vhdl_adaptor;
            } else {
                result = new sn_simulator_agent;
            };
        };
    };
    
    is_verilog_name(name: string): bool is {
        // For possible extensions 'is first'
        if result == FALSE {
            result = (name ~ "/VERILOG/") or 
                (name ~ "/MTI_VLOG/") or 
                (name ~ "/NCVLOG/") or 
                (name ~ "/VCS/") or 
                (name ~ "/XL/") or 
                (name ~ "/SPD/");
        };
    };
    
    is_vhdl_name(name: string): bool is {
        if result == FALSE {
            result =  (name ~ "/VHDL/") or 
                (name ~ "/MTI_VHDL/") or 
                (name ~ "/NCVHDL/")   
        };
    };
};
    
extend simple_port {
    !drive: string;   -- The drive param (as in "verilog variable")
    !drive_hold: string; -- The drive_hold param (as in "verilog variable")
};


struct verilog_adaptor like sn_simulator_agent {
    
    -- Create the stubs for this adaptor by calling the append_stub() method.
    -- Called upon "write stubs -esi" for all relevant adaptors.
    write_stubs(units: list of any_unit, ports: list of any_port) is {

	var l: list of string;  -- For the e code
	
	-- Create "verilog variable" for each channel that needs it
	for each  (p) in ports do {
            
            if (p is a simple_port){ 
                
                var sp:= p.as_a(simple_port);
                
                -- Create a "verilog variable" statement for this signal if needed
                var options: list of string;
                if sp.drive != "" then {
                    options.add(append("drive=", quote(sp.drive)));
                };
                if sp.drive_hold != "" then {
                    options.add(append("drive_hold=", quote(sp.drive_hold)));
                };
                if not options.is_empty() then {
                    
                    var unit_name:= sp.get_enclosing_unit(any_unit).type().name;
                    -- ## Note: probably need a builtin, unit_name() method
                    l.add({
                        append("extend ", unit_name, " {");
                        append("    verilog variable '", sp.hdl_path(), "' using ", 
                            str_join(options, ", "), ";");
                        "};";
                    });
                };
            } else if (p is a call_port) {
                
               var cp := p.as_a( call_port);
               for each (m) in cp.methods {
                   add_method_code(cp, m);
               }; 
            };
        };

	-- If the list is not empty, add it to stubs.e.
	-- Always create a "write stubs -verilog" command
	if not l.is_empty() then {
	    l = {"\n<'"; l; "'>\n"};
	    simulator.agent_manager.append_stub("stubs.e",l);
	};
        simulator.agent_manager.append_stub("stubs.ecom", {"write stubs -verilog"});
        
    };

    
    add_method_code(p: call_port, m: port_method_struct) is {
	
        var key:= append("verilog subprogram ", m.method_name);
        var definition_in_unit:= verilog_subprogram_definition_string(p,m); 
        
        -- Add to the parent unit the  subprogram definition 
        var l:={"\n<'";definition_in_unit; "'>\n"};
        
        simulator.agent_manager.append_stub("stubs.e",l);
            
        -- Add the method category, if needed
        if not simulator.agent_manager.has_method_code(m, key) then {
            var code:= verilog_subprogram_call_string(p,m);
	    p.add_method_code(m,key,code);
        };
    };
    
    
    verilog_subprogram_call_string(p: call_port, m: port_method_struct):string is{
        
        var is_return         : string;
        var subprogram_path   : string;
        var subprogram_name   : string;
        var subprogram_params : string;

        
        subprogram_path = p.hdl_path();
        subprogram_name = m.method_name;
        
        var rf_method := rf_manager.get_struct_by_instance(m).get_method("call");
        var rf_params := rf_method.get_parameters();
        var rf_result := rf_method.get_result_type();
        
        for each (par)in rf_params do{
            if(index == 0){
                subprogram_params = par.get_name(); 
            }else{   
                subprogram_params = append(subprogram_params,",",par.get_name());
            };
        };
       
        if(rf_method.get_result_type() == NULL){
            is_return = "" ;
        }else{
            is_return = "return " ;
        };
            
        var s := append (is_return,"'",subprogram_path,".",subprogram_name,"'(",subprogram_params,");");
        return s;
    };
    
    
    
    verilog_subprogram_definition_string(p: call_port, m: port_method_struct):string is{
        
        var task_or_function           : string;
        var subprogram_path            : string;
        var subprogram_name            : string;
        var subprogram_params_and_mode : string;
        var subprogram_result          : string;  // with ':'
        
        
        subprogram_path   = p.hdl_path();
        subprogram_name   = m.method_name;

        -- Create the parametrs string
        var params := rf_manager.get_struct_by_instance(m).get_method("call").get_parameters();
        var param_bit_size: uint;
        
        for each (par)in params do{

            // ##VY How to check it is a scalar?
            // if (!par.get_type().is_scalar()){
            //     error("Only scalar parameters are supported for Verilog call ports"); 
            // } else { }
            param_bit_size = par.get_type().as_a(rf_scalar).get_size_in_bits();

            var mode :string;
            
            if(par.is_by_reference()){
                mode = "inout";
            }else{
                mode = "in";
            };
            
            var s :=  append(par.get_name(),":",param_bit_size,":",mode);   
            if(index == 0){
                subprogram_params_and_mode = s;
            }else{   
                subprogram_params_and_mode = append(subprogram_params_and_mode,",",s);
            };
            
        };
        
        -- Create the result string , and find if the subprogam is task or function 
        var res_type:= rf_manager.get_struct_by_instance(m).get_method("call").get_result_type();
        if(res_type != NULL){
            
            task_or_function = "function";
            subprogram_result = append(":" ,res_type.as_a(rf_scalar).get_size_in_bits().as_a(string));
          //  subprogram_result = append(":" ,res_type.get_size_in_bits().as_a(string));
          //  if(!res_type.is_scalar()){
                //error!!!              
         //   };
                
        }else{
            
            task_or_function = "task";
        };
        

        var s := append("verilog ",task_or_function," '",subprogram_path ,".",subprogram_name,"'(",subprogram_params_and_mode,")",subprogram_result ,";");
        
        return s;
        
    };
    
    
    get_description(port: any_port): list of string is {
	if port is a simple_port (sp) then {
	    return {
	        append("path=", sp.hdl_path());
		append("full_path=", sp.full_hdl_path());
		append("wire=", sp.wire);
		-- ## Finish all
	    };
	};
    };
};

struct vhdl_adaptor like sn_simulator_agent {
};

'>

o Agent stuff

<'

struct sn_hdl_agent_port_struct like any_agent_port_struct {
    !tms: translated_model_struct;
    !vec: sn_vec;
    
    // ##VY I copy all the below fields from the port to speed up the run-time
    // ##VY It is temporary because ideally these fields must be used only once
    
    data_type : hdl_data_type;
    full_name : string;
    port_type : type_descriptor;

    get_simple_port(port: simple_port): untyped is {
        // ##VY Add support for trace sim_read/sim_write
        
        if tms == NULL {
            error ("Internal error: Trying to access sn_hdl_agent_port when tms is NULL");
        } else {
            //  var uint_res: uint = tms.get_sig(full_path, vec);
            
            if (session.update_count != tms.gput_count) { // ##VY should be avoided for memories
                tms.fetch_value(TRUE); 
                tms.gput_count = session.update_count;
            };
            
            var uint_res: uint = tms.get_sig(full_name, vec);
            
            if (session.sim_read_traced) {
                simulator.trace_get_model(port.hdl_path(),full_name,uint_res,vec);
            };
            
            if data_type == uint_type {
                result = uint_res.unsafe();
            } else if data_type == long_type {
                result = vec.unsafe();
            } else if data_type == bit_list_type {
                result = vec.bytes.unsafe();
            } else if data_type == list_type {
                var res:= make_new_list(port_type.d_index); // ##VY Should be a more effective way
                
                compute packing.unpack_item(packing.low,vec.bytes.unsafe(),
                    res.unsafe(),port_type,0,FALSE);
                result = res.unsafe();
            };
        };
    };
    
    
    put_simple_port(port: simple_port,the_value :untyped) is{

        if tms == NULL {
            error ("Internal error: Trying to access sn_hdl_agent_port when tms is NULL");
        } else {
              
            if (session.update_count != tms.pput_count) {
                if (not tms.name_par.to_release) {   
                    tms.fetch_value(FALSE); 
                    tms.pput_count = session.update_count;
                    if (!simulator.to_set_driver_now ) { // Mirit//??
                        simulator.add_put_to_list(tms);
                    }; 
                };
            };
            
            if (session.sim_write_traced) {
                var str_value: string;
                if data_type != uint_type {
                    str_value = vec.numeric_to_string(print_options);
                    // ##VY Does not work for x/z literals (see hdl_access.e
                } else {
                    str_value = scalar_to_string(the_value.unsafe(),type(uint),FALSE);
                };
                session.on__sim_write(port.hdl_path(),full_name,str_value);
            };
            if data_type == uint_type {
                tms.put_sig(full_name,the_value.unsafe(),0.unsafe()); //##VY move call outside of if's
                
            } else if data_type == long_type{

                tms.put_sig(full_name,0,the_value.unsafe()); -- the_value is already long because of the formal parameter type
                
            } else if data_type == bit_list_type {
                
                vec.bytes = the_value.unsafe();
                tms.put_sig(full_name,0,vec);
                
            } else if data_type == list_type {
                packing.pack_item(packing.low,vec.bytes,the_value,port_type,TRUE);
                tms.put_sig(full_name,0,vec);
            };
        }; 
    };
    
    
    elaborate(port: any_port) is {

        if(port.external){
            full_name = simulator.normalize_hdl_path(port.full_hdl_path());
            
            // ##VY check that everything here is correct: separtors, normalization, decanonization etc.
            tms = simulator.add_translated_struct(port.hdl_path(), port.get_parent_unit(), full_name, FALSE, "");
            // ##VY add source line above
            
            var info       : hdl_exp_info = new;
            var curr_calc  : curr_calc    = new;
            var name_param : name_param   = new;
            
            tms.info      = info;
            tms.curr_calc = curr_calc;
            tms.name_par  = name_param;
            tms.curr_calc.calc_not_slice(tms,tms.min,tms.max,tms.size); // ##VY This must be a method of tms
            
            
            // ##VY Probably we can move data_type field from port to this struct
            data_type = port.as_a(simple_port).data_type;
            if data_type in [long_type..list_type] {
                vec = new;
                vec.attributes = 0;
            } else if data_type == other {
                error ("Unsupported port type for simulator agent ", port.internal_get_agent().name());
            };
            port_type = port.as_a(simple_port).element_type;
        };
    };
};

// Mirit - When you write stub -esi the agent is NULL-SIM - The real agent is only a string in the port ??
extend sn_simulator_agent {
    
    new_agent_port_struct(port: any_port) : any_agent_port_struct is {
        // ##VY This is a temporary solution
        if(port.external){
            result = new sn_hdl_agent_port_struct;
        };

'>
##VY Once we provide during write stub real agents (and not NULL_SIM) we will be able to do the
     following:

     if (is_verilog() or is_vhdl()) {
         result = new sn_hdl_agent_port_struct;
     } else {
         error ("Internal error: new_agent_port_struct() is not implemented for agent ", me.name());
     };
<'
    };

};

'>
