#! /bin/csh -f

#----------------------------------------------------------------------
# Check SPECMAN_HOME 
#----------------------------------------------------------------------
if (! ($?SPECMAN_HOME)) then
    echo "*** Error: env. var. SPECMAN_HOME is not set."
    exit -1
endif



# get jspecview java exe from env var JSPECVIEW_JAVA_EXE: one of:
# JRE1.1   :  sn_jspecview.sh will run java 1.1 from jre
# JRE1.2   :  sn_jspecview.sh will run java 1.2 from jre
# DISABLED : sn_jspecview.sh will do nothing
# the default is JRE1.1 
if (! ($?JSPECVIEW_JAVA_EXE)) then
  set JSPECVIEW_JAVA_EXE = "JRE1.1"
endif

# get jspecview main classpath from env var JSPECVIEW_CP_TYPE: one of:
# JAR
# CLASS
# CLASS_D
# the default is JAR
if (! ($?JSPECVIEW_CP_TYPE)) then
  set JSPECVIEW_CP_TYPE = "JAR"
endif


# ------------------------------------------
# init general vars
# ------------------------------------------
set is_verbose = 0
set tmp_file="/tmp/_sn_help_tmp.$$"
set tick="'"
set quote='"'
set qmark='?'
set comma=','
set ampersent = '&'
set cmd_line = "$0 $*"
set err_msg = ""
set good_help = 0
set internal_use = 0
set curr_dir = `pwd`
set user_classpath = ""
if ($?JSPECVIEW_VERBOSE) then
  set is_verbose = 1
endif

if ($is_verbose) echo "pwd:"
if ($is_verbose) echo "V> CMD: $cmd_line"

# ------------------------------------------
# init specific vars
# ------------------------------------------
set portfile = ""
set comm_ready_file = ""
set port    = ""
set parent_pid = ""
set host = ""
set jvm_params = ""
set appl_params = ""
set minimal_args = ""
#=======================================================
#===   first loop - get options  =======================
#=======================================================
set argc = $#argv
while ($argc > 0)
    # get first character
    set flag = $1

    switch( $flag )
        case -h:
        case -help:
            set good_help = 1
	    goto USAGE
            breaksw
       
        case -verbose:
        case -v:
	    shift
	    set is_verbose = 1
            breaksw

        case -minimal:
	    shift
            set appl_params = "$appl_params -minimal"
            breaksw


        case -jverbose:
        case -jv:
            shift
            set jvm_params = "$jvm_params -verbose"
            breaksw

        case -internal_use:
            shift
            set internal_use = 1
            breaksw

	case -portfile:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set portfile = "$1"

            set appl_params = "$appl_params -portfile $portfile"

            shift
            breaksw

	case -comm_ready_file:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set comm_ready_file = "$1"

            #set appl_params = "$appl_params -comm_ready_file $comm_ready_file"

            shift
            breaksw

	case -user_classpath:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set user_classpath = "$1"

            shift
            breaksw

	case -prop_file:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set prop_file = "$1"

            set appl_params = "$appl_params -prop_file $prop_file"

            shift
            breaksw

	case -temp_dir:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set temp_dir = "$1"

            set appl_params = "$appl_params -temp_dir $temp_dir"

            shift
            breaksw

	case -port:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set port = "$1"
            set appl_params = "$appl_params -port $port"
            shift
            breaksw

	case -parent_pid:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set parent_pid = "$1"
            set appl_params = "$appl_params -parent_pid $parent_pid"
            shift
            breaksw

	case -host:
            shift
	    if ("$1" == "") then
	      goto USAGE
            endif
	    set host = "$1"
            set appl_params = "$appl_params -host $host"
            shift
            breaksw


        default:
            # extra parameters
            set appl_params = "$appl_params $1"
            breaksw

    endsw
    set argc = $#argv
end

#=======================================================
#===   END first loop            =======================
#=======================================================

START:

if ("$JSPECVIEW_JAVA_EXE" == "DISABLED") then
  echo "Note: sn_jspecview is currently disabled. Please start JSpecview manually."
  exit 0
endif


if (  ("$portfile" == "") && ("$port" == "") ) then
  echo "*** No port-file nor port-number were specified"
  goto USAGE
endif
 
if (  ("$portfile" != "") && ("$port" != "") ) then
  echo "*** You cannot specify both port-file and port-number"
  goto USAGE
endif
 
if ("$host" == "") then
  set host = `hostname`
  set appl_params = "$appl_params -host $host"
endif
# SPECMAN_HOME
set appl_params = "$appl_params -bin_dir $SPECMAN_HOME/bin"

if ($?JSPECVIEW_LNF) then
    set appl_params = "$appl_params -lnf $JSPECVIEW_LNF"
    echo "Using $JSPECVIEW_LNF Look-And-Feel"
endif

ENVIRONMENT:
  set specman_bin = $SPECMAN_HOME/bin
  set specman_arch_sh = $specman_bin/sn_arch.sh
  if (! -e $specman_arch_sh) then 
    echo "*** Can't find utility file: $specman_arch_sh"
    goto USAGE
  endif

  # get OS
  set sn_os = `$specman_arch_sh`
  if ( ($sn_os != "solaris") && ($sn_os != "hpux") && ($sn_os != "linux")) then
    echo "*** Platform '$sn_os' is not supported."
    goto USAGE
  endif
  if ($sn_os == "hpux") then
            set jvm_params = "$jvm_params -green"
  endif
  if ($is_verbose) echo "V> sn_os: $sn_os"

# run with/without "-nojit"
# default : on linux      -  use -nojit
#           on solaris/hp -  don't use -nojit.

set use_jit = 1
if (($?SN_JAVA_DONT_USE_JIT) || ($sn_os == "linux")) then
  set use_jit = 0
endif
 

if ($use_jit == "0") then 
  set jvm_params = "$jvm_params -nojit"
endif 

# if using JRE1.1 decied which jre to run (on linux, there are 2 available jres...)

if($?JSPECVIEW_JRE) then 
  set jre_name = "$JSPECVIEW_JRE"
else  if (-e $SPECMAN_HOME/japp/$sn_os/which_jre)then          
          if($is_verbose) then
             echo "running tje jre  :  " 
             echo `$SPECMAN_HOME/japp/$sn_os/which_jre`
          endif 
          set jre_name = `$SPECMAN_HOME/japp/$sn_os/which_jre`
      else
          set jre_name = "jre"
      endif       
endif
  



JAVA:

set jdev_name = "jdev"
if ($?JSPECVIEW_JDEV) then
  set jdev_name = $JSPECVIEW_JDEV
endif


set japp = $SPECMAN_HOME/japp
set jdev = $SPECMAN_HOME/$jdev_name
set OTHER_LIBS = ""
set ICE_LIBS   = "${japp}/jar/ice5.06.jar"

if ("$JSPECVIEW_JAVA_EXE" == "JRE1.3") then
  echo "Running JAVA 1.3"
  setenv JAVA_HOME "/proj/Spider/java/jre13/solaris/j2re1_3_0beta"
  set JAVA = "$JAVA_HOME/bin/java"
  set JAVA_LIBS = ""
  set OTHER_LIBS = ""
else 
  if ("$JSPECVIEW_JAVA_EXE" == "JRE1.2") then
    echo "Running JAVA 1.2"
    setenv JAVA_HOME "/proj/Spider/jbuilder30/jdk1.2.2/jre"
    set JAVA = "$JAVA_HOME/bin/java"
    set JAVA_LIBS = ""
    set OTHER_LIBS = ""
  else 
    setenv JAVA_HOME "$japp/$sn_os/$jre_name"
    set JAVA = "$JAVA_HOME/bin/jre"
    set JAVA_LIBS = "$JAVA_HOME/lib/rt.jar"
    set OTHER_LIBS = "$japp/jar/swing.jar{:}$japp/jar/motif.jar{:}$japp/jar/windows.jar"
  endif
endif

#if ("$JSPECVIEW_JAVA_EXE" == "PROFILER") then
#  echo "Running JAVA with profiler"
#  setenv JAVA_HOME "/proj/Spider/jbuilder30/jdk1.2.2/jre"
#  set JAVA = "$JAVA_HOME/bin/java"
#  set JAVA_LIBS = ""
#  set jvm_params = "$jvm_params -green -Djava.compiler= -Xrunhprof:cpu=samples,depth=6,thread=y" 
#endif



#java exec
if ($is_verbose) echo "V> JAVA: $JAVA"

if (! -e "$JAVA") then
  echo "*** Can't find java run-time file at: $JAVA"
  goto USAGE
endif

# main class
set jspecview_class = "com.verisity.jspecview.JSpecview";
if ($is_verbose) echo "V> JSPECVIEW_CP_TYPE $JSPECVIEW_CP_TYPE"

if ($?JSPECVIEW_JAR_FILE) then
    set jspecview_class_file = "$JSPECVIEW_JAR_FILE"
    set JSPECVIEW_LIBS = "$JSPECVIEW_JAR_FILE"
    if ($is_verbose) echo "USING JAR: $JSPECVIEW_JAR_FILE"
else
    switch( "$JSPECVIEW_CP_TYPE" )
      case "JAR":
        set jspecview_class_file = "$japp/jspecview/jar/jspecview.jar"
        set JSPECVIEW_LIBS = "$japp/jspecview/jar/jspecview.jar"
        breaksw
  
      case "CLASS":
        set jspecview_class_file = "$jdev/class/com/verisity/jspecview/JSpecview.class";
        set JSPECVIEW_LIBS = "$jdev/class"
        breaksw
    
      case "CLASS_D":
        set jspecview_class_file = "$jdev/class_d/com/verisity/jspecview/JSpecview.class";
        set JSPECVIEW_LIBS = "$jdev/class_d"
        breaksw
    
      default:
        echo "*** Unrecognized value for JSPECVIEW_CP_TYPE: $JSPECVIEW_CP_TYPE"
        goto USAGE
    
    endsw
endif

if (! -e "$jspecview_class_file") then
  echo "*** Can't find java class file $jspecview_class_file"
  goto USAGE
endif



if ($?JSPECVIEW_RES_LIB) then
    set RESOURCE_LIB = ${JSPECVIEW_RES_LIB}
    if ($is_verbose) echo "USING RESOURCES: $RESOURCE_LIB"
else
    set RESOURCE_LIB = "$SPECMAN_HOME/resources{:}$SPECMAN_HOME/resources/com.verisity.utils.ui{:}$japp/resource{:}$japp/ipcover/resource"
endif
set classpath = "$JSPECVIEW_LIBS{:}$JAVA_LIBS{:}$RESOURCE_LIB{:}$ICE_LIBS{:}$OTHER_LIBS"

# add to classpath
if ($?SN_USER_CLASSPATH) then
  set classpath = $classpath{:}$SN_USER_CLASSPATH
endif

if ("$user_classpath" != "") then
  set classpath = $classpath{:}$user_classpath
endif




if ($is_verbose) echo "V> classpath: $classpath"

# run it!
if ($is_verbose) then
  echo "V> Running:"
  echo "V>   $JAVA $jvm_params -classpath $classpath $jspecview_class $appl_params"
endif

if ($?JSPECVIEW_LOG) then
  echo "Redirecting output to: $curr_dir/$JSPECVIEW_LOG"
  (cd $japp; $JAVA $jvm_params -classpath $classpath $jspecview_class $appl_params > $curr_dir/$JSPECVIEW_LOG &)
else

  (cd $japp; $JAVA $jvm_params -classpath $classpath $jspecview_class $appl_params >& /dev/null &)
endif

exit 0

USAGE:
if ($internal_use == 0) then
    echo ""
    echo "Usage: "
    echo ""
    echo " sn_jspecview.sh   -portfile <portfile-name> | -port <port-num>   [-host <host-name>] [-prop_file <properties-file>]"
    echo ""
    echo " -h : Shows this help screen "
    echo ""
    echo " To set JAVA exe use:"
    echo "  setenv JSPECVIEW_JAVA_EXE JRE1.1"
    echo "  setenv JSPECVIEW_JAVA_EXE JRE1.2"
    echo "  setenv JSPECVIEW_JAVA_EXE DISABLED"
#    echo "  setenv JSPECVIEW_JAVA_EXE PROFILER"
    echo ""
    echo " To control classpath use:"
    echo "  setenv JSPECVIEW_CP_TYPE JAR"
    echo "  setenv JSPECVIEW_CP_TYPE CLASS"
    echo "  setenv JSPECVIEW_CP_TYPE CLASS_D"
    echo "  setenv JSPECVIEW_JDEV xxx"
    echo " To set resources:"
    echo "  setenv JSPECVIEW_RES_LIB resource-dir"
    echo ""
    echo " To set verbose on:"
    echo "   setenv JSPECVIEW_VERBOSE TRUE"
    echo " To set log file :"
    echo "   setenv JSPECVIEW_LOG <file-name>"
    echo " To set JSpecview auto-launch :"
    echo "   setenv JSPECVIEW_AUTO TRUE"
    echo " To set extra parameters for sn_jspecview.sh :"
    echo "   setenv JSPECVIEW_EXTRA <extra-args>"
    echo " To set special Look & Feel (window, motif, metal)"
    echo "   setenv JSPECVIEW_LNF metal"
    echo " To add user's classpath:"
    echo "   setenv SN_USER_CLASSPATH <classpath>"
    echo ""

    echo " To unsetenv all the above:"
    echo "   unsetenv JSPECVIEW_*"

endif

if ($good_help) then
  echo 444
  exit 0
endif
echo 333
exit -1
