<'

'>
The following data structure represents the reflective information of Specman type system
<'

-- semantical entity in the type system
struct entity_rep like sn_struct {
    comment: string;
    containing_module: string;
    
    print_doc() is empty;
};

-- a type entity with a name
struct named_entity_rep like entity_rep {
    name: string;
};

-- struct representation
struct struct_rep like named_entity_rep {
    parent: string;
    field_list: list of field_rep;       
    method_list: list of method_rep;
    extention_list: list of entity_rep;
    
    print_doc() is {
        for each (f) in field_list do {
            f.print_doc();
        };
        for each (m) in method_list do {
            m.print_doc();
        };
    };
};

-- method representation
struct method_rep like named_entity_rep {
    return_type: string;
    parameter_list: list of parameter_rep;
    extention_list: list of entity_rep;
    
    print_doc() is {
        for each (p) in parameter_list do {
            p.print_doc();
        };
    };
};

-- method parameter representation
struct parameter_rep like named_entity_rep {
    type_name: string;
};

-- field representation
struct field_rep like named_entity_rep  {
    type_name: string;
    property_kind: [attribute, association];
    cardinality: [scalar, vector];
};


'>
The following is an extention of Specman type system structs to support creation of reflective information.
<'

extend source_item {
   
    is_relevant_for_doc(): bool is {
        result = dcm.is_relevant_for_doc(me);
    };
    
    extract_comment(): string is {
        try {
		var full_path: string = source_module().file_name;
		var srcline: string =  source_line_no(full_path,source_line());
		if str_match(srcline,"/^.*(--|//)(.*)\n$/") { 
		    -- item has rest-of-line comment
		    result = $2;
		    return result;
		};
        
		srcline = source_line_no(full_path,source_line()-1);
		if str_match(srcline,"/^[ \t]*(--|//).*$/") {
		    -- read consequent commented lines in upward order, break upon non-comment
		    for curr_line from source_line()-1 down to 1 do {
			srcline = source_line_no(full_path,curr_line);
			if !str_match(srcline,"/^[ \t]*(--|//)(.*)\n$/") { 
			    break;
			};
			result = append($2," ",result);
		    };
		    return result;
		};
        
		if str_match(srcline,"/^[ \t]*<'[ \t]*\n$/") {
		    -- read the whole text segment 
		    for curr_line from source_line()-2 down to 1 do {
			srcline = source_line_no(full_path,curr_line);
			if str_match(srcline,"/^[ \t]*'>[ \t]*\n$/") { 
			    break;
			};
			result = append(srcline,result);
		    };
		};
	} else {
	       result = "";
	};
    };
	    
    init_entity_rep(er: entity_rep): entity_rep is {
        er.comment = extract_comment();
        er.containing_module = source_module().name;
        result = er;
    };
};

extend struct_descriptor {
    
    get_rep(): struct_rep is {
        result = init_entity_rep(new struct_rep).as_a(struct_rep);
        result.name = name;
        result.parent = parent.name;
        
        for each (ex) in extensions_refs do {
            var si: source_item = new source_item;
            si.source_ref = ex;
            result.extention_list.add(si.init_entity_rep(new entity_rep));
        };
        
        for each (f) in fields.all(it!=NULL && it.Context_struct==me) do {
            if f.is_relevant_for_doc() {
                result.field_list.add(f.get_rep());
            };
        };
        
        for each (m) in methods.all(it!=NULL && it.Context_struct==me) do {
            if m != NULL && m.is_relevant_for_doc() {
                result.method_list.add(m.get_rep());
            };
        };
   };
};

extend struct_member {
    is_relevant_for_doc(): bool is also {
        result = result && Context_struct.is_relevant_for_doc();
    };
};

extend field {
    get_rep(): field_rep is {
        result = init_entity_rep(new field_rep).as_a(field_rep);
        if is_a_struct() || is_a_struct_list() {
            result.property_kind = association;
        } else {
            result.property_kind = attribute;
        };
        result.name = name;
        if is_a_struct_list() {
            result.type_name = base_type.as_a(list_descriptor).base_type.name;
        } else {
            result.type_name = base_type.name;
        };
    };
};

extend method {
    
    extract_definition_comment(): string is {
        if previous_method == NULL {
            result = extract_comment();
        } else {
            result = previous_method.extract_definition_comment();
        };
    };
    
    get_extentions(): list of entity_rep is {
        if previous_method != NULL {
            result = previous_method.get_extentions();
            result.add(init_entity_rep(new entity_rep));
        };
    };
    
    get_rep(): method_rep is {
        result = init_entity_rep(new method_rep).as_a(method_rep);
        result.name = name;
        result.comment = extract_definition_comment();
        result.extention_list = get_extentions();
        for each (p) in get_params() do {
            result.parameter_list.add(p.get_rep());
        };       
    };
    
    get_params(): list of local is {
        for each (l) in scope.locals do {
            if l.name!="me" && l.name!="result" {
                result.add(l);
            };
        };
    };
};

extend local {
    get_rep(): parameter_rep is {
        result = new parameter_rep;
        result.name = name;
        result.type_name = base_type.name;
    };
};

struct doc_content_manager {
    
    struct_list: list (key: name) of struct_descriptor;
    
    module_pattern: string;

    init() is also {
        module_pattern = "/sn_.*/";
    };
    
    is_relevant_for_doc(si: source_item): bool is {
        result = si.source_module().name ~ module_pattern;
    };

    add_struct(name: string) is {
        var lt: type_descriptor = all_types.key(name);
        if lt != NULL and  lt is a struct_descriptor {
            struct_list.add(lt.as_a(struct_descriptor));
        } else {
            out(append("warning: struct ",name," does not exist"));
        };
    };
    
    remove_struct(name: string) is {
        if struct_list.key_exists(name) {
            struct_list.delete(struct_list.key_index(name));
        };
    };
    
    add_structs(names: list of string) is {
        for each (n) in names {
            add_struct(n);
        };
    };
    
    add_struct_hierarchy(name: string) is {
        var lt: type_descriptor = all_types.key(name);
        if lt != NULL and  lt is a struct_descriptor {
            struct_list.add(lt.as_a(struct_descriptor));
            for each (s) in lt.as_a(struct_descriptor).children {
                struct_list.add(s);
            };
        } else {
            out(append("warning: struct ",name," does not exist"));
        };        
    };

    remove_struct_hierarchy(name: string) is {
        for each (s) in struct_list.key(name).children {
            remove_struct(s.name);
        };
        remove_struct(name);
    };
    
};

extend global {
    dcm: doc_content_manager;
    
    init() is also {
        dcm = new;
    };
    
    print_doc(file_name: string) is {
        for each (sd) in dcm.struct_list do {
            sd.get_rep().print_doc();
        };
    };
};

'>
