	eb_gen.e - does generation-related issues for EB
	================================================

o Implements:

 - The eb_gen command

 - Converting of calls from add_field() to add_field_with_*()
   (using static analysis info)

 - The various eb_manager show_*() methods


Extending eb_node (defined in eb_elements.e)
--------------------------------------------

<'
extend eb_node {
    !p_node_role: field;
	-- The generating field (or NULL for eb_main)

    !help_lines: list of string;
    !last_group: string;
	-- See "explain" feature below

    get_role(): string is only {
	if p_node_role == NULL then {
	    return "";
	} else {
	    return p_node_role.name;
	};
    };

    -- Set p_node_parent, p_node_role for this node
    set_parent(parent: any_struct, field_name: string) is {
	p_node_parent = parent.as_a(eb_node);
	if p_node_parent == NULL {return};
	var f:= parent.type().fields.first(.name == field_name);
	if f == NULL then {
	    error("EB error in eb_gen: no field called '", field_name,
	          "'");
        };
	p_node_role = f;

	eb_manager.last_generated = me;
	   -- For the reset logic
    };

    -- An ommision in the original files
    set_group(group_name: string) is {
	add_group(group_name);
    };
    
    pre_body() is {
        
        if ( p_ebuilder.eb_manager.target_dir == "" ) then {
            var target_dir   := 
                p_ebuilder.eb_manager.p_main_node.as_a(eb_main).target_dir;
            var override_dir := 
                p_ebuilder.eb_manager.p_main_node.as_a(eb_main).override_dir;
            p_ebuilder.eb_manager.set_target_dir(target_dir,override_dir);
        };
    };
};
'>

o Implementation of eb_gen:

eb_gen is somewhat complex. What we need is two
things:

 - Normally, we want an eb_gen to generate the field if it
   is NULL.

 - When the user clicks "reset", this calls ebi_reset() of
   the parent struct, which should generate the corresonding
   field in the parent unconditionally.

Here is how it is done:

 - "eb_gen <foo>" both extends the method eb_gen_field()
   and calls it.

 - ebi_reset() calls eb_gen_field() of its parent.

<'
extend eb_node {

    -- Return the parent of this node
    parent(): eb_node is {
	return p_node_parent;
    };

    -- Reset me by re-generating me from my parent, except in the
    -- special case where we need to re-generate eb_main
    -- Also puts the last generated in last_generated
    ebi_reset() is {

	-- If this is the main, regenerate it
	if p_node_parent == NULL then {
	    --eb_manager.create_eb_main();
            eb_manager.start_ebuilder(me.type().name);
            
	-- Else call it's parent eb_gen_field()
	} else {
	    p_node_parent.eb_gen_field(p_node_role.name, FALSE);
	};
    };
	

    -- Generate this field (unless it exists and keep_old is TRUE)
    eb_gen_field(fname: string, keep_old: bool) is empty;
};

define <eb_gen'action> "eb_gen <field'name>" as {
    eb_gen_field("<field'name>", TRUE);
    fields {
	eb_gen_field(fname<?>: string, keep_old<?>: bool) is first {
	    if fname<?> == "<field'name>" then {
		if <field'name> != NULL and keep_old<?> then {
		    return;
	        } else {  -- Must generate
	    	    var t<?>:= pgen3.rnd_policy;
    		    try {
			pgen3.rnd_policy = min;
                        //var s_old_id: string = "";
                        //if (<field'name> != NULL) {
                        //    s_old_id = <field'name>.p_s_id;
                        //};
			gen <field'name>;
                        // restore id
                        //if (s_old_id != "") {
                        //    <field'name>.p_s_id = s_old_id;
                        //};
			<field'name>.set_parent(me, "<field'name>");
			pgen3.rnd_policy = t<?>;
    		    } else {
			pgen3.rnd_policy = t<?>;
			fail();
		    };
		    return;
	     	};
	    };
	};
    };	
};

define <eb_gen_list'action> "eb_gen_list <field'name>" as {
    eb_gen_field("<field'name>", TRUE);
    fields {
	eb_gen_field(fname<?>: string, keep_old<?>: bool) is first {
	    if fname<?> == "<field'name>" then {
		if not <field'name>.is_empty() and keep_old<?> then {
		    return;
	        } else {  -- Must generate
	    	    var t<?>:= pgen3.rnd_policy;
    		    try {
			pgen3.rnd_policy = min;
			gen <field'name>;
		        for each (p<?>) in <field'name> {
			    p<?>.set_parent(me, "<field'name>");
			};
			pgen3.rnd_policy = t<?>;
    		    } else {
			pgen3.rnd_policy = t<?>;
			fail();
		    };
		    return;
	     	};
	    };
	};
    };	
};
'>


o Integration issues with eb_top_gui:

<'
extend global {
    eb_manager: eb_manager;

    init() is also {
	eb_manager = p_ebuilder.eb_manager;
    };
};

'>


File writing etc.
-----------------

<'
extend eb_manager {
    !main: eb_main;

    target_dir: string;  -- Top level 
    override_dir: bool;

    set_target_dir(target_dir: string, override_dir: bool) is {
	me.target_dir = target_dir;
        me.override_dir = override_dir;
    };

};


struct eb_manager_helper {
    get_main_by_name(node_name:string): eb_main is empty;
    eb_gen_field(fname: string, keep_old: bool) is empty;
};

extend eb_manager {

    start_ebuilder(node_name:string) is {
       
        var tmp:= new eb_manager_helper;
        main = tmp.get_main_by_name(node_name);
        
        ASSERT (main != NULL);
            
        eb_manager.p_main_node = main;
        
        // get the node to select
        me.p_selected_node = p_ebuilder.p_ebuilder_gui.find_first_empty();
        if (me.p_selected_node == NULL) {
            me.p_selected_node = main;
        };
        
        ASSERT (me.p_selected_node != NULL);
        
        // open ebulder window
        p_ebuilder.get_jservice().open_ebuilder_window(
            p_ebuilder.p_ebuilder_gui.get_full_tree(), me.p_selected_node.p_s_id);
        
    };
};


define <eb_main'statement> "eb_main <node'name> <block>" as {

    struct <node'name> like eb_main {
        
        <block>;
        
        keep soft target_dir == "<node'name>";
    };
    
    extend eb_manager_helper {
        
        !<node'name>: <node'name> ;

        get_main_by_name(node_name:string): eb_main is also {
            
            if ( node_name == "<node'name>" ) then {
                specman("gen");
                eb_gen <node'name>;
                return <node'name>;
            };
        };
    };
};
    
struct eb_main like eb_node {

    target_dir : string;
    override_dir: bool;
    
    pre_create_form() is {
        
        if ( parent() == NULL ) {
            add_group("Target directory");
            add_field("target_dir", "The target directory to write into", 
                "The target directory to write into", TRUE, FALSE);
            add_field("override_dir", "Delete target direcory if exists", 
                "Delete target direcory if exists", TRUE, FALSE);
            add_group("");
        };
    };
    

    

    
    body_wrapper() is also {
	  -- After calling body()
	write_all_files();
    };

    write_all_files() is {

	-- Checking
	var s:= eb_manager.target_dir;
	if s == "" or s == "." then {
	    error("EB: bad target directory '", s, "'");
	};
	if files.file_exists(s) then {
	    if eb_manager.override_dir then {
		var new_s:= append(files.new_temp_file(), "dir");
		out("Moving old '", s, " to '", new_s, "'");
		out(output_from_check(append("/bin/mv ", s, " ", new_s)));
	    } else {
		error("EB: Cannot create directory '", s, 
		      "': it exists and 'override' was not specified");
	    };
	};

	-- Actual writing
	out("Creating directory '", s, "' and writing files ...");
	out(output_from_check(append("mkdir ", s)));
	write_files_to_dir(s);
	
	-- Create local load.ecom and target.txt files
	-- ## Note: this is the current kludge for running
        -- the second Specview from within the first
	files.write_string_list("load.ecom", {"restore";
		append("@", s, "/load.ecom")});
	files.write_string_list("target.txt", {s});

	-- Show result
	out("\nFiles created in '", s, "': ");
	out(output_from(append("ls -l ", s)));
    };

    write_files_to_dir(dir: string) is undefined;
	-- To be defined later in evm phase
	-- Does the actual writing, then shows dir in netscape
};
'>


Converting calls from add_field() to add_field_with_*()
-------------------------------------------------------

This uses pgen3.role_get_field_rl() to replace a call to
add_field() with the (perhaps narrower) call to add_field_with_*().

This is done by:

 - Finding the gen_role of this struct (created at eb_gen time).

 - Calling the pgen3.role_get_field_rl() in eb_ranges.e (new generation
   functionality), to see if there is a "better" representation 
   for the range of this field.

 - Calling the corresponding method accordingly.

## Issues:

 - Cannot call with string (seems to disallow calling with strings)

 - Issue with bools (we do not want a list)

<'
define EB_MAX_NUMERIC_COMBO 16;  -- Max numeric values to go in combo box

extend eb_node {

    -- New implementation of add_field():
    add_field(field_name: string, label: string, tool_tip: string,
        initialy_enabled: bool, notify_changes: bool) is first {
	-- Compute gen_role
	var role_str: string;
   	if p_node_parent == NULL then {
	    role_str = me.type().name;  -- ## Might have 'when' issues
	} else {
	    role_str = append(p_node_parent.type().name, ".",
				p_node_role.name);
					-- ## Might also have 'when' issues
	};

	-- Declare return params and call it
	var all_values_in_list: bool;  -- (or just min, max etc.)
	var values: list of string;  -- Actual values to be returned
        var is_string: bool;  -- Return value: is it a string

	-- If is_string, currently fall through (a gen restriction)
	-- ## Need to fix for string and bool
	var base_type:= me.type().fields.first(.name == field_name)
				.base_type.short_type;

        compute pgen3.role_get_field_rl(role_str, field_name,
		EB_MAX_NUMERIC_COMBO, all_values_in_list,
		values, is_string);

	if all_values_in_list then {
	    add_field_with_list(field_name, label, tool_tip, 
                initialy_enabled, notify_changes, values);
	    return;
	   
	} else if is_string then {
  	    ;  -- Do nothing
	
	} else {
	    add_field_with_range(field_name, label, tool_tip, 
			initialy_enabled, notify_changes, values[0], values[1]);
	   return;
	};
    };
};
'>


