
<'

extend sn_ebuilder_node_data {
    pprint() is empty;
};

extend eb_manager {
    
    //-------------------------
    // Methods
    //------------------------
    
    
    ebi_register_eb_node(node: eb_node) is {
        node.p_s_id = get_unique_id();
        p_hash_eb_nodes.add(node);
    };
    
   
    //--------------------------------
    // implement eb_manager interface
    //------------------------------
    get_unique_id(): string is {
        p_eb_node_counter += 1;
        
        result = p_eb_node_counter.as_a(string);
    };

    find_eb_node(s_id: string): eb_node is {
        result = p_hash_eb_nodes.key(s_id);
    };

    // next button was clicked in the map
    // the manager 'simulates'  next click on the form
    ask_move_to_next(user_input :sn_jform_page_input): sn_ebuilder_new_state is {
        
        ASSERT (me.p_selected_node != NULL);
        
        var s_err: string = p_selected_node.apply_user_input(user_input);
        
        if (s_err == NULL) {
            // apply was succesful - now find next node
            var node_next : eb_node = 
                p_ebuilder.p_ebuilder_gui.find_next_empty(me.p_selected_node);
            
            if (node_next != NULL) {
                // set new model and selection
                p_set_selected_node(node_next);
                result = NEW sn_ebuilder_new_state.ctor_model_sel(
                    p_ebuilder.p_ebuilder_gui.get_full_tree(),
                    node_next.p_s_id);

            } else {
                // no new selection - set new model, and mark completed
                result = NEW sn_ebuilder_new_state.ctor_model(
                    p_ebuilder.p_ebuilder_gui.get_full_tree());
                result.n_status = EB_STATUS_COMPLETED;
            };
        } else {
            result = NEW sn_ebuilder_new_state.ctor_err(s_err);
        };
        
    };

    // a different node was clicked in the map
    // the manager 'simulates'  next click on the form
    ask_move_to_node(node_next: eb_node, 
        user_input :sn_jform_page_input): sn_ebuilder_new_state is {
        
        ASSERT (me.p_selected_node != NULL);
        ASSERT (node_next != NULL);
        
        if (user_input != NULL) {
            var s_err: string = p_selected_node.apply_user_input(user_input);
        
            if (s_err == NULL) {
                // set new model and selection
                p_set_selected_node(node_next);
                result = NEW sn_ebuilder_new_state.ctor_model_sel(
                    p_ebuilder.p_ebuilder_gui.get_full_tree(),
                    node_next.p_s_id);

            } else {
                result = NEW sn_ebuilder_new_state.ctor_err(s_err);
            };
        } else {
            // no input ==> user didn't want any change
            p_set_selected_node(node_next);
            result = NEW sn_ebuilder_new_state.ctor_model_sel(NULL, node_next.p_s_id);
        };
        
    };

    // next button was clicked in the map
    // the manager 'simulates'  next click on the form
    ask_apply(user_input :sn_jform_page_input): sn_ebuilder_new_state is {
        print me.p_selected_node;
        ASSERT (me.p_selected_node != NULL);
        
        var s_err: string = p_selected_node.apply_user_input(user_input);
        
        if (s_err == NULL) {
            // set new model and selection
            var new_root := p_ebuilder.p_ebuilder_gui.get_full_tree();
            //new_root.pprint();
            result = NEW sn_ebuilder_new_state.ctor_model(new_root);


        } else {
            result = NEW sn_ebuilder_new_state.ctor_err(s_err);
        };
        
    };

    // next button was clicked in the map
    // the manager 'simulates'  next click on the form
    ask_reset(): sn_ebuilder_new_state is {
        
        ASSERT (me.p_selected_node != NULL);
        
        me.p_selected_node.full_reset();
        
        var new_root := p_ebuilder.p_ebuilder_gui.get_full_tree();
        
        //new_root.pprint();
        result = NEW sn_ebuilder_new_state.ctor_model(new_root);
        
    };

    p_set_selected_node(node: eb_node) is {
        me.p_selected_node = node;
    };
    
    get_selected_node(): eb_node is {
        return me.p_selected_node;
    };
    
    get_root() : eb_node is {
        
        result = p_main_node;
    };
    
    
    ebuilder_closed() is {
        // notification for closing the ebuilder
        p_ebuilder.is_open = FALSE;
    };
    
    ask_fill_defaults(user_input :sn_jform_page_input): sn_ebuilder_new_state is {
        
        ASSERT (me.p_selected_node != NULL);
        
        // first - apply user input, if required
        if (user_input != NULL) {
            var s_err: string = p_selected_node.apply_user_input(user_input);
        
            if (s_err != NULL) {
                result = NEW sn_ebuilder_new_state.ctor_err(s_err);
                return result;
            };
        } else {
            // no user input:
            // we check if there is any empty node - 
            // if not, we simply return 'completed'.
            if (p_ebuilder.p_ebuilder_gui.find_next_empty(me.p_main_node) == NULL) {
                
                // no empty node - return as 'completed'
                result = NEW sn_ebuilder_new_state.ctor_status(EB_STATUS_COMPLETED);
                return result;
            };
        };
        
        // now go over all nodes - we know that there is at 
        // least one empty (see above)
        var start_node : eb_node = me.p_main_node;
        var node_to_select: eb_node = me.p_selected_node;
        var curr_node : eb_node;
        var done : bool = FALSE ;
        
        while ( done == FALSE ) do {
            
            if ( start_node.p_status != EB_STATUS_FULL ) then {
                curr_node = start_node;
            } else {
                curr_node = p_ebuilder.p_ebuilder_gui.find_next_empty(start_node);
                if ( curr_node == NULL ) then {
                    done = TRUE;
                    break;
                } else {
                    // curr node is not filled yet.
                    node_to_select = curr_node;
                };
            };
            
            //out("FILL DEF ID: " , curr_node.p_s_id);
            
            var s_err: string = curr_node.validate_form();
            
            if ( s_err != NULL ) then {
                //out("FILL DEF ERR: ", s_err);
                p_set_selected_node(curr_node);

                result = NEW sn_ebuilder_new_state.ctor_model_sel_err(
                    p_ebuilder.p_ebuilder_gui.get_full_tree(), 
                    curr_node.p_s_id,
                    s_err);

                return result;
            };

            curr_node.ebi_accept_form();
            start_node = curr_node;
        };
        
        p_set_selected_node(node_to_select);
        result = NEW sn_ebuilder_new_state.ctor_model_sel(
            p_ebuilder.p_ebuilder_gui.get_full_tree(), node_to_select.p_s_id);
        result.n_status = EB_STATUS_COMPLETED;
    };
};


extend eb_node {
    
    get_title():string is {
        result = me.title;
    };
    
    get_name():string is {
        
        --result = append(me.title , " " , me.get_role()); 
	result = me.title;
    };
    
    get_desc(): string is {
        result = appendf("Node id:%s stat:%d - '%s'", p_s_id, p_status, get_name());
    };
    
};


'>
