=============== INTERFACE DEFINITION ===============

<'
// Errors
//      o when no roles exists - probably no static analysis was done
//      o when 'role_str' does not exists or has a strange format
//      o when field is not televant in role (do-not-gen or under
//              impossible when)
//      o when field is a list or non scalar
//      o when 'int_limit' is negative or > 10k

extend pgen3 {
    role_get_field_rl(
                role_str: string,
                                // e.g., "sys.s1"
                field_str: string,
                                // e.g. "F1"
                int_limit: count,
                                // if size of rl <= int_limit
                                // for numeric fields
                                // return also list of values as strings 
                                // in 'values'
                                // for enum & bool fields parameter ignored
                all_values_in_list: *bool,
                                // TRUE if field is enum or bool
                                // or number of values in range <= int_limit
                values: *list of string,
                                // if field is enum or bbool or 
                                // or number of values in range <= int_limit
                                // return strings for all values
                                // otherwise return 2strings - for min and max
                is_string: *bool
                                // TRUE if field is string
        ): any_rl is undefined;


};
'>

======= SPECMAN EXTENSIONS =============

<'
extend simple_gen_constraint {
    has_runtime_fixed_par(): bool is only {
        if one_path then {
            return FALSE;
        };
        for each (cp) in constraint_paths {
            if cp.is_runtime_fixed(me) then {
                return TRUE;
            };
        };
    };
};

extend role_manager {
    find_role(
        my_sd: struct_descriptor,
        first_sd: struct_descriptor,
        fls: list of field_holder, fromf: index, upto: index,
        start_matching_from: index, gotf_ind: index): a_role is  {

        assert fromf == 0;
        var pigp:= UNDEF;

        if gotf_ind > 0 then {
            // create (probably a dummy) role for the
            // same role fields but which is not gotf
            var gi:= pgen3.gotf_infos[gotf_ind];
            pigp =  gi.get_pos_of_fls(0, fls, upto);
            if find_role(
                my_sd, first_sd, fls, fromf, upto,
                start_matching_from, UNDEF) == NULL then {
                return NULL;
            };
        };

        if my_sd.sgi.roles.size() == 0 {
            my_sd.create_type_role();
        };

        assert start_matching_from < my_sd.sgi.roles.size();

        init_the_edge(my_sd.sgi.roles[start_matching_from],upto,fromf);
        
        var pref_fls:= e.end_point.a_role.fields;
        var start_sd:= e.end_point.a_role.start_type;
                
        rm_trace.trace_start(fls,gotf_ind,TRUE);

        find_re(fls, gotf_ind, pigp, first_sd);
        
        rm_trace.trace_e(FIND_OR_CREATE);

        var the_re := e.end_point;

        case e.er {
            exact {
                assert the_re != NULL;
                assert the_re.a_role != NULL;
                assert e.fls_pos < fromf;
                result = the_re.a_role;
            };
            default {
                result = NULL;
            };
        };  -- case
    }; -- find_or_create
};
'>

<'
extend scalar_rl {
    rl_to_string_list_typed(all: bool, t: type_descriptor): list of string
        is undefined;
};
 
define <more_scalar_rl_to_string'statement>
    "scalar_rl more to_string <type'name> <type2'name>" as {
    extend <type'name>_rl {
        rl_to_string_list_typed(all: bool, t: type_descriptor): 
                list of string is {

            if rl_list.is_empty() then {
                return result;
            };

            var di := t.d_index;

            if !all then {
                var lo := rl_list[0];
                var hi := rl_list.top();
                result.add(any_to_string(lo.unsafe(),di));
                result.add(any_to_string(hi.unsafe(),di));
            } else {

              for i from 0 to rl_list.size()-2 step 2 {
                var lo := rl_list[i];
                var hi := rl_list[i+1];

                for k from lo to hi do {
                    // due to stupid bug of converting for- local var to int
                    var kk: <type2'name> = k.as_a(<type2'name>);
                    result.add(any_to_string(kk.unsafe(),di));
                };
              };

            };

        };
    };
};

scalar_rl more to_string int int;
scalar_rl more to_string uint uint;
scalar_rl more to_string long sn_vec;

'>

=============== NEW METHODS =====================

<'
extend struct_descriptor {
    match_fields_to_role(sts: list of string, from: index, upto: index,
                fls: list of field_holder,
                start_sd: struct_descriptor,
                tds: list of struct_descriptor): a_role is {
        -- from <= index
        var s:= sts[from];

        assert upto >= 1;
        assert from >= 1;
        for each (f) in {fields; sgi.temp_fields} do {
            if f.name ~ s then { //
                var ssd: struct_descriptor;
                var is_list:= FALSE;
                case {
                    f.is_a_struct() {
                        ssd = f.struct_type();
                    };
                    f.is_a_struct_list() {
                        ssd = f.elem_struct_type();
                        is_list = TRUE;
                    };
                    default {
                        continue;
                    };
                };
                    
                if tds[from] == NULL then {
                    tds[from] = ssd; //f.struct_type();
                };
                if fls[from-1] == NULL then {
                    fls[from-1] = a new field_holder with {
                        .field = f;
                        .role = is_list ? it_exp : none;
                    };
                };
                if from == upto then {
                    var curr:= symtab.all_roles.size();
                    result = pgen3.role_manager.find_role(
                            //f.struct_type(),
                            ssd,
                                start_sd, fls, 0, upto-1, 0,  UNDEF);
                } else {
                    result = //f.struct_type().match_fields_to_role(sts,
                            ssd.match_fields_to_role(sts,
                                from+1, upto, fls, start_sd, tds);
                };
            };
        };
    };
};

extend a_role {
    fname_to_index(fs: string): index is private {
        result = UNDEF;
        for each (f) in {sd.fields; sd.sgi.temp_fields} do {
            if f.name ~ fs then {
                return index;
            };
        };
    };

    get_ffi_by_name(fs: string): final_fgr_info is {
        if final_fgr_infos.is_empty() then {
            return NULL;
        };

        var find: index = fname_to_index(fs);

        if find == UNDEF then {
            return NULL;
        };
        if find >= final_fgr_infos.size() then {
            return NULL;
        };

        return final_fgr_infos[find];
    };

    get_rl_of_ffi(ffi: final_fgr_info,
                  s: *string): any_rl is {
        if ffi.last_fgr != NULL then {
            result = ffi.last_fgr.get_rl(regular);
                //.rl_to_string(TRUE));
        } else {
            if ffi.f.fgi.default_rl == NULL then {
                ffi.f.fgi.default_rl = ffi.f.get_rl();
            };
            result = ffi.f.fgi.default_rl;
        };
        s = result.rl_to_string_typed(TRUE, ffi.f.base_type());
    };
};
'>

<'
extend pgen3 {

// given a role sting name <type>{.<field>....} return the role
    get_role(s: string, start: index): a_role is {

        if symtab.all_roles.is_empty() then {
            error("No Roles: maybe static analysis was not performed yet");
        };

        var fls: list of field_holder;

        var ss:= str_split(s, ".");
        if ss.is_empty() or ss[0] == "" then {
            error("get_role: illegal string '", s, "'");
        };

        var td:= type_called(ss[start]).as_a(struct_descriptor);
        if td == NULL then {
            error("get_role: no type called '", ss[0], "'");
        };

        var l:= ss.size()-1;
        if l == 0 then {
            return td.sgi.me_roles[0];
        };
        var tds: list of struct_descriptor;
        tds.resize(ss.size());
        fls.resize(ss.size());
        tds[0] = td;

        for i from 1 to l do {
            assert tds[i-1] != NULL;
            result = tds[i-1].match_fields_to_role(ss, i, l, fls, tds[i-1], 
tds);
            if result != NULL then {
                return result;
            };
        };
        if tds[l] != NULL then {
            result = tds[l].sgi.me_roles[0];
        };
        if result == NULL then {
            error("get_role: no role matching '", s, "' was found");
        };
    };


    role_get_field_rl(
                role_str: string,
                                // e.g., "sys.s1"
                field_str: string,
                                // e.g. "F1"
                int_limit: count,
                                // if size of rl <= int_limit
                                // for numeric fields
                                // return also list of values as strings 
                                // in 'values'
                                // for enum & bool fields parameter ignored
                all_values_in_list: *bool,
                                // TRUE if field is enum or bool
                                // or number of values in range <= int_limit
                values: *list of string,
                                // if field is enum or bbool or 
                                // or number of values in range <= int_limit
                                // return strings for all values
                                // otherwise return 2strings - for min and max
                is_string: *bool
                                // TRUE if field is string
        ): any_rl is {

        values.clear();
        all_values_in_list = FALSE;
        is_string = FALSE;

//
        assert int_limit >= 0 && int_limit < 10k;
//
        var s: string;
        var r:= get_role(role_str, 0);
        var ffi:= r.get_ffi_by_name(field_str);
        if ffi == NULL then {
            error("role_get_field_rl(): field '",
                                field_str, "' is not relevant in '",
                                role_str, "'");
        };

        var rl:= r.get_rl_of_ffi(ffi, s);

        if rl == NULL then {
            error("role_get_field_rl(): no range list for field '",
                                field_str, "'");
        };

// special handling for strings
        var t:= ffi.f.base_type;
        is_string = (t.short_type == STRING_TYPE);
        if is_string then {
            return NULL;
        };

// real stuff
        if rl is a scalar_rl (srl) then {
            if srl is a list_rl then {
                error("role_get_field_rl(): field '", field_str, "' is a list");
            };
            var sz:= srl.rl_size();
            var enum_or_bool:=
                (t.short_type == ENUM_TYPE
                        or t.short_type == BOOL_TYPE);
            var all:= enum_or_bool or (sz <= int_limit);
            values.add(srl.rl_to_string_list_typed(all, t));
            all_values_in_list = all;
            result = srl;
        } else {
            error("role_get_field_rl(): field '", field_str, "' is not scalar");
        };
    };

};
'>


