<'


extend sys {

    // The ATM that is being currently handled
    !curr_atm: atm;
    
    // Unit ATM1 - This is the main unit
    ATM1: atm is instance;
    
    // Unit ATM2 - This is secondary unit
    ATM2: red atm is instance;
    
    // The next packet waiting to be sent
    next_packet: sn_packet;
    
    post_generate() is also {
        curr_atm = ATM1;
    };
};

// ATM is the main unit in this design.
unit atm {
    // main events
    
    event clk;
    event cycle;
    event low_end;
    event green_start;

    
    // whether it is a low atm
    low: bool;
    
    // The ATM's color
    color: [red, green, blue];
    
    // the kind
    kind: [good, bad];
    

    // The cahnnel list
    channels: list of channel;
    keep channels.size() == 4;

    // The general switch to be used
    general_switch : switch is instance;

    // extend do_pack() 
    run() is also {
    };
    
    // fetch packet by name
    fetch(name: string) @sys.any is {};
    
    // read a packet by number
    read(num: int): sn_packet @sys.any  is {};
    
    // write packet to port
    write(pkt: sn_packet, port_num: int, name: string, time: time) @sys.any is {};

    // an undefined method
    foo() is undefined;
    
    
    cover clk is {
        item color;
        item kind;
    };
    
    cover cycle is {
        item low;
    };
    
    when low atm {
        // A low switch - in case we are in a low ATM
        low_size : [low, medium, big];
        
        event ev_low_atm;
        
        print_low() is {};
        
        cover low_end is {
            item low_size;
        };
        
        // extend do_pack() 
        run() is also {
        };
    };

    when green atm {
        
        event ev_green_atm;
        
        // A green switch - in case we are in a green ATM
        green_num : int (bits: 2);
        print_green(n: int) is {};
        
        cover green_start is {
            item green_num;
        };
        
    };

    when red atm {
        // A blue switch - in case we are in a blue ATM
        red_switch : switch is instance;
        keep red_switch.max_channels == 3;
        
    };

    
    when red low atm {
        // A red and low fla for this ATM
        RED_LOW: bool;
        n_red_low: uint;
        keep RED_LOW == TRUE;
        keep n_red_low < 100;
        
    };

    // current switch - using: keep curr_switch == general_switch;
    curr_switch: switch;
    keep curr_switch == general_switch;
    
    my_switch: switch is instance;
};

unit abs_switch {
    max_channels: int;
    keep max_channels < 6;
    keep max_channels > 2;
    
    lst_channels: list of channel is instance;
    keep lst_channels.size() <= max_channels;
    keep lst_channels.size() >= 2;
    keep for each in lst_channels {it.id == index};
    
    go() is undefined;
    
    supply(counter: int) is {};
    
    
};

unit special_switch like abs_switch {
    advanced: bool;
    color: [red, green, blue];

    when advanced special_switch {
      recent: bool;
    };

    when red special_switch {
       paint() is {};
    };

   when green special_switch {
       paint() is {};
    };

    when blue advanced special_switch {
       paint() is {};
    };


    go() is {};
};

unit switch like abs_switch {
    fast: bool;
    big: bool;
    static: bool;

    when fast switch {
       speed: int;
    };

    when big switch {
       size: int;
    };

    when big static switch {
       location: string;
    };

    when static switch {
       addr: int;
    };
       
    name: string;
    
    get_name() is {};
    
    go() is {};
    
};

    
unit channel {
    len: uint;
    addr: uint;
    id: int;

    atm_parent: atm;
    my_switch: switch is instance;
    switches: list of switch;
    next_cell: sn_cell;
    curr_cell: sn_cell;

    event ready;
    event done;

    cover ready is {
      item len;
      item addr;
    };

    read() is {}; 
    write() is {}; 
    find() is {}; 
};

struct sn_packet {
    x: int;
    y: int;
    cells: list of sn_cell;
    
    init() is also {
        x = 99;
        y = 888;
        cells.add(NEW sn_cell.ctor("111"));
        cells.add(NEW sn_cell.ctor("222"));
        cells.add(NEW sn_cell.ctor("333"));
    };
};

struct sn_cell {
    a: string;
    b: [red, green, blue];
    
    init() is also {
        a = "A Cell";
        b = green;
    };
    
    ctor(s: string) is {
        a = s;
    };
    
    when red sn_cell {
        purple: bool;
    };
};
        

extend global {
    init() is also {
        sys.next_packet = new sn_packet;
    };
};

display user menu separator;
display user menu "Example: show class atm" "show class atm";

'>
