		vt_ebuilder.e - the ebuilder-related VT code
		============================================

The explain() method
--------------------

o What explain() does:

This method is called upon clicking the "explain" button (or automatically,
if in auto-explain mode.

By default, this method does vt.show("vt/vt_default_explain.html").

vt_default_explain.html shows a header, and then puts all the
help_lines gathered from form (see below) in a box.

However, you can (and in fact you are encouraged) to create your
own explain pages. To do that, create (in the corresponding evm
directory) an html file (by convention we suggest calling
it <eb_node_name>_explain.html), and then show it, by modifying
your explain() method to look e.g. like the following:

	explain() is only {
	    vt.show(eb_ahb_evc_explain.html);
        };


o About the help_lines():

The user creating the form itself can use any of the following 
methods:

 add_text() - Adds text to the form

 add_group() - Creates a new group

 add_field_*() - Adds an actual control (checkbox etc.)

Each of these als adds the corresponding text (somewhat formatted) 
to a list in the eb_node called help_lines.

To get the corresponding lines, use the following method:

	get_help_lines(wildcard: string): list of string

This method will return all lines matching the wildcard format,
and remove them from the list, so subsequent get_help_lines()
will not return them.

vt_default_explain.html uses "get_help_lines(...") to get them all.

Finally, you can also use the method:

	add_help(s: string)

which simply adds s to the help_lines, whithout doing anything
to the form. This is in case you'd rather create your explain
page mainly in your .eb file and not using an html editor.


o Creation:

Each "create_form()" is enveloped in "create_form_started()"
and "create_form_ended()".

create_form_started() clears help_lines. create_form_ended()
currently does not do anything.


o Formatting of text and fields:

 - Text is by group. Currently, group header in big font

        <H1>...</H1>

 - Default group will be called "Parameters"

 - Each field has title and range list and Balloon help

        <H2>....</H2>

 - Add_text() just adds the text
        <p>
        ...
        </p>

   (maybe with <I> </I> with italics

 - add_help() is only gets added to the help.

        same

 - In the text/help, one can put HTML, and also:

 - link_url("pppp", "oo"), link_file() ...

<'
extend eb_node {

    -- Called before creating any form
    create_form_started() is also {
        help_lines.clear();
        last_group = "";
    };

    add_group(label: string) is also {
        if label != last_group then {
            add_help(|<<#:
<H3><(label)></H3>
            end #);
            last_group = label;
        };
    };

    add_field(field_name: string, label: string, tool_tip: string,
      initialy_enabled: bool, notify_changes: bool) is first {
        add_help(|<< #:
<H4>Parameter: <(label)></H4>

<(tool_tip)>
        end #);
    };

    add_text(s: string) is first {
        add_help(s);
    };

    add_help(s: string) is {
        s = str_replace(s, "/\n\n/", "\n<br>\n");
        s = str_replace(s, "/^\n/", "<br>\n");
        help_lines.add(s);
    };

    -- Return matching help line
    -- ### Note: Currently no removal
    get_help_lines(wildcard: string): list of string is {
	if wildcard == "" then {
	    wildcard = "...";
	};

	result = help_lines.all(it ~ wildcard);
    };

    explain() is {
	vt.show("vt/vt_default_explain.html");
    };
};
'>


o An eBuilder button:

This button leads to eBuilder. Note that it currently has
a special colour - should really have an icon.

<'
extend sn_vt {
    b_ebuilder(main_form: string, file_to_load: string, text: string)
								: string is {
	vt.connect = TRUE;
	var command:= (file_to_load == "" ? append("ebuilder ", main_form)
	     : append("load -if ", file_to_load, "; ebuilder ", main_form));
	return append(
 	  "<font color=\"FF6600\"><input type=\"button\" name=\"", text, 
	  "\"  value=\"", text, "\" onClick=\"specmanCommand('",
	  "display.specman_command(", 
	  vt.quote_command(quote(command)), ")')\"></font>");
    };
};
'>


o Actual writing of the files:

When the build button is clicked, eb_main.write_all_files()
creates the directory if needed, and then calls:

	write_files_to_dir(dir: string)

This method should write the files to the directory, and then 
show the post_build screen (by calling post_build(dir)).

By default, post_build() shows vt/vt_default_post_build.html, 
but you may choose to change that to create a customized HTML
file which e.g. walks the user through the generated code.

<'
extend eb_main {
    write_files_to_dir(dir: string) is {
	dir = files.get_full_file_name(dir);
	vt.write_files_to_dir(dir);
	//post_build(dir);
    };

    post_build(dir: string) is {
	vt.show("vt/vt_default_post_build.html");
    };
};

extend sn_ebuilder_service {
    build() is also {
        var root : eb_main = p_ebuilder.eb_manager.p_main_node.as_a(eb_main);
        var dir : string = files.get_full_file_name(p_ebuilder.eb_manager.target_dir);
        root.post_build(dir); // calls root post_build 
    };
};


'>


o Searching in the target dir:


o Searching:

There are four user-visible search routines:

o vt.search(files: string, search_string: string)

This method searches for each occurrance of 'search_string' in each
of the files matching 'files'.

If only one is found, then we jump to it directly (in the source
window). Otherwise, if none is found, we issue an error message.
Finally, if several are found, we create an HTML file with the
line and the file name, which is clickable.

o vt.l_search(files: string, search_string: string, text: string): string

Returns a link which when clicked will perform such a search.

o vt.target_search(), vt.l_target_search():

Same params, but search in the eBuilder target directory.

<'
extend sn_vt {
 
    search(files: string, search_string: string) is {

	var s:= append("/", str_to_partial_re(search_string), "/");
	var links: list of string;
	var command: string;

	-- Search in all files
	for each file (f) matching files do {
	    for each line (l) in file f do {
		if l ~ s then {
		    var f1:= str_replace(f, "/^.*//", "");
		    command = dec("source ", f, " ", index + 1);
		    links.add(append(
			vt.l_command(command, f1), ": ", l, "<br>"));
		};
	    };
	};

	-- See how many matches we have
	if links.size() == 0 then {
	    error("No match of '", search_string, "' found in '", files, "'");
	} else if links.size() == 1 then {
	    display.specman_command(command);
	} else {
	    -- Create list and show it 
	    connect = TRUE;
	    var title:= append("Matches of '", search_string, "' in ", files);
	    vt.show("vt/vt_search.html");
	};
    };

    l_search(files: string, search_string: string, text: string): string is {
	if text == "" then {
	    text = unquote(search_string);
	};
	return vt.l_command(append("vt.search(", quote(files), ",", 
		quote(search_string), ")"), text);
    };

    target_search(files: string, search_string: string) is {
	search(append(eb_manager.target_dir, files), search_string);
    };

    l_target_search(files: string, search_string: string, text: string)
								: string is {
        return l_search(append(eb_manager.target_dir,"/", files), 
                        search_string, text);
    };
};
'>
		
