		The top-level eVM stuff
		=======================

General description
===================

An eVM directory should always be a subdirectory of a 
directory in SPECMAN_PATH, e.g. "./foo.evm".

The eVM directory (whose name ends in ".evm") should 
contain:

  main.info - contains the info about that evm (in a standard format)

  main.html - The main HTML page

Here is how you do it:

  Create a .evm directory

  Put main.html there

  Click on "edit evm_info" to create/edit its main.info

  When done, this will update the evm_index file in the 
  parent directory and erase all evm_info


eVM-related stuff
=================


o The description of an eVM:


<'
struct sn_evm_descriptor {
    name: string;  -- Name of directory, without the ".evm"
    title: string;  -- Descriptive title;
    category: string;  -- E.g. "eVC descriptions"
    creator: string;  -- Email of creator
    change_date: string;  -- Of last change
    version: string;  -- Of this eVM
    has_ebuilder: bool;  -- TRUE if it has ebuilder components
    keywords: string;  -- Keywords separated by blanks.
   	               -- We may want to change it to a list
    under_directory: string;  -- The directory under which it resides

    -- Returns the age of the main.html file (or 0 if exists)
    -- We may consider changing this method later for performance
    -- and for unification with change_date. ###
    age(): int is {
	return files.file_age_if(append(under_directory, "/", name, ".evm"));
    };
};
	
-- This is the struct which resides in evm_index
struct sn_evm_holder {
    descriptors: list of sn_evm_descriptor;
};

extend sn_vt {
    evm_descriptors: list of sn_evm_descriptor;
	-- All those found by read_all_evm_index_files()

    -- This method makes sure we have all the eVM info in place
    -- Note: The list can be empty
    get_evms(): list of sn_evm_descriptor is {
	if evm_descriptors.is_empty() then {
	    read_all_evm_index_files();
	};
	return evm_descriptors;
    };

    -- This method makes sure we have all the eVM info in place
    -- then returns the eVM whose name this is, else error
    get_evm(name: string): sn_evm_descriptor is {
	if evm_descriptors.is_empty() then {
	    read_all_evm_index_files();
	};
	result = evm_descriptors.first(.name == name);
	if result == NULL then {
	    error("No currently known eVM called '", name, "'");
	};
    };

    -- This method finds all eVMs along the SPECMAN_PATH
    read_all_evm_index_files() is {
	
	-- Go through all dirs in SPECMAN_PATH
	evm_descriptors.clear();
	var dirs:= 
		str_split(final_path, "/:/").apply(it == "" ? "." : it);
   	var found_dirs: list of string;  -- Where we found something

	for each (dir) in dirs do {
	    var fname:= append(dir, "/evm_index");
	    if files.file_exists(fname) and not (dir in found_dirs) then {
	        found_dirs.add(dir);
		try {
		    var h: sn_evm_holder = files.read_binary_struct(fname,
			"sn_evm_holder", FALSE).as_a(sn_evm_holder);
		    assert h != NULL;
		    for each in h.descriptors do {
			.under_directory = dir;
			evm_descriptors.add(it);
		    };
		} else {
		    out("Warning: file '", fname, "' corrupted - skipping:");
		    write_the_error();
		};
	    };
	};
    };

    -- Update (or create) the index file for this directory.
    -- Also clears the global evm_descriptor cache, to force
    -- an update.
    update_evm_index_file(dir: string) is {

	var h:= new sn_evm_holder;
	
	if dir == "" then {
	    dir = ".";
	};
	for each file (fname) matching append(dir, "/*.evm") do {
	    var info_fname:= append(fname, "/main.info");

	    -- Read the struct and add to the list, or error
	    try {
		var d:= files.read_ascii_struct(info_fname,
				"sn_evm_descriptor").as_a(sn_evm_descriptor);
		assert d != NULL;
		h.descriptors.add(d);
	    } else {
		out("Warning: file '", info_fname, "' is corrupted:");
		write_the_error();
	    };
	};

	var index_fname:= append(dir, "/evm_index");
	out("Writing '", index_fname, "' (", h.descriptors.size(), " eVMs)");
	files.write_binary_struct(index_fname, h);

	evm_descriptors.clear();  -- To force an update
    };
};
'>


o Showing the eVM categories:

This is the main evm window. It shows the list of all categories
as well as some pointers to misc utilties.

<'
extend sn_vt {

    -- Show all categories matching wildcard
    evm_categories(wildcard: string) is {

	if wildcard == "" then {wildcard = "..."};

	var evms:= vt.get_evms();
	var categories:= evms.category.all(it ~ wildcard)
						.sort(it).unique(it);
	var texts: list of string;

	for each (c) in categories {
	    texts.add(dec(c, " (", evms.count(.category == c), 
		")<br>"));
	};

	var links:= categories.apply(vt.l_command(
		append("vt.evm_list(", quote(it), ")"), texts[index]));

	vt.show("vt/vt_evm_categories.html");
    };

    -- Show the full list of eVMs matching this category or categories 
    evm_list(wildcard: string) is {

	var title: string;

	if str_empty(wildcard) then {
	    wildcard = "...";
	    title = "eVMs for all categories (starting with most recent)";  
	} else {
	    title = append("eVMs for category '", wildcard, "'");
	};

	low_level_evm_list(vt.get_evms().all(.category ~ wildcard), title);
    };


    -- Low level routine to actually display the list
    low_level_evm_list(evms: list of sn_evm_descriptor, title: string) is {

	evms = evms.sort(.age()).reverse();
	
	var list: list of string;

	for each (evm) in evms do {
	    list.add(<< #:
<(vt.l_evm(evm.name))><br><br>
<(vt.evm_table(evm.name))>
<br>
	    end #);
	};

	vt.show("vt/vt_evm_list.html");
    };

    -- Show all eVMs matching this search string
    search_evms(search_string: string) is {

	var s:= append("/", str_to_partial_re(search_string), "/");
	var evms:= vt.get_evms().all(.title ~ s or .category ~ s or 
		.version ~ s or	.creator ~ s or .name ~ s or .keywords ~ s);
	low_level_evm_list(evms, append("All eVMs matching '", search_string, 
									"'"));
    };

    -- Create a link pointing at that eVM
    l_evm(name: string): string is {	
	var evm:= vt.get_evm(name);
	return vt.l_command(append("vt.show_evm(", 
		quote(name), ")"), append("<b>", evm.title, "</b>"));
    };

    -- Create a link pointing at an eVM list matching some categories
    l_evm_list(wildcard: string, text: string): string is {	
	return vt.l_command(append("vt.evm_list(", 
		quote(wildcard), ")"), text);
    };

    -- Show that particular eVM
    show_evm(name: string) is {
	var evm:= vt.get_evm(name);
	var fname:= append(evm.under_directory, "/", name, ".evm/main.html");
	var command:= append("vt.show(", fname, ")");
	display.evm_shown(evm.name, evm.title);
	display.specman_command(command);  -- Must be done as a command
    };
};


extend display {
    evm_shown(name: string, title: string) is empty;
};
'>


o The main "changes" page:




o Editing the main.info file:

This form lets you change the main.info file in an eVM directory
(or create it if needed).

You activate it via:

	vt.evm_edit(evm_name: string)

Where we add the ".evm" to the name if missing. It then edits
this file for you, and writes out the result.

If the directory does not exist, it will be created under "."
when "update" is clicked.


<'
extend sn_vt {

   -- Returns the file name of the info file for that evm
   evm_info_fname(evm_name: string): string is {
	var fname:= files.add_file_type(evm_name, ".evm", TRUE);
	return files.get_full_file_name(append(fname, "/main.info"));
   };

    -- Bring up the edit menu
    evm_edit(evm_name: string) is {

	var dir_message: string;  -- What to display about the dir
	var info_fname: string;   -- The full name of the main.info file

	-- Find or prepare to create the dir
	try {
	    info_fname = evm_info_fname(evm_name);
	    var fname:= str_replace(info_fname, "//[^/]+$/", "");
	    dir_message = append("See the directory ", 
			l_file(fname, append(fname, "/")));
	} else {
	    var fname:= append(files.curr_full_dir(), "/", evm_name, ".evm");
	    info_fname = append(fname, "/main.info");
	    dir_message = append("Directory '", fname, 
		"' does not exist - 'update' will create it");
	};

	var d: sn_evm_descriptor;

	-- If the file exists, read it
	if files.file_exists(info_fname) then {
	    d = files.read_ascii_struct(info_fname,
				"sn_evm_descriptor").as_a(sn_evm_descriptor);
	    assert d != NULL;

	-- Else create it
        } else {
	    d = new;
	    d.creator = files.get_email_address();
	    d.name = evm_name;
	    var x:= str_split(date_time(), "");
	    x.delete(0);
	    x.delete(2);
	    d.change_date = str_join(x, "_");
	    d.version = "1.0";
	};

	var categories:= {
	    "-- Select category --";
	    vt.get_evms().category.sort(it).unique(it);
	};

	vt.show("vt/vt_evm_edit.html");
    };

    -- Update to the file
    evm_update(name: string, title: string, cat_select: string, 
	category: string, creator: string, date: string, version: string,
	keywords: string) is {

	-- Deal with category selection
	if cat_select != "-- Select category --" then {
	    if category !~ "/^\s*$/" and category != cat_select then {
		error("Unclear if the category is '", category, "' or '", 
		      cat_select, "'");
	    };
	    category = cat_select;
	};

	var fname: string;  -- evm directory
	var info_fname: string;

	-- Locate or create
	try {
	    info_fname = vt.evm_info_fname(name);
	    fname = str_replace(info_fname, "//[^/]+$/", "");
	} else {
	    fname = append(files.curr_full_dir(), "/", name, ".evm");
	    out("Creating '", fname, "'");
	    out(output_from_check(append("mkdir ", fname)));
	    info_fname = append(fname, "/main.info");
	    var html_fname:= append(fname, "/main.html");
	    var evm_table_pattern:= append("<(vt.evm_info(", quote(name),
			"))>");
	    files.write_string_list(html_fname, << vt/vt_main_template.html);
	};

	var d:= new sn_evm_descriptor;
	d.title = title;
	d.category = category;
	d.creator = creator;
	d.name = name;
	d.change_date = date;
	d.version = version;
	d.keywords = keywords;
	files.write_ascii_struct(info_fname, d,
	   "eVM description", TRUE, 10, 10);
	var top_file:= str_replace(fname, "//[^/]+$/", "");
	vt.update_evm_index_file(top_file);
	vt.show_file(fname);
    };	


    -- Shows general utilities (tar, create etc.), and optionally specific
    -- stuff for the specific eVM called 'name'.
    evm_utils(name: string) is {

	-- If name is not empty, fill specific_evm
	var specific_evm: list of string;
	if name != "" then {
	    var f:= files.add_file_type(name, ".evm", TRUE);
	    f = files.get_full_file_name(f);

	    specific_evm = <<#:
<h4> Modifying the '<(name)>' eVM:</h4>

<(vt.b_command(append("vt.evm_edit(\"", name, "\")"), "Modify info"))><br>
<(vt.b_command(append("shell xemacs ", f, " &"), "Edit directory with xemacs"))>
	    end #;
	};

	vt.show("vt/vt_evm_utils.html");
    };

    -- A misc routine added just to avoid handling the title in HTML,
    -- and also because display.add_evm_bookmark() is defined later
    add_evm_bookmark(name: string) is empty;

    -- Prints out the table for a single evm
    evm_table(evm_name: string): list of string is {
	var e:= vt.get_evm(evm_name);
	var op1:= vt.l_command(append("vt.add_evm_bookmark(", quote(e.name), 
		  ")"), "Bookmark it");
	var op2:= vt.l_command(append("vt.evm_utils(", quote(e.name), ")"), 
		  "Modify it");
	var op3:= vt.l_command("vt.evm_categories(\"\")", "Top");

	result = <<#:
<table BORDER COLS=3 WIDTH="100%" BGCOLOR="#FFFF99" NOSAVE >
<tr NOSAVE>
<td NOSAVE>Category: 
<(vt.l_command(append("vt.evm_list(", quote(e.category), ")"), e.category))>
</td>
<td>Date: <(e.change_date)></td>
<td>Name: <(vt.l_file(append(e.under_directory, "/", e.name, ".evm"), e.name))>
</td>
</tr>
<tr NOSAVE>
<td NOSAVE>Creator: 
<(append("<a href=\"mailto:", e.creator, "\">", e.creator, "</a>"))></td>
<td>Version: <(e.version)></td>
<td><(append(op1, " / ", op2, " / ", op3))></td>
</tr>
</table>
	end #;
    };

    -- The following gets included in each evm main
    evm_info(evm_name: string): list of string is {
	result = {
	    evm_table(evm_name);
	    vt.c_hidden("dir", append(evm_name, ".evm"));
	};
    };
};

extend files {

    -- Return the user's email address
    -- Lifted from the bug button code
    get_email_address(): string is {

	var domain: string = "";
        try {
            var conf_file: string = "/etc/resolv.conf";
            if (files.file_exists(conf_file)) {
                for each line (L) in file conf_file {
                    if (L ~ "/^\s*domain\s+(\S+)\s*$/") {
                        domain = $1;
                        break
                    };
                };
            };
        };
        
        return append(output_from("whoami")[0], "@", domain);
    };
};
'>

        


Stripe charts
-------------

## Note: This part is here temporarily. Should be in vt_extras.e
or something.

o Description:

This defines the vt.stripe() method and the vt.box() macro.

Basically, you call vt.box() several times, building up the
specific boxes of which the stripe is composed. Then you call
vt.stripe(), which returns a list of string describing the
whole stripe.

A stripe chart is created by creating multiple stripes.


o vt.box():

 vt.box(title: string, value: string, color: string)

This defines a new box for the "current stripe", whose title string
is 'title', whose value is 'value' and whose color is 'color'.

Note that 'title' and 'value' can be of any type, and are converted
to string (or even list of string).

Note that vt.box() does not return anything. It just adds a data
structure to some hidden list, so that vt.stripe() can collect them.

If all three parameters are empty, this creates a divider.

Examples:

  vt.box("Packet kind", kind, "green");
  vt.box("", "", "");  -- Add a divider


o vt.stripe():

  vt.stripe(indent: int): list of string

This is a real method, which returns piece of HTML (represented as a list
of string) which will show a stripe containig all the boxes defined
so far. It also erases all those boxes.

'indent' specifies at what level this stripe resides (0 or more). The
whole stripe will be indented to the right accordingly.


o Basic data structures:

<'
-- A single box
struct sn_vt_box {
    is_divider: bool;
    title: string;
    value: string;
    color: string;
};

extend sn_vt {
    boxes: list of sn_vt_box; 
	-- vt.box() adds, vt.stripe() uses and clears.

    add_box(title: string, value: string, color: string) is {
	var t:= new sn_vt_box;
	t.title = title;
	t.value = value;
	t.color = translate_color(color);
	t.is_divider = (title == "" and value == "" and color == "");
	boxes.add(t);
   };

   -- Returns strings for some colors appropriate as background
   translate_color(color: string): string is {
	color = str_lower(color);
	case color {
	    "white1" 	{return "white"};  -- For compatibility
	    "pink1"  	{return "#ffcccc"};
	    "orange1"	{return "#ffcc99"};
	    "yellow1" 	{return "#ffff99"};
	    "green1"	{return "#99ff99"};
	    "blue1"	{return "#66ffff"};
	    "purple1"	{return "#ffccff"};
	    "gray1"	{return "#cccccc"};
	};
	return color;
    };
};
'>


o vt.box():

<'
define <vt_box'exp> "<exp>.box(\(<any>\))" as computed { 

    -- Initial parsing
    var all:= str_replace(str_expand_dots(<2>), "/^.(.*).$/", "\\1");
    var parts:= str_split(all, "/,[ ]*/");
    if parts.size() != 3 then {
	error("The syntax is 'vt.box(title, value, color)'\n", 
	      "Found: 'vt.box(", all, ")'");
    };
    var title:= parts[0];
    var value:= parts[1];
    var color:= parts[2];
    
    return append("vt.add_box(append(", title, "), append(", value, 
						"), ", color, ")");
};
'>


o vt.stripe():

Each stripe is a single-row HTML table, where each part (between dividers)
is a sub-table.

<'
extend sn_vt {
    stripe(indent: int): list of string is {
	assert indent >= 0 and not boxes.is_empty();

	var num_of_pieces:= boxes.count(.is_divider) + 2;


	result = <<#:
	    <table border=0 width=10 cellspacing=3 cols=<(num_of_pieces)>>
	    <tr> 
            <td width=<(indent * 40 + 1)>></td>
	end #;
	
	while not boxes.is_empty() {
	     var next:= boxes.first_index(.is_divider);
	     var l1:= (next == UNDEF ? boxes : boxes[0..next - 1]);
	     boxes = (next == UNDEF ? {} : boxes[next + 1..]);
		
	     var row1:= l1.apply(append(
			"<td nowrap bgcolor=", .color, ">", .title, "</td>"));
	     var row2:= l1.apply(append("<td nowrap>", .value, "</td>"));
	     result.add(<<#:
		<td>
		<table cols=<(l1.size())> border=1 cellspacing=0>
		<tr align=center><(row1)></tr>
		<tr align=center bgcolor=white><(row2)></tr>
	        </table>
		</td>
	     end #);
	};

	result.add(<<#:
	    </tr></table>
	end #);
    };
};
'>


o vt.create_and_show():

Creates an HTML file and shows it. 
'title' is the title of the file
'contents' is the contents of the file, as a list of HTML strings.

<'
extend sn_vt {
    create_and_show(title: string, contents: list of string) is {
	var lines:= <<#:
<html>
<head>
<title><(title)></title>
</head>
<body>
<center><h3><(title)></h3></center>
<(contents)>
</body>
</html>
	end #;
	vt.show_expanded(lines);
    };
};
'>


o vt.l_struct():

This method returns an HTML link to the struct (will issue the "tree" 
command on that struct).

'title' is the hyperlink title - if empty, use e.g. "@12".


<'
extend sn_vt {
    l_struct(the_struct: any_struct, title: string): string is {
	var str:= the_struct.to_string();
	var s:= str_replace(str, "/^.+ (.*)$/", "\\1");  -- No initial blanks
	var s1:= str_replace(s, "/^.+(@.+)$/", "\\1");  -- The default title
	title = (title == "" ? s : title);
	return vt.l_command(append("print ", s, " using window, title=NULL"), 
								title);
    };
};
'>


	
o vt.field_boxes():


 vt.field_boxes(the_struct: any_struct, fields_wildcard: string,
	        physical_only: bool, color: string): list of string;

This method creates a contiguous list of boxes, from left to right, 
one for each matching field in 'the_struct'.

Matching means that the full field name matches 'fields_wildcard',
and that the field is physical (if physical_only is TRUE).

Note that fields under when are not displayed if that when is not
open, even if their names match.

There is a maximal number of chars in each box (default: 20), after
which we add "...".

## Initially: Physical only, scalar only, nos support for chopping.

<'
extend sn_vt {
    field_boxes(the_struct: any_struct, fields_wildcard: string,
			        physical_only: bool, color: string) is {

	-- Preliminary stuff
	if the_struct == NULL then {
	    return;
	};
	if not physical_only then {
	    error("vt.field_boxes() with non-physical fields is ",
		  "not implemented yet");
	};

	-- Compute it	
	var t:= the_struct.type();
	var names: list of string;
	var bits: list of string;
	var values: list of string;
	var next_bit:= 0;

	for each (f) in t.fields {
	    if (physical_only and not f.is_physical) or not 
		field_exists(the_struct, f) or
		f.name !~ fields_wildcard then {
		continue;
	    };

	    -- Compute value
	    var value: untyped = 
			sn_get_field_by_no(the_struct, f.field_no).unsafe();
	    var value_s:= any_to_string(value, f.base_type.d_index);

	    -- Compute bit
	    var high:= next_bit + f.base_type.size_in_bits - 1;
	    var bit:= (high == next_bit ? dec(high) 
					: dec(high, ":", next_bit));	    
	    next_bit = high + 1;

	    -- Compute name
	    var name:= str_replace(f.name, "/^.*'/", "");

	    -- Add to lists
	    values.add(value_s);
	    names.add(name);
	    bits.add(bit);
	};

	-- Finally, do the boxes
	for each (v) in reverse values do {
	    vt.add_box(append(names[index], " [" , bits[index], "]"), v, color);
	};
    };
};
'>
