	eb_expand.e - implements the text expansion constructs
	======================================================

Defining sn_vt
--------------

o Define sn_vt:

Definee the sn_vt struct (of which there is one instance, global.et)
to contain all necessary methods.

<'
-- The struct representing a file buffer (before string replacement)
struct sn_vt_buffer {
    name: string;  -- The file name

    lines: list of string;  -- After expansion

    module;  -- Where it was first defined
    line_no: int;  -- Where it was first defined

    prev: sn_vt_buffer;  -- Points at the previous one or NULL
};

extend global {
    vt: sn_vt;
    init() is also {
	vt = new;
    };
};

-- The main struct of this feature 
struct sn_vt {
    buffers: list (key: name) of sn_vt_buffer;
	-- The file buffers

    list_buffers: list of sn_vt_buffer;
	-- The quoted list buffers - handled as a stack

    current_buffer: sn_vt_buffer;
	-- The current buffer we are adding to, or NULL


    et_dir: string;  
	-- The directory to write the files to, e.g. "foo_1.etd"

    trace: [off, on];  -- Whether we do tracing
};
'>


File-creation macros
--------------------

o The file-creation actions:

Here are the macros defining those actions:

  file <fname> <<{line;...}, file <<{line;...}

Note: Converting the "file <fname> <<#: <lines> end #;" construct into 
the corresponding "file <<{line; ...}" construct is done automagically 
by the Specman pre-processor.


o What file should do:

For example:

	file test.e <<#:
	    var x: int;
	    var y: <(t1)>;
	end #;

Should be translated into:

    {    
	new_file("test.e");
	vt.line(  "       var x: int;");
	vt.expand({"       var y: "; "<(t1)>"; ";"}, {dec(t1)});
    };


The "file <<#" construct should be translated similarly, with an
empty file name. 

New_file creates a new buffer (or takes over an old one by that
name and makes it empty).
Line() and expand() add to the current buffer.


Finally, there is also the:

   file <file> << <orig-file>

syntax, which uses the "<< <orig-file>" syntax (later)


o Implementation:

Note: I am currently hijacking the "C code" action, to be used for that
purpose. Thus, it currently cannot be used inside et itself.

## Still needs to implement "file <file> << <from'file>" via the exp.


<'
define <file_from_file'action> "file <file> \<\<[ ]<second'file>" as {
    var t<?>:= << <second'file>;
    vt.new_file((<< {<file>})[0]);
    vt.current_buffer.lines.add(t<?>);
};

define <file'action> "file <file> \<\< <block>" as {
    vt.new_file((<< {<file>})[0]);
    C code <block>;
};

define <file_empty'action> "file \<\< <block>" as {
    vt.new_file("");
    C code <block>;
};


extend <c_code'action> {
    do_parse() is also {

	var strings:= <strings>.x.string;  -- The original lines
	var lines:= vt.create_expand_code(strings);
	var final:= append("{", str_join(lines, "; "), "}");
	reparse(me, final);
    };
};
'>

o Creating the code for expanding the lines:

The main action here is to replace each "<(...)>" with
the corresponding code.

Two issues deserve special attention:

 - HTML editors often replace "<" and ">" with "&lt;" and "&gt;",
   we need to replace back.

 - HTML editors often decide for themselves where to split lines.
   Hence, a "<(...)>" may end up split between two lines.

Here is how we treat this second issue:

 - We first split_all() on the "<(...)>" pattern.

 - If the very last part in the split (i.e. beyond the last)
   matches "...<(...", then we append it as is to the next line (adding
   a single " " before the next line) and remove it from the current line.

Note that both of these "solutions" may be wrong for non-HTML files.
We probably want to change this routine to handle HTML and non-HTML
files differently, but this is non-trivial and is delayed for later ###.

<'
extend sn_vt {

    create_expand_code(strings: list of string): list of string is {

	-- Convert the strings to a list of action lines	
	var lines: list of string;  -- Will contain the calls to et.*()
        var reminder: string;  -- From previous line

	for each (sline) in strings do {

	    var s:= (str_empty(reminder) ? sline 
				: append(reminder, " ", sline));
	    s = str_replace(s, "/&lt;/", "<");
	    s = str_replace(s, "/&gt;/", ">");
			
  	    -- See if there is a need to expand
	    var parts:= str_split_all(s, "/<\\(.`\\)>/");

	    -- See if we need to create a new reminder 
	    if parts.top() ~ "/<\\(/" then {
		reminder = parts.pop();
		if str_len(reminder) > 1k then {
		    error("Matching ')>' missing for too long: '", 
			reminder, "'");
		};
	    } else {
	        reminder = NULL;
	    };
	    if parts.is_empty() then {
		continue;
	    };

	    -- If no embedded expressions, add vt.line() call
	    if parts.size() == 1 then {  
	        lines.add(append("vt.line(", quote(s), ")"));

	    -- Else add one vt.expand() call
	    } else {

		var values: list of string;

		-- Create the pieces
		for each (part) in parts do {

		    -- Get the original string
		    if even(index) then {
			continue;  -- No need to expand;
		    };

		    assert part ~ "/^<\\((.*)\\)>$/";
		    values.add(append("dec(", $1, ")"));
		};

		-- Creatre the vt.expand() action
		var s1:= str_join(parts.apply(quote(it)), "; ");
		var s2:= str_join(values, "; ");
		lines.add(append("vt.expand({", s1, "}, {", s2, "})"));
	    };
	};	    
	
	return lines;
    };
};
'>



File creation methods
---------------------

<'
extend sn_vt {

    -- Clear all
    clear_files() is {
	-- Clear everything
	buffers.clear();
	list_buffers.clear();
#ifdef USING_ET {
	current_buffer = NULL;
	set_config(cover, mode, normal);
	set_check("... did not match test.patterns ...", ERROR_AUTOMATIC);
		-- To make sure we write coverage
	out("");
};
    };


#ifdef USING_ET {
    -- Write those buffers out to the proper directory
    write_files() is {

        -- An initial check
	if buffers.is_empty() then {
	    error("ET: No files defined - cannot do an et 'test'");
	};

	-- Compute the et_dir, create it
	et_dir = dec(symtab.modules.top().name, "_", get_config(gen, seed), 
		     ".etd");
	if files.file_exists(et_dir) then {
	    spawn("/bin/rm -r ", et_dir);
	};
	spawn("mkdir ", et_dir);

	-- Write all the files there
	for each (b) in buffers do {

	    -- If this is a .e file, auto-add "<''>" as a convenience		
	    if b.name ~ "/\.e$/" then {
		b.lines = {"<'"; b.lines; "'>"};
	
	    -- For shell script, auto-add "unset SPECMAN_DISPLAY", to allow 
	    -- running under Specview (otherwise a nested Specman goes crazy 
	    -- thinking it is in display mode)
	    } else if b.name ~ "/\.sh$/" then {
		b.lines = 
		     {"#!/bin/csh -f"; "unsetenv SPECMAN_DISPLAY"; b.lines};
	    };

	    files.write_string_list(append(et_dir, "/", b.name), b.lines);
	};
    };
};


    -- Write the files to the specified directory
    -- Assumes directory has been created previously
    write_files_to_dir(dir: string) is {

        -- An initial check
	if buffers.is_empty() then {
	    error("EB: No files defined - cannot write them");
	};

	et_dir = dir;

	-- Write all the files there
	for each (b) in buffers do {
	    files.write_string_list(append(et_dir, "/", b.name), b.lines);
	};
    };


    -- Returns TRUE if this is a new file, and further sets current_buffer
    -- to the new file. If this is really a list_buffer (noted by a "--"
    -- fname), then simply add to the list_buffers stack.
    new_file(fname: string) is {

	-- If a list_buffer file, add to list_buffers stack
	if fname == "--" then {
	    var prev:= current_buffer;
	    current_buffer = new;
	    current_buffer.name = fname;
	    current_buffer.prev = prev;
	    list_buffers.add(current_buffer);
	    if trace == on then {
		out("Expanding << #: ...\n");
	    };
	    
	-- Else if the fname is empty, use current_buffer
	} else if str_empty(fname) then {
	    if (current_buffer == NULL and buffers.is_empty()) or
	          current_buffer.name == "--" then {
		error("ET: 'file' specified with no <filename>, but ",
		      "no previous file exists to append to");
	    };
	    
	-- Else find if this buffer already exists
	} else if file_exists(fname) then {
	    var prev:= current_buffer;
	    current_buffer = buffers.key(fname);
	    current_buffer.prev = prev;
	    current_buffer.lines.clear();
	    if trace == on then {
	        out("Replacing ", current_buffer.name, " ...\n");
	    };

	-- Else a new file
	} else {
	    var prev:= current_buffer;
	    current_buffer = new;
	    current_buffer.name = fname;
	    current_buffer.prev = prev;
	    buffers.add(current_buffer);
	    if trace == on then {
	        out("Expanding ", current_buffer.name, " ...\n");
	    };

	};
    };


    -- Add a line to the current buffer
    line(s: string) is {
	current_buffer.lines.add(s);
    };

   
    -- Expand the parts using the values, and add to the current buffer.
    -- The values correspond to the odd parts.
    expand(parts: list of string, values: list of string) is {

	-- Initial stuff
	if trace == on then {
	    out("   expand: '", str_join(parts, ""), "'");
	};

	-- Replace odd parts with replacment
	for each (s1) in values do {
	
	    var s:= s1;  -- Because cannot assign
	    var parts_index:= 2 * index + 1;

#ifdef USING_ET {
	    -- If this looks like a struct or list of struct, get 
            -- it's .text strings it is an et_goal
	    if s ~ "/^(.* )?[a-zA-Z0-9_]+-@\d+$/" then {

		-- Create a list of all the indices (numbers after "@")
		var splits:= str_split_all(s, "/[a-zA-Z0-9_]+-@\d+/");
		splits = splits.all(odd(index));
		var indices:= splits.apply(
			str_replace(it, "/^.*@/", "").as_a(int));
		var lines: list of string;
		for each (i) in indices do {
		    var struct: base_struct = get_printed_struct(i);
		    assert struct != NULL;
		    var st1:= struct.as_a(et_goal);
		    if st1 == NULL then {
			error("'", splits[i], "' is not an et_goal, hence ",
			     "cannot get its '.text' field");
		    };
		    var l1:= st1.text;
		    if not l1.is_empty() then {
			lines.add(l1);
		    };
		};			
		s = str_join(lines, "\n");
	    };
};
	    -- If has \n in it, need special multi-line expand
	    if s ~ "/\n/" then {
		var pad_size:= parts[0..parts_index - 1].sum(str_len(
				symtab.untabify(it)));
		if pad_size != 0 then {
		    var rep:= append("\n", str_repeat(" ", pad_size));
		    s = str_replace(s, "/\n/", rep);
		};
	    };
	    parts[parts_index] = s;  -- Finally do the replacement
	};

	-- Add it
	var new_l:= str_join(parts, "");
	if trace == on then {
	    out("     into: '", new_l, "'\n");
	};
	current_buffer.lines.add(new_l);
    };


    -- Return TRUE if this file is already defined via "file foo <<#"
    -- Note: use this to prevent the default (override) behaviour
    file_exists(fname: string): bool is {
	return buffers.key_exists(fname);
    };


    -- ## This routine is to be replaced by display.get_printed_struct
    -- when it becomes mainstream
    get_printed_struct(id: int): base_struct is {
	var index:= display.printed_struct_index.key_index(id);
	if index == UNDEF then {
	    error("@", dec(id), " not a legal name");
	};

	return display.printed_structs[index];
    };
};
'>


Misc. macros
------------


o The << #: macro:

This is an expression, not an action. It is used e.g. as follows:

     var x: list of string = << #:
	var x: <(t1)>;
	var y: int;
     end #;

<'
define <et_quote'exp> "\<\< <block>" as {
    first {
	var s<?>: list of string;
	vt.new_file("--");
	C code <block>;
	s<?> = vt.get_lines();
    } then (s<?>);
};

-- A special workaround for "~" problem
define <list_eq_workaround'exp> "<exp> == \<\< <any>" as {
    <exp> == (<< <any>)
};

extend sn_vt {

    -- Return the stuff associated with 
    get_lines(): list of string is {
	var b:= list_buffers.pop();
	assert b.name == "--";
	current_buffer = b.prev;
	return b.lines;
    };
};
'>


o The apply_all() macro:

This is needed for:

  list1.apply_list(<list-exp>)

Typically, for:

 <(hdl_vars.apply_list(.init_lines()))>

<'
define <apply_list'exp> before <type'exp> 
				"<list'exp>.apply_list\(<exp>\)" as delayed {
    first {
        var l<?>:= (<list'exp>)!parse_check(list, *, 
					"The 'apply_list()' object"); 
        var v<?>: list of typeof(first { var it: typeof(l<?>[0]); var index: int;} then ((<exp>)[0]));
        for each in l<?> do {
                v<?>.add(<exp>);
            };
    } then (v<?>)
};
'>


o eb trace [on|off]

<'
define <eb_trace'command> "eb tr[ace][ (on|off)]" as {
    vt.trace = <3|on>;
};
'>
 


The <<file constructs
---------------------

This defines two constructs:

  - The expression '<< file'

  - The action 'file <file1> << file2'.

The action is implemented using the expression.

The expression works much like the "<<#:" expression,
except that the file is not inline.

How is that done?

One simple solution is to include it all inline.

<'
define <eb_list_from_file'exp> before <et_quote'exp> "\<\<[ ]<file>" as 
								computed {
    var fname:= <2>;
    var strings:= each line in file fname;
    var lines:= vt.create_expand_code(strings);
    var s:= append(
	    "first {var s<?>: list of string; vt.new_file(\"--\"); {",
	    str_join(lines, "; "), 
	    "}; s<?> = vt.get_lines();} then (s<?>)");
   
    return s;
};
'>



The various show_*() methods
----------------------------

These use display.open_web_location(string, web_location_kind).


o show_file(filename: string):

Shows this file (as either ASCII or HTML, depending on its extension)
in Netscape. Name is relative to current directory. To show
something in a component directory, use e.g.:

    show_file("$SPECMAN_EB_TOP/router/router_overview.txt");

<'
extend sn_vt {
    show_file(filename: string) is {
	
	-- Use add_file_type() to work around problems with original
	-- method not knowing about ~foo etc.
	filename = files.add_file_type(filename, "", TRUE);
	filename = files.get_full_file_name(filename);
	show_url(append("file://localhost", filename));
    };

    l_file(filename: string, text: string): string is {
	if text == "" then {
	    text = unquote(filename);
	};
	try {
	    filename = files.add_file_type(filename, "", TRUE);
	    filename = files.get_full_file_name(filename);
            if text == "" then {
	       text = files.file_pure_name(filename);
            };
	    result = append("<A HREF=\"file://localhost", filename, "\">", 
				text, "</A>");
	} else {
	    result = append("*** File not found: '", filename, "' (for link '",
			text, "') ***");
	};
    };

    -- Remove quotes (HTML cannot take them)
    unquote(s: string): string is {
	return str_replace(s, "/[\"']/", "");
    };
};


extend sn_vt {
    -- Returns the full name of local file (i.e. one residing where 'me'
    -- was defined).
    local_file(file: string): string is {
	return  append(files.file_dir(me.type().source_module().file_name), 
							file);
    };
};
'>


o show_url(url: string):

Shows the specified URL in a netscape window.

<'
// patch, in order to overcome the differences between
// various views
extend sn_vt {
    show_url(the_url: string) is {
	display.open_web_location1(the_url);
    };

    l_url(url: string, text: string): string is {
	return append("<A HREF=\"", url, "\">", text, "</A>");
    };

};
'>


o show_vadvisor(chapter_file: string):

Shows that chapter in Vadvisor in a netscape window. Example:

   show_vadvisor("check_scoreboard/check_scoreboard.html");

<'
extend sn_vt {
    show_vadvisor(chapter: string) is {
	show_file(append("$SPECMAN_HOME/docs/vadvisor/", chapter));
    };

    l_vadvisor(chapter: string, text: string): string is {
	return vt.l_file(append("$SPECMAN_HOME/docs/vadvisor/", chapter),
						text);
    };
};
'>


o show_manual(chapter_file: string):

Shows that chapter in the eref/sref etc. in a netscape window. Example:

   show_manual("control_flow_actions12.html");

To find out the name of the chapter_file, when in Vadvisor/manual,
click on right button and choose "show frame in another window",
then see what file it is showing.

<'
extend sn_vt {
    show_manual(chapter: string) is {
	show_file(append("$SPECMAN_HOME/docs/index_html/", chapter));
    };

    l_manual(chapter: string, text: string): string is {
	return vt.l_file(append("$SPECMAN_HOME/docs/index_html/", chapter),
						text);
    };
};
'>


o The "|<<" syntax:

This syntax is similar to the list quote syntax, but:

 - It returns a string with \n

 - Removes all initial blanks

Useful as in:

	add_text(|<< #:
....
....
	end #;

<'
define <et_list_as_string'exp> "\|[ ]\<\<[ ]<block>" as {
    str_join((<< <block>).apply(str_replace(it, "/^[ \t]+/", "")), "\n")
};
'>


The vt_* constructs
---------------------

<'
extend sn_vt {

    connect: bool;  -- If TRUE, add Specman connection code

    -- Show an expanded version of this list
    show_expanded(lines: list of string) is {

	-- If we connect, then we add the connect applet, and also
	-- put it all in one big form. Note that sometimes we did
	-- turn connect on directly, but we still need it (see match below).
	if not connect and lines.has(it ~ "/specmanCommand/") then {
	    connect = TRUE;
	};
	if connect then {
	    var n_port: int = display.start_applet_connection(UNDEF);
	    var s_codebase:= display.get_applet_class_dir();
            if (n_port < 0) {
                sn_error("Cannot start applet connection server");
            };
--	    compute system(append(
--		           "cp applet/SpecviewConnectionApplet.class ",
--			   files.get_temp_dir()));
	    for each (l) in lines {
	        if l ~ "/(</HEAD>|</head>)/" then {  -- ##Fix
		    var in_string:= display.get_applet_in_header();
		    var post_string:= 
				display.get_applet_post_header(n_port, 
								s_codebase);

		    lines[index] = str_replace(l, "/(</HEAD>|</head>)/",
		        append(in_string, "</HEAD>", post_string));
	        };

	        if l ~ "/(<body[^>]*>|<BODY[^>]*>)/" then {
		    lines[index] = str_replace(l,
			 "/(<body[^>]*>|<BODY[^>]*>)/", "\\1<form>");
		};

	        if l ~ "/(</body>|</BODY>)/" then {
		    lines[index] = str_replace(l,
			 "/(</body>|</BODY>)/", "</form>\\1");
		};

	    };
	};

	var x:= append(files.new_temp_file(), "html");
	files.write_string_list(x, lines);
	vt.show_file(x);
    };
};
'>



o <vt_report'exp>:

example:

   vt.report(my_list, .x, .y)

This returns a list of string which defines the corresponding
html table. It mimics the <report'action> logic.

<'
define <vt_report'exp> "<exp>.report(\(<any>\))" as computed { 

    -- Initial parsing
    var all:= str_replace(str_expand_dots(<2>), "/^.(.*).$/", "\\1");
    var parts:= str_split(all, "/,[ ]*/");
    var list:= parts.pop0();
    var titles: list of string;
    var titles_string:= str_replace(parts.pop0(), "{...}", "\\1");
    if titles_string != "" then {
	titles_string = str_replace(titles_string, "/[\"]/", "");
        titles = str_split(titles_string, "/; */");
    } else {
        titles = parts.apply(str_replace(str_add_dots(it), "/^\./", ""));
    };
    var t1:= append("<tr BGCOLOR=\"#99ff99\"><td>", str_join(titles, 
		"</td><td>"), "</td></tr>");
    var t2:= append("\"<tr><td>\", ", str_join(parts, ", \"</td><td>\", "),
		    ", \"</td></tr>\"");
      -- Should become:  "<td>", .name, "</td><td>", .index, "</td>"
    -- Start creating the code
    var lines:= {
	dec("first {var l<?>:= {\"<table BORDER COLS=", parts.size()+1, 
	    " WIDTH=10>\"}");
    };
    lines.add(append("l<?>.add(", quote(t1), ")"));
    lines.add(append("l<?>.add(", list, ".apply(append(", t2, ")))"));
    lines.add("l<?>.add(\"</tr></table>\")} then (l<?>)");

    return str_join(lines, ";");
};
'>


Expand issues
-------------

o vt.show(file, [exp])

Shows file as expanded with "me" assigned to the required struct
expression.

If exp appears, then it serves as "me" during the expansion.
If exp is not used, no assignment is made to "me".

The file can appear with or without quotes.

<'
define <vt_show'action> "<exp>.show(\(<any>\))" as computed { 

    -- Initial parsing
    var all:= str_replace(str_expand_dots(<2>), "/^.(.*).$/", "\\1");
    var parts:= str_split(all, "/,[ ]*/");
    if parts.size() not in [1..2] then {
	error("Bad syntax for vt.show(): '", all, "'");
    };

    var fname:= str_replace(parts[0], "/\"(.*)\"/", "\\1");
    //print 1, fname;
    fname = files.add_file_type(fname, ".html", TRUE);
    //print 2, fname;
    fname = str_replace(fname, "/[\\/][\\/]/", "/");
    //print 3, fname;
    if parts.size() == 1 then {
	return append(
	 "{vt.connect = FALSE; vt.show_expanded(<<", fname, ")}");

    } else {
	return append("{var t<?>:= ", parts[1], "; var me:= t<?>;",
              "vt.connect = FALSE; vt.show_expanded(<<", fname, ")}");
    }
};
'>

o vt.embed(file, [exp])

Expands the file (with optional exp, as in vt.show()), but rather
than showing it, returns it as a list of string, usually to be
embedded within another file.

The reutrned list includes only the stuff inside the body
(excluding all before and including <body>, and all after and
including </body>).

The "exp" is supplied as "me" (no change to me if no exp).

The file has to appear without quotes.

<'
define <vt_embed'exp> "<exp>.embed(\(<any>\))" as computed { 

    -- Initial parsing
    var all:= str_replace(str_expand_dots(<2>), "/^.(.*).$/", "\\1");
    var parts:= str_split(all, "/,[ ]*/");
    if parts.size() not in [1..2] then {
	error("Bad syntax for vt.expand(): '", all, "'");
    };

    var fname:= str_replace(parts[0], "/\"(.*)\"/", "\\1");
    fname = files.add_file_type(fname, ".html", TRUE);
    if parts.size() == 1 then {
	return append("vt.embed_list(<<", fname, ")");

    } else {
	return append("first {var t<?>:= ", parts[1], 
		"; var me:= t<?>; var r<?>:= vt.embed_list(<<", fname, 
		")} then (r<?>)");
    }
};

extend sn_vt {
   
    -- Return just the proper lines from the list (inside the body)
    embed_list(lines: list of string): list of string is {
        var inside: bool;

	for each (l) in lines {
	    if inside {
	        if l ~ "^(.*)</(body|BODY)>/" then {
		    result.add($1);
		    inside = FALSE;
		} else {
		    result.add(l);
		};
	    } else {
	        if l ~ "/<(body|BODY)[^>]*>(.*)$/" then {
	            inside = TRUE; 
		    result.add($2);
	        };
	    };
        };
    };
};
'>


o vt.modules(wildcard):

Returns the hyperlinked list of files:

<'
extend sn_vt {
    modules(wild: string): list of string is {
	return symtab.modules.all(print_options.in_good_module(it)).
	        all(.name ~ wild).apply(vt.l_file(.file_name, ""));
    };
};
'>

o image:

<'
extend sn_vt {
    image(file: string): string is {
	file = files.get_full_file_name(files.add_file_type(file, "", TRUE));
	return append("<img src=\"file://localhost", file, "\">");
    };
};
'>


o vt.command_output()

   vt.command_output(command: string, lines: int, chars: int): list of 
								string;

Creates a scrolling list with the output from that command.

E.g.

  <(vt.command_output("show break", 3, 20))>

<'
extend sn_vt {
    command_output(command: string, lines: int, chars: 
						int): list of string is {
	var x:= append(files.new_temp_file(), "elog");
	specman("set log ", x, " {", command, "}");
	var text:= (each line in file x)[2..];
	return {
	    dec("<form><textarea readonly rows=", lines, " cols=", chars, 
								">");
	    text;
	    "</textarea></form><p>"
	};
    };
};
'>


Connecting to Specman
---------------------

This method causes connecting Specman. It should appear
somewhere at the top of the file.

<'
extend sn_vt {
    b_command(command: string, text: string): string is {
	if text == "" then {
	    text = unquote(command);
	};
	vt.connect = TRUE;
	return append("<INPUT TYPE=\"button\" NAME=\"", text, 
	  "\"  VALUE=\"", text, "\" onClick=\"specmanCommand('", 
	  vt.quote_command(command), "')\">");
    };

    l_command(command: string, text: string): string is {
	if text == "" then {
	    text = unquote(command);
	};
	vt.connect = TRUE;
	return append("<a href=\"javascript:specmanCommand('", 
		vt.quote_command(command), "')\">", text, "</a>");
    };

    -- Cause a vt.show(file, struct) when clicked
    l_show(file: string, struct: string, text: string): string is {
	return vt.l_command(append("vt.show(", file, ", ", struct, ")"), text);
    };
};
'>


o Handling controls:

Controls are created via e.g.:
 
  vt.c_text(name: string, initial: string)

  vt.c_checkbox(name: string, initial: bool)

  vt.c_group(name: string, value: string, initial: bool)

  vt.c_select(name: string, values: list of string, initial: string)

  vt.c_hidden(name: string, value: string)


<'
extend sn_vt {
    c_text(name: string, initial: string): string is {
	return append("<input name=\"", name, "\" type=\"text\" value=",
		  quote(initial), ">");
    };

    c_checkbox(name: string, initial: bool): string is {
	return append("<input name=\"", name, "\" type=\"checkbox\"",
	          " value=\"sn_checkbox\"",
		  (initial ? " checked" : ""), ">");
    };

    c_group(name: string, value: string, initial: bool): string is {
	return append("<input name=\"", name, "\" type=\"checkbox\"",
	          " value=\"", value, "\"",
		  (initial ? " checked" : ""), ">");
    };

    c_select(name: string, values: list of string, initial: string)
							: list of string is {


	-- Compute the index of initial
	if values.is_empty() then {
	    error("Empty list for 'c_select()'");
	};
	var ind:= (str_empty(initial) ? 0 : values.index(initial));
        if ind == UNDEF then {
	    error("Initial value '", initial, "' for c_select() ",
		  "is not in option list:", values);
	};
	
	return {
	    append("<select name=\"", name, "\" value=\"", initial, "\">");
	    values.apply(append("<option ", (index == ind ? "selected " : ""),
			 "value=", quote(it), ">", it));
	    "</select>"
	};
    };

    c_hidden(name: string, value: string): string is {
	return append("<input name=\"", name, "\" type=\"hidden\" value=\"", 
		value, "\">");
    }
};
'>


o The vt.quote_command() method:

It should take e.g. "sys.show(%foo, \"%foo\", %bar)"
And return:         "sys.show('+cget(form.foo)+...

<'
extend sn_vt {
    quote_command(s: string): string is {
        var parts:= str_split_all(s, "/%(\w+)/");
        for each in parts do {
            if even(index) then {
                parts[index] = display.quote_for_javascript(it);
            } else {
                parts[index] = str_replace(it, "/%(\w+)/", 
                                           "'+cget(form.\\1)+'");
            };
        };
        return str_join(parts, "");
    };
};
'>




Misc utilities
--------------

o Load on demand:

<'
define <load_if_missing'command> "load<opt>if[_missing] <any>" as {
    load_if <any>
};
'>


o Executing a command:

By default, we just pass it to display.specman_command().
This (unlike calling specman()), makes sure that we do not 
have problems if this is a GUI command (which opens some window),
or it is a restore / load command etc..

If the size of the string is > 256, there is a GUI problem (TCL
cannot receive a command this long). In such a case, we create
a command file with the command, and execute it.

<'
extend display {
    html_execute(s: string) is only {
	if str_len(s) > 200 then { -- Being on the safe side
	    var f:= append(files.new_temp_file(), "ecom");
	    files.write_string_list(f, {s});
	    s = append("@", f);
	};
	display.specman_command(s);
    }
};
'>


Plotting charts
---------------

This is done using the public domain "Plod" package.


o Syntax:


  vt.chart(type, list, labels, values, ...);


o Examples:

  vt.chart("line", all_types[0..40], {}, .name, .size_in_bytes, 
							.fields.size());

o Note:

 "line" and "bar" work. "pie" does not, yet.

 Also, just 5 by 50 max items (auto-chop beyond 50).

<'
define <vt_chart'exp> "<exp>.chart(\(<any>\))" as computed { 

    -- Initial parsing
    var all:= str_replace(str_expand_dots(<2>), "/^.(.*).$/", "\\1");
    var parts:= str_split(all, "/,[ ]*/");
    if not (parts.size() in [4..8]) then {
	error("vt.plot() must have 4 to 8 parameters");
    };
    var type:= parts.pop0();
    if type == "" {
	type = "line";
    };
    var list:= parts.pop0();
    var titles: list of string;
    var titles_string:= str_replace(parts.pop0(), "{...}", "\\1");
    var lables:= parts.pop0();
    if titles_string != "" then {
	titles_string = str_replace(titles_string, "/[\"]/", "");
        titles = str_split(titles_string, "/; */");
    } else {
        titles = parts.apply(str_replace(str_add_dots(it), "/^\./", ""));
    };
    var colors:= {"blue";"red";"black";"green";"cyan"};
      -- Possibles: red, green, blue, yellow, purple, cyan, white, gray, black

    -- Start creating the code
    var f:= append(get_symbol("SPECMAN_GUI_EXTRAS"), "/jb/e/vt");
    var t1:= append(
      "append(\"<applet code=Plod.class codebase=\\\"file:", f, "\\\"",
      " width=\", vt.chart_width(t<?>.size()), \" height=250>\")");

    var t2:= append("<param name=bgcolor value=lightyellow>",
	      "<param name=plot value=", type, ">");
    -- Should be append("<param value=", str_join(t<?>.apply(..),","), ">");
    var t3:= append("append(\"<param name=label value=\", quote(str_join(", 
	      "t<?>.apply(str_chop(append(", lables, 
	      "), 11)), \",\")), \">\")");

    -- Initial part
    var lines:= {
	append("first {var t<?>:= ", list, 
	   "; if t<?>.size() > 50 {t<?> = t<?>[0..49]};",
	   "var l<?>:= {", t1, "}");
	append("l<?>.add(", quote(t2), ")");
	append("l<?>.add(", t3, ")");
    };


    -- Add all "y" lines
    for each (p) in parts do {
	var post:= (index == 0 ? "" : dec(index));
	var t:= append(
		"append(\"<param name=color", post, " value=", colors[index],
		 "> <param name=legend", post, " value=", titles[index],
		 "> <param name=y", post, " value=\", quote(str_join(", 
		 "t<?>.apply(dec(", p, ")), \",\")), \">\")");
	lines.add(append("l<?>.add(", t, ")"));
    };
	
    -- Finally
    lines.add("l<?>.add(\"</applet>\")} then (l<?>)");

    return str_join(lines, ";");
};

extend sn_vt {
    chart_width(items: int): string is {
	items = min(items, 50);
	var x:= items * 75;
	if x < 550 {x = 550};
	result = dec(x);
    };
};
'>

