package com.verisity.jspecview.comm;

import java.io.*;
import java.net.*;

import com.verisity.utils.snorb.*;
import com.verisity.specview.model.gencode.*;
import com.verisity.utils.debug.*;
import com.verisity.utils.debugmode.*;
import com.verisity.icover.*;
import com.verisity.specview.model.*;
import com.verisity.specman.model.SpecmanCommandModel;
import com.verisity.jspecview.*;
import com.verisity.jspecview.model.gencode.*;


public class AppletConnection  implements Runnable,IDebugMode{


  static private AppletConnection __connection;
  private ServerSocket             _server  = null;
  private PrintWriter              _out = null;
  private BufferedReader           _in = null;
  private Socket                   _clientSocket= null;

  private int              _nPort ;
  private String           _sHost = null;

  private boolean          _clientIsConnected = false ;
  private boolean          _threadIsAlive     = false;

  // an applet_command sewrvice instance
  private sni_japplet_command  _appletCommand = null;  

  static final private String OUT_MESSAGE_PREFIX = "J!!";
  static final private String IN_MESSAGE_PREFIX = "APPLET!!";
  static final private String OUT_MESSAGE_SUFFIX = "@@@";
  static final private int IN_MESSAGE_PREFIX_LEN = IN_MESSAGE_PREFIX.length();
  static final private int OUT_MESSAGE_SUFFIX_LEN = OUT_MESSAGE_SUFFIX.length();
  static final private int INITIAL_PORT = 10001;

  static final String COMMON_REPLY               = "ok";


  private AppletConnection () throws IOException {
    if (JAppletService.DEBUG) System.out.println("Creating new Applet Connection");
      openAppletConnection(INITIAL_PORT);
  }



  static public AppletConnection getConnection() {
    if (__connection == null) {

      try {
	__connection = new AppletConnection();
      } catch (Exception e) {
	__connection = null;
      }
    }

    return  __connection ;
  }

  // getting the command service - lazy init.
  private sni_japplet_command getAppletCommandService(){
	if  (_appletCommand == null){
	  _appletCommand=
		(sni_japplet_command)Snorb.lookupService(sni_japplet_command.DESCRIPTOR, JAppletService.COMMAND_SERVICE_NAME);   
	  if(DEBUG) Assert.assert( _appletCommand != null); 
	}
	return _appletCommand ; 
  }

  
 

  private void openAppletConnection(int nInitialPort) throws IOException {
    
    if (JAppletService.DEBUG) System.out.println("Opening Applet Connection");
    int nCurrPort = nInitialPort;
    boolean wasOpened = false;

    int counter = 0;
    while(!wasOpened ) {

      if(counter > 1000)
        throw new IOException();

      counter ++;
      if (JAppletService.DEBUG) System.out.println("nPort =  " + nCurrPort);

      wasOpened = openServerSocket(nCurrPort);
      if(!wasOpened) {

	if (JAppletService.DEBUG) System.out.println("Can't Open Connection on port: " 
						     + nCurrPort + " host:" + getHostName());
        nCurrPort++;

      } else {
	if (JAppletService.DEBUG) System.out.println("Connection  Opened on port: " 
						     + nCurrPort + " host:" + getHostName());
      }

    }

    _nPort = nCurrPort;
    _sHost = internalGetHost();
    
    // Start listening
    new Thread(this).start();

  }

  public  boolean openServerSocket(int nPort){
   try {
      _server = new ServerSocket(nPort);
      return true;
    } catch(Exception e) {
      return false;
    }
  }

  public sn_japplet_data getData() {
    
    return new sn_japplet_data(
			       getPort(),
			       getHostName(),
			       getHostAddress());
  }

  public String getHostName() {
    try {
      return InetAddress.getLocalHost().getHostName();
    } catch (Exception e) {
      return null;
    }

  }

  public String getHostAddress() {
    try {
      return InetAddress.getLocalHost().getHostAddress();
    } catch (Exception e) {
      return null;
    }
  }

  public int getPort() {
    return _nPort;
  }

  private String internalGetHost() throws IOException {
    return InetAddress.getLocalHost().getHostAddress();
  }



 // handel single client session:
  private void handleAClient (){	
    
    while(_clientIsConnected){
      
      try{

	String sMessage = readMessage();
	String reply = handleSpecviewMessage(sMessage);
	writeMessage(reply);
	
      }catch(IOException e ){
		_clientIsConnected = closeConnectionToClient();
      }
      
    }
  }	
  

  //listen to incomming clients - until the closeConnectin is called.
 public void run() {

   
   _threadIsAlive = true;
   
   while(_threadIsAlive){ 
	 
	 //wait for connection from client 
	 try {
	   
	   if (JAppletService.DEBUG) System.out.println("Server is waiting to client on port: " + 
													_nPort + " host:" + getHostName());
	   _clientSocket = _server.accept();
	   if (JAppletService.DEBUG) System.out.println("A new client was connected");
	   
	   _out = new PrintWriter(_clientSocket.getOutputStream(), true);
	   _in = new BufferedReader(new InputStreamReader(_clientSocket.getInputStream()));
	   
	   _clientIsConnected = true;
	   
	 } catch(IOException e) {
	   System.out.println("IOException: "+e.getMessage());
	 }  
	 
	 handleAClient();	  
   }
   
 }
  



  /* public void run() {
    Socket clientSocket = null;
    
    //wait for connection from client 
    try {
      
      if (JAppletService.DEBUG) System.out.println("Server is waiting to client on port: " + _nPort + " host:" + getHostName());
      clientSocket = _server.accept();
      if (JAppletService.DEBUG) System.out.println("A new client was connected");

      
    } catch(IOException e) {
      System.out.println(e.getMessage());
    }
    
    try {
      _out = new PrintWriter(clientSocket.getOutputStream(), true);
      _in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
    } catch(IOException e) {
      System.out.println("IOException: "+e.getMessage());
    }    
    
    _threadIsAlive  = true;
    //communication with client, most of the comm. stops with "end"
    try{	   
      while( _threadIsAlive ){
	
	System.out.println("In server before read...");
	String sMessage = readMessage();
	
	System.out.println("server read "+sMessage );
	String reply = handleSpecviewMessage(sMessage);
	writeMessage(reply);
	
	
	System.out.println("In server after read...");
	//}
      }
    }catch(IOException e) {
      System.out.println(e.getMessage());
    }
  }*/







  private String readMessage() throws IOException {

    String sMessage = null;

    try {

      if (JAppletService.DEBUG) Assert.assert(_in != null);

      String sFullLine = _in.readLine();

      if (JAppletService.DEBUG) System.out.println("message from Applet was read : " + sFullLine );

      // if we got a null from client 
      //it indicates that the client has closed the connection.
      if (sFullLine == null)
		throw new EOFException();
	  
      if (DEBUG) Assert.assert(sFullLine.startsWith(IN_MESSAGE_PREFIX));
	  
      sMessage = sFullLine.substring(IN_MESSAGE_PREFIX_LEN);
	  
	  if(JAppletService.DEBUG)System.out.println("READING from socket :" + sMessage);
	  
      } catch (EOFException e) {

		// in higher level - the exception will cause a client connection closing.
        if (DEBUG) System.err.println("EOFException: readMessage(): Specview Socket");
		throw new IOException();
		
      } catch (IOException e) {
		
        if (DEBUG) System.err.println("Couldn't read data from Specview on port: " +
									  _nPort + " host: " + _sHost);
        throw new IOException(e.getMessage());
      }
    
	if(sMessage != null)
	  sMessage = sMessage.substring(0,sMessage.length()-OUT_MESSAGE_SUFFIX_LEN);
    return sMessage;
  }
  
  private void writeMessage(String sMessage) {
    
    String sFullLine = OUT_MESSAGE_PREFIX + sMessage + OUT_MESSAGE_SUFFIX + "\n";
    _out.write(sFullLine);
    _out.flush();
    
    if(JAppletService.DEBUG)System.out.println("WRITING into socket :" + sMessage);
    
  }
  
  public String handleSpecviewMessage(String sMessage){
    
    //SpecmanCommandModel.runSpecmanCommand(sMessage);
	try{
	  getAppletCommandService().command_from_japplet(sMessage);
	}catch(OrbException e){
	  if(DEBUG)Assert.print("snorb err in _appletCommand.command_from_japple");
	}
    return  COMMON_REPLY;
  }
  


 private boolean  closeConnection(){
 
	try{
	  _server.close();
	  
	}catch(IOException e){
	}
	return false;
  }
  


  private boolean closeConnectionToClient(){
	
	// _clientIsConnected = false;
    
    try{
      if(_clientSocket != null)
	_clientSocket.close();	
      if(_in != null)
	_in.close();
      if(_out != null)
	_out.close();
    }catch(IOException e){
      System.out.println("failed closing ClientConnection" +_clientSocket );
    }
	return false;
  }
  
  
}


