#------------------------------------------------------------------------
# The global data structure that holds the info is 
#            "the_global_config_categories" !!!!
#
# This is a list of information for each category :
#   Category name
#   "visited" flag - to ensure that data of only visited pages will be applied
#   list of options when every option is :
#         {
#          tcl value name 
#          e_var_name 
#          init_value 
#          validation_type(UNUMERIC,NUMERIC,QUATED - used only in entry widget)
#          label
#          tool tip
#          min_range 
#          max_range 
#          regulat expression 
#          list of values / for enums or radio buttons              
#          group title / title of the group of the option - for validation msg
#          }
#  
#
#  Communication with specman
#
#    Sn -> Tcl/TK
#
#      After pressing the "config" button in specview Specman start to perform
#      the following actions:
#
#    if the configuration window is already opened just raise its required page:
#    
#       new_sn_raise_config_window (window , page)
#     
#
#    ELSE (should build the dialog) :
# 
#     new_sn_init_config_window (window, location, required_category, title)   
#     
#     For each category
#        new_sn_build_config_page( window ,  category_name , category_label)
#
#     new_sn_finish_build_pages ( window)
#
#     for each page/category
#         new_sn_start_build_page( window , required_page)
#         for each group in category 
#            new_sn_build_group(window, category ,group_name, 
#                               options_num, option_list)
#
#    remark : option list is a list of all options belong to group.
#    each option is :
#      ui_type , e_field_name , initial value , is_numeric , 
#      tool_tip , min range , max range , 
#      regular expression , enum values
#
#
#     new_sn_finish_build_config( window , required_page)
#
#
#   Whenever the set_config or config command is performed ( in the prompt)
#   Spacman calls Tk :
#
#   new_sn_refresh_values(window , category , var_name , new_val)
#
#------------------------------------------------------------------------


#------------------------------------------------
#------   CONFIG  WINDOW
#-------------------------------------------------

#-----------------------------------------------------------------
#-- Creates the full config window, with the different pages.
#-- Each page represent a different confguration category.
#-----------------------------------------------------------------

set max_int_string MAX_INT
set min_int_string MIN_INT
set max_uint_string MAX_UINT
set undef_string UNDEF

set config_wname   ""
set legal_input    1
set curr_page ""

set group_border_width  2
set items_vertical_specing 5
set external_group_hspecing 5
set external_group_vspecing 10
set group_item_hspecing 10
set dialog_item_hspecing 5
set entry_default_width 25


set category_name_idx 0
set category_visited_idx 1
set category_options_list_idx 2

set option_tcl_var_idx 0

# same idx for the TCL option parameters and the order of params passed from SN
set option_e_var_name_idx 1
set option_orig_val_idx 2
set option_type_idx 3
set option_label_idx 4
set option_tooltip_idx 5
set option_min_idx 6
set option_max_idx 7
set option_re_idx 8
set option_values_idx 9
set option_group_idx 10

# building_dialog is 1 if new dialog is buing opened.
# in case the dialog was already opened - should be 0
set building_dialog 1

#----------------------------------------------------------------------------
# The TCl command that is called from specman when the "config" menu option is
# activated and the config dialog should be opened
#----------------------------------------------------------------------------
proc new_sn_init_config_window { wname location page title max_int min_int undef e_to_tcl_newline open_on_last_page e_m_val e_k_val} {

    global config_book_name
    global config_wname
    global building_dialog
    global max_int_val
    global min_int_val
    global max_uint_val
    global undef_val
    global newline
    global last_page
    global curr_page
    global m_val
    global k_val

    #counter that enables to create unique name for widgets that are not
    # assosiated with tcl_vars
    global unique_name_counter 

    set unique_name_counter 0
    set max_int_val $max_int
    set min_int_val $min_int
    set undef_val $undef
    set newline $e_to_tcl_newline
    set last_page $open_on_last_page
    set max_uint_val 4294967295  
    set m_val $e_m_val
    set k_val $e_k_val
  
    if { $page == $last_page} {
	set page $curr_page
    }
    if { $page == "" } {
	set page "print"
	set curr_page $page
    }

    # if there's already a config window open - just raise it
    if  { [winfo exists $wname] != 0 } {
	raise $wname
	new_sn_raise_page $config_book_name $page
	set building_dialog 0
	return
    }

  
    set building_dialog 1

    #-- create the window
    toplevel $wname
    set config_wname $wname
    create_window_icon $wname

    
    #-- set some initial window values
    wm title $wname $title
    wm geometry $wname $location
    wm minsize $wname 300 300
    wm protocol $wname WM_DELETE_WINDOW "sn_config_close $wname 1"

    #-- catch 'destroy'
    catch "destroy $wname.config_book"

    sn_build_balloon $wname

    
    #-- build book
    set config_book_name $wname.config_book
    tixNoteBook $config_book_name

    return
}


#----------------------------------------------------------------------------
# The TCl command that is called from specman when the "config" menu option is
# activated and the config dialog is opened
#----------------------------------------------------------------------------
proc new_sn_raise_config_window {wname page } {
 
    global config_book_name
    global building_dialog

    # if there's already a config window open - just raise it
    if  { [winfo exists $wname] != 0 } {
	raise $wname
	new_sn_raise_page $config_book_name $page
	set building_dialog 0
	return
    } else {
	config_error_message "Internal error - dialog status"
    }
    
}

proc new_sn_raise_page { config_book_name page } {

    $config_book_name raise $page
    new_sn_update_raised_page_params $page

    
}


proc new_sn_update_raised_page_params {page } {

    global category_visited_idx
    global the_global_config_categories
    global curr_page

    set curr_page $page

    # update is_visited flag
    set config_option [get_category_info_by_name $page]

    set idx $category_visited_idx
    set config_option [lreplace $config_option $idx $idx 1] 

    set category_idx [get_category_global_idx $page] 
    set the_global_config_categories [lreplace $the_global_config_categories $category_idx $category_idx $config_option]

}

#------------------------------------------------------------------------------
# Add page to the tab control ( the gUI effect) AND
# add a new category_option to the the_global_config_categories list (the data)
#------------------------------------------------------------------------------
proc new_sn_build_config_page { wname category_name page_label} {

    global building_dialog
    global config_book_name
    global the_global_config_categories

    if { $building_dialog == 0 } {
	return
    }
   
    $config_book_name add $category_name -label "$page_label" -raisecmd "on_page_change"

    lappend category_info $category_name 0 
    lappend the_global_config_categories $category_info

    return
}

#-------------------------------------------------------------------------
# This proc is called from specman after all categories were added to the
# tab control but the pages are still empty .... ( bofore adding the widgets)
#-------------------------------------------------------------------------
proc new_sn_finish_build_pages { wname } {
    
    global building_dialog
    global config_book_name
    global the_global_config_categories
    global GenericGroupIdx

    if { $building_dialog == 0 } {
	return
    }

    #-- pack book
    pack $config_book_name -expand 1 -fill both -side top

    #-- Build the global buttons
    new_sn_build_config_buttons $wname

    set GenericGroupIdx 1
   
    return
}


#-------------------------------------------------------------------------
# This proc is called from specman for each page just before staring to add
# its options 
#-------------------------------------------------------------------------
proc new_sn_start_build_page_options { wname category } {

    global building_dialog

    if { $building_dialog == 0 } {
	return
    }

    global PageName
    global ColFrameName
    global ColIdx
    global group_border_width

    set ColIdx 0
    set PageName [$wname.config_book subwidget $category ]

    create_new_col

    return
}

#-------------------------------------------------------------------------
# This proc is called from specman for each group of the current category
# ( The category which is currently built).
# This proc adds a group to the curr page and adds all its options .
#-------------------------------------------------------------------------
proc new_sn_build_group { wname category group options_num options} {

    global building_dialog

    if { $building_dialog == 0 } {
	return
    }

    global GenericGroupIdx
    global ColFrameName
    global group_border_width
    global external_group_hspecing
    global external_group_vspecing
    global dialog_item_hspecing
    global group_item_hspecing
    global col_options_num
    
    if { $group == "" } {  # options without group

	set generic_group_name genericGroup$GenericGroupIdx
	set group_frame_name $ColFrameName.$generic_group_name 
	set group_title  ""
	
	set v_external_specing 0
	set xpading $dialog_item_hspecing

	frame  $group_frame_name -border $group_border_width -relief flat

	incr GenericGroupIdx  	

    } else {

	set group_title $group

	set tokens [split $group_title]
	set group_name [join $tokens _ ]
	set group_frame_name $ColFrameName.a$group_name
	# add 'a' to group name to ensure lower case

	set v_external_specing $external_group_vspecing
	set xpading  $group_item_hspecing 

	frame  $group_frame_name -border $group_border_width -relief groove
    }

    pack $group_frame_name -fill x -anchor w -side top -padx $external_group_hspecing -pady $v_external_specing
    
    if { $group != ""} {
	set group_label_name $group_frame_name.label
	label $group_label_name -text $group_title -relief ridge
	pack $group_label_name -anchor w -side top
	
	
    }

    set category_idx [get_category_global_idx $category]

    for { set i 0} {$i < $options_num} {incr i} {
	set option [lindex $options $i] 
	new_sn_build_option $category $category_idx $option $group_frame_name $xpading $group_title

    }
    
    # handle dialog window size. make sure its height is less than 0.5
    # of screen size 
    #set screen_h [winfo screenheight $wname]
    #set dialog_h [winfo reqheight $wname]
    #puts $dialog_h
    #puts [winfo geometry $wname]
    #if { [expr $dialog_h * 2 ] > $screen_h } {
    #	set ColFrameName $PageName.col1

    #	frame $ColFrameName -border $group_border_width
    #	pack $ColFrameName -anchor w -expand 1 -side left
    #}

    set col_options_num [ expr $col_options_num + $options_num ]
    if { $col_options_num >= 6 } {
	create_new_col	
    }
    return
}


#-------------------------------------------------------------------------
# option is a string that was send by Specman and describes option :
#     widget_type
#     Specman var_name
#     config_all value ( the init value )
#     label & tool tip
#     min & max ( for range checking)
#     regular expression ( for string matching)
#     values ( used for enum) 
#---------------------------------------------------------------------
proc new_sn_build_option { category category_idx option parent padx grp_title} {

    global the_global_config_categories
    global category_name_idx 0
    global category_visited_idx 1
    global category_options_list_idx
    global option_e_var_name_idx 
    global option_orig_val_idx
    global option_type_idx
    global option_label_idx 
    global option_tooltip_idx 
    global option_min_idx 
    global option_max_idx 
    global option_re_idx 
    global option_values_idx
    global option_group_idx
    global undef_val
    global max_uint_val
    global undef_string
    global max_uint_string
    global entry_default_width
    global entry_min_width
    global newline

    # seed entry is a special case due to its "Random" buttom and
    # it is handled not in the sutomatic mechanizem
    set BuildSpecialForSeed 0

    set widget_type [lindex $option 0]
    
    set var_name [lindex $option $option_e_var_name_idx]

    set width $entry_default_width  

    set validation_type [lindex $option $option_type_idx]
    
    set init_val [lindex $option $option_orig_val_idx]

    set is_numeric 0
    if { [string compare $validation_type "NUMERIC"] == 0 } {
	set is_numeric 1
    }
    set is_u_numeric 0
    if { [string compare $validation_type "UNUMERIC"] == 0 } {
	set is_u_numeric 1
    }

    if { $init_val == $undef_val } {	
	if { $is_numeric == 1 } {
	    set init_val $undef_string
	}
    }
    
    if { $init_val == $max_uint_val } {	
	if { $is_u_numeric == 1 } {
	    set init_val $max_uint_string
	}
    }
    
    set init_val [special_values_to_string $init_val $validation_type]
      
    set label [lindex $option $option_label_idx]
    set tool_tip [lindex $option $option_tooltip_idx]
    if { $tool_tip == "" } {
	set tool_tip $label
    }

    regsub -all $newline $tool_tip \n tool_tip
    regsub -all "\;" $tool_tip "\n" tool_tip
    
    set min_range [lindex $option $option_min_idx]
    set max_range [lindex $option $option_max_idx]
    set re [lindex $option $option_re_idx]
    set values [lindex $option $option_values_idx]

    if { $var_name != "dummyField" } {
    	#append tool_tip "\n(Role: " $var_name 

	if { $is_numeric == 1 } {
	    set tt_min_range [special_values_to_string $min_range "NUMERIC"]
	    set tt_max_range [special_values_to_string $max_range "NUMERIC"]
	}
	if { $is_u_numeric == 1 } {
	    set tt_min_range [special_values_to_string $min_range "UNUMERIC"]
	    set tt_max_range [special_values_to_string $max_range "UNUMERIC"]
	}
    
	if { $is_numeric == 1 || $is_u_numeric == 1 } {
	    append tool_tip "\n (Range: " $tt_min_range ".." $tt_max_range ")"
	} 

	# append tool_tip ")" 
    }

    global tclVal$category$var_name
    set tcl_var_name tclVal$category$var_name
    set $tcl_var_name $init_val

    if { $var_name == "seed"} { 
	set BuildSpecialForSeed 1
    }
    
    # create the list of parameters of the new option
    lappend new_option_info  $tcl_var_name $var_name $init_val $validation_type $label $tool_tip $min_range $max_range $re $values $grp_title

    # append the new_option parameters to list of previous options
    # first - get the category  option in the_global_config_categories
    # than - get the options info of the category 
    # and append to existing options the new option
    set category_info [lindex $the_global_config_categories $category_idx]  
    set options_info [lindex $category_info $category_options_list_idx]
    lappend options_info $new_option_info
    
    # create updated category_info entry in the format of category_info
    # in the_global_config_categories in order to replace the updated info 
    # with the existing one
    set category_name [lindex $category_info $category_name_idx]
    set visited  [lindex $category_info $category_visited_idx]
    lappend new_category_info $category_name $visited $options_info
    
    #replace the list of options for a category with the new list
    set the_global_config_categories [lreplace $the_global_config_categories $category_idx $category_idx  $new_category_info]
   
    if { $BuildSpecialForSeed == 1 }  {
	sn_build_config_entry $parent $label $var_name $tcl_var_name $tool_tip\
	    $padx 14 1
    } else {
	switch -exact $widget_type {
	    label {
		sn_build_config_static_text $parent $label
	    }
	    button {
		sn_build_config_button $parent $re $label $tool_tip \
		    $validation_type
		# PATCH for buttons - the re holds the name of the button
		# This name is send to specman and idntify the button
	    }
	    entry {	   
		sn_build_config_entry $parent $label $var_name $tcl_var_name \
		    $values $tool_tip $padx $entry_default_width
	    }
	    checkbox {
		sn_build_config_checkbox $parent $label $var_name $tcl_var_name\
		    $values $tool_tip $padx
	    }
	    radiobutton {
		sn_build_config_radiobutton $parent $label $var_name \
		    $tcl_var_name $values $tool_tip
	    }
	    dropdown {
		sn_build_config_dropdown $parent $label $var_name $tcl_var_name\
		    $values $tool_tip $padx 
	    }
	    scale {
	    }
	    default {
	    }
	}
    }

    return
}

proc new_sn_finish_build_config { wname page } {

    global config_book_name

    global curr_page
    global the_global_config_categories
    global last_page

    if { $page == $last_page } {
	set page $curr_page
    }
    if { $page == "" } {
	set page "print"
	set curr_page $page
    }

    #-- Move to wanted page
    new_sn_raise_page $config_book_name $page
    
    #-- Make it modal
    focus $wname
    
    return
}


#-------------------------------------------------
#------   HANDLE BUTTONS
#-------------------------------------------------
proc sn_config_close {wname from_button} {

    global the_global_config_categories

    focus .main0.input_field

    set command_result [sn_command "display.close(\"$wname\", \
	\"[wm geometry $wname]\", 1)" specman 0 $from_button $from_button] 

    unset the_global_config_categories

    return
}

proc new_sn_config_apply {wname } {

    
    new_collect_config_values $wname FALSE
    
    return
}



#-----------------------------------------------------------------------------
# Before appling 2 validation checks must be performed :
# page checking for the curr page to ensure that values of this page that are
# dependant in each other - are OK.
# tab checking - to ensure that items in different pages that their values are
# dependant - are OK
#-----------------------------------------------------------------------------
proc new_collect_config_values {wname close_dlg} {

    global specman_tmp_dir
    global the_global_config_categories
    global curr_page
    global legal_input
    global category_visited_idx
    global is_config_modified
    global file

    set result 0
    set legal_input 1
    set is_config_modified 0

    set status [perform_validation]
    if { $status == 0 } {
	# can not apply - some field is invalid    
	return $status
    }


    set status [validate_page $curr_page ] 
    if { $status == 0 } {
	# can not apply - the current page is invalid    
	return $status
    }

    set status [validate_tab]
    if { $status == 0 } {
	# can not apply - the current page is invalid    
	return $status
    }
 
    # open ecom file with all config commands in the temp dir
    set config_ecom_file_name config_tmp.ecom
    set file_name [concat $specman_tmp_dir$config_ecom_file_name]

    for {set i 0} { $i < [ llength $the_global_config_categories] } {incr i} {
	set category_info [lindex $the_global_config_categories $i]
	set is_visited [lindex $category_info $category_visited_idx]
	if { $is_visited == 1 } {
	    collect_category_info $category_info $file_name
	}
    }
    
    if { $is_config_modified == 1 } {
	
	# remove the file in the end (self destruction)
	puts $file "session.sn_config.already_validate = FALSE -- SILENT"
	puts $file "files.file_delete(\"$file_name\") -- SILENT"
	close $file
	
	set sn_config_ecom_file_name "$file_name"
	
	set result [sn_command "session.apply_config_values(\"$sn_config_ecom_file_name\",  $close_dlg , \"$wname\")" specman 0 1]
	
    } else {
	#if no changes were done specman isn't activated & the close window
	# is made by TCL.
	#(if there are changes - specman controls the update and close win)
	if { $close_dlg == "TRUE" } {
	    sn_config_close $wname 1
	    set result 1
	}
    }

    return $result
}

#---------------------------------------------------
# OK action is like apply and than close 
#--------------------------------------------------
proc new_sn_config_ok {wname } {

    new_collect_config_values $wname TRUE

    return
}


#-----------------------------------------------------------------------------
# Whenever the set_config or config command is performed in specview prompt
# the config window must be updated
# The 'new_values_list' is list of {category var_name new_val}
#-----------------------------------------------------------------------------
proc new_sn_refresh_values {new_values_list} {

    set refreshed_options_num [llength $new_values_list]


    for { set i 0 } {$i < $refreshed_options_num } {incr i} {

	set refreshed_option [lindex $new_values_list $i]
	set category [lindex $refreshed_option 0]
	set var_name [lindex $refreshed_option 1]
	set new_val [lindex $refreshed_option 2]
	set validation_type [lindex $refreshed_option 3]
	# replace special values to their string
	set new_val [special_values_to_string $new_val $validation_type]

	global tclVal$category$var_name 
	set tcl_var_name tclVal$category$var_name
	set $tcl_var_name $new_val

	# update orig_value in option info - to be the new value.
	global option_orig_val_idx
	global category_options_list_idx
	global the_global_config_categories

	set category_info [get_category_info_by_name $category]
	set options_list [lindex $category_info $category_options_list_idx]
	set option_info [get_option_info_by_category_name $category $var_name]

	#set option_idx [lsearch $options_list $option_info]
	set option_idx [get_option_idx_by_var_name $options_list $var_name]
	if { $option_idx == -1 } {
	    config_error_message "Internal error - invalid refresh"
	}
	#set category_idx [lsearch $the_global_config_categories $category_info]
	set category_idx [get_category_global_idx $category]
	if { $category_idx == -1 } {
	    config_error_message "Internal error - invalid refresh"
	}
	


	# replace orig val in option_info list
	set option_info [lreplace $option_info $option_orig_val_idx $option_orig_val_idx  $new_val]
	#replace option_info in options list
	set options_list [lreplace $options_list $option_idx $option_idx $option_info]
	#replace options_list in category info
	set category_info [lreplace $category_info $category_options_list_idx $category_options_list_idx $options_list]
	# replace category in config options global list
	set the_global_config_categories [lreplace $the_global_config_categories $category_idx $category_idx $category_info]

    }
}


#--------------------------------------------
# Private methods
#---------------------------------------------
proc special_string_to_value { val } {
    
    global m_val
    global k_val
    
    # regular exp for <num>M or <num>K
    set m_val_regexp  "^( |\t)*(-)?(0b\[01]\[_01]*|0o\[0-7]\[_0-7]*|0\[hx]\[0-9a-f]\[_0-9a-f]*|\[0-9]\[_0-9]*\[mM])( |\t)*\$"
    set k_val_regexp  "^( |\t)*(-)?(0b\[01]\[_01]*|0o\[0-7]\[_0-7]*|0\[hx]\[0-9a-f]\[_0-9a-f]*|\[0-9]\[_0-9]*\[kK])( |\t)*\$"
    
    if { [regexp $m_val_regexp $val] > 0 } {
	regsub "m" $val " " val 
	regsub "M" $val " " val
	set val [expr $val*$m_val]
    }

    if { [regexp $k_val_regexp $val] > 0  } {
	regsub {[kK]} $val " " val
	set val [expr $val*$k_val]
    }
    
    return $val
}


proc special_values_to_string {value validation_type} {
    
    global max_int_val
    global min_int_val
    global max_uint_val
    global undef_val
    global max_int_string
    global min_int_string
    global max_uint_string
    global undef_string
    global newline
    

    set ret_val $value

    if { $value == $max_int_val} {
	set ret_val $max_int_string
    }
    if {$value == $undef_val && $validation_type == "NUMERIC"} {
	set ret_val $undef_string
    }
    if {$value == $min_int_val } {
	set ret_val $min_int_string
    }
    if {$value == $max_uint_val && $validation_type == "UNUMERIC"} {
	set ret_val $max_uint_string
    } 

    regsub -all $newline $ret_val {\n} ret_val
    
   
    return $ret_val
}

#-------------------------------------------------
#------   INPUT VALIDATION
#-------------------------------------------------

#------------------------------------------------------------------------------
# When pressing Apply or Ok button - the apply action should be performed.
# before creating the ecom file with the new updated options - every option
# (in all categories) that has been assigned  with a new value must be checked.
#------------------------------------------------------------------------------
proc perform_validation {} {

    global the_global_config_categories
    global category_visited_idx
    global legal_input
    global config_book_name
    global category_name_idx

    for {set i 0} { $i < [ llength $the_global_config_categories] } {incr i} {
	set category_info [lindex $the_global_config_categories $i]
	set is_visited [lindex $category_info $category_visited_idx]
	set category_name [lindex $category_info $category_name_idx]
	if { $is_visited == 1 } {
	    set legal_input [validate_category $category_info]
	    if {$legal_input == 0 } {
		$config_book_name raise $category_name
		break;
	    }
	}
    }

    return $legal_input
}

#-------------------------------------------------------------
# Validate all modified options in the category 
#-------------------------------------------------------------
proc validate_category {category_info} {

    global category_options_list_idx
    global option_e_var_name_idx
    global category_name_idx
    global legal_input

    set options_list [lindex $category_info $category_options_list_idx]
    set category_name [lindex $category_info $category_name_idx]
    for { set i 0 } { $i< [llength $options_list] } {incr i} {
	set option_info [lindex $options_list $i]
	set option_name [lindex $option_info $option_e_var_name_idx ]
	set legal_input [on_leave_entry $category_name $option_name]
	if { $legal_input == 0 } {
	    break;
	}
    }

    return $legal_input
}

proc numeric_validation {value category_name var_name old_val txt} {
    global config_wname

    set path "^( |\t)*(-)?(0b\[01]\[_01]*|0o\[0-7]\[_0-7]*|0\[hx]\[0-9a-f]\[_0-9a-f]*|\[0-9]\[_0-9]*\[mMkK]?)( |\t)*\$"
    set num_of_parm [llength $value]

        
    if { $num_of_parm <= 0 } {
	append txt "is missing." 
	config_validation_error_message $txt $category_name $var_name $old_val
	#grab $config_wname 
	return 0

    } elseif { $num_of_parm > 1 } { 
	append txt "has too many parameters."
	config_validation_error_message $txt $category_name $var_name $old_val
	#grab $config_wname
	return 0

    } elseif { [regexp $path $value match prefix the_value postfix] == 0 } {

	append txt "is illegal ('" $value "').\n"
	append txt "A numeric value is expected."
	config_validation_error_message $txt $category_name $var_name $old_val 
	#grab $config_wname
	return 0
	    
    } else { 

	return 1
    }
}

proc numeric_range_validation { min max val type } {

    if { $min == "NA" && $max == "NA" } {
	set legal_range 1
    } else {

	regsub -all "_" $val "" val

	set fmin "${min}E0"
	set fmax "${max}E0"
	set fval "${val}E0"


	if { $fval >= $fmin && $fval <= $fmax || $fmin == 0 && $fmax == 0 } {
	    set legal_range  1
	} else {
	    set legal_range 0
	} 
    }

    # DO NOT PERFORM VALIDATION ON TCL SIDE ....
    set legal_range  1
    return $legal_range

}

proc string_re_validate { string re } {

    set validation_stat 1

    if { $re != ".*"} {
	set validation_stat [ regexp "$re" "$string" ]
    }

    return $validation_stat
}


#-------------------------------------------------
#------   ERRORS
#-------------------------------------------------
proc show_config_error_message {title txt} {
    global config_wname

    set dlg_name $config_wname.err_msg_dlg

    toplevel $dlg_name
    wm title $dlg_name $title
    wm geometry $dlg_name +600+100  

    button $dlg_name.b -text "OK" -command "destroy $dlg_name"
    message $dlg_name.msg -text $txt -aspect 1000
    pack $dlg_name.msg -padx 20 -pady 5 -side top
    pack $dlg_name.b -padx 20 -pady 10 -side top

    focus $dlg_name.b
    grab $dlg_name
    tkwait window $dlg_name
    grab release $dlg_name
}

proc config_error_message {txt} {
    bell
    show_config_error_message "Error" $txt
}

proc show_config_validation_error_message {title txt category_name var_name orig_val} {
    global config_wname

    set dlg_name $config_wname.validiation_err_msg_dlg

    toplevel $dlg_name
    wm title $dlg_name $title
    wm geometry $dlg_name +600+100  

    button $dlg_name.ok -text "OK" -command "destroy $dlg_name"
    button $dlg_name.restore_val -text "Restore Value" -command "restore_value $dlg_name $category_name $var_name \"$orig_val\""
    message $dlg_name.msg -text $txt -aspect 1000
    pack $dlg_name.msg -padx 20 -pady 5 -side top
    pack $dlg_name.ok  -padx 20 -pady 10 -side left 
    pack $dlg_name.restore_val -padx 20 -pady 10 -side right
   
    bind $dlg_name <ButtonPress-1> {
	raise %W
    }

    focus $dlg_name.ok
    grab $dlg_name
    tkwait window $dlg_name
    grab release $dlg_name
    
}

proc restore_value { dlg_name category_name var_name original_value } {
 
    global tclVal$category_name$var_name

    set tcl_var_name tclVal$category_name$var_name
    set $tcl_var_name "$original_value"

    destroy $dlg_name
    
}

proc config_validation_error_message { txt category_name var_name orig_val} {
    
    global config_wname
    global config_book_name

    $config_book_name raise $category_name
    
    bell 
    show_config_validation_error_message "Invalid Value" $txt $category_name $var_name "$orig_val"

    grab $config_wname
   
}



proc create_new_col {} {
    
    global PageName
    global ColFrameName
    global ColIdx
    global group_border_width
    global col_options_num

    set col_options_num 0 
    set ColFrameName $PageName.col$ColIdx
   
    frame $ColFrameName -border $group_border_width
    pack $ColFrameName -anchor n -side left

    incr ColIdx
}


#------------------------------------------------------------------------
# While applying - collect all option in category that were modified and
# write them to the corresponding command in the ecom file that will be
# activated by specman , in order to update the config values
#-----------------------------------------------------------------------
proc collect_category_info { category_info file_name} {

    global category_options_list_idx
    global option_orig_val_idx
    global category_name_idx
    global option_type_idx
    global option_tcl_var_idx
    global option_e_var_name_idx
    global is_config_modified
    global file

    set is_modified 0

    set category_name [lindex $category_info $category_name_idx]
    set sn_config_cmd "config "

    append sn_config_cmd $category_name " " 
    set options_list [lindex $category_info $category_options_list_idx]

    for { set i 0 } { $i< [llength $options_list] } {incr i} {
	set option_info [lindex $options_list $i]
	set option_name [lindex $option_info $option_e_var_name_idx ]
	set option_type [lindex $option_info $option_type_idx]

	set new_val_name [lindex $option_info $option_tcl_var_idx]
	global $new_val_name
	set new_val [set $new_val_name]

	set original_val [lindex $option_info $option_orig_val_idx]
	
	if { $new_val != $original_val } {

	    set tmp_val ""
	    set is_modified 1 
	    
	    if { [string match QUATED $option_type] } {
		regsub -all {\"} $new_val {\\"} new_val
                append tmp_val "\"" $new_val "\""
                set new_val $tmp_val
	    }

	    append sn_config_cmd  " -" $option_name "=" $new_val
	}
    }

    
    if {  $is_modified == 1 } {
	
	# open ecom file if this is the first modification 
	if { $is_config_modified == 0 } {
	    set file  [open $file_name w 0666]
            puts $file "session.sn_config.already_validate = TRUE -- SILENT"
	    set is_config_modified 1
	}
	puts $file [concat $sn_config_cmd "-- SILENT"]
    }
}




proc on_leave_entry {category_name var_name} {
  
    global option_type_idx
    global option_min_idx
    global option_max_idx
    global option_re_idx
    global option_tcl_var_idx
    global option_orig_val_idx
    global option_label_idx
    global option_group_idx
    global max_int_val 
    global min_int_val
    global max_uint_val
    global max_int_string 
    global min_int_string
    global max_uint_string
    global undef_val
    global undef_string
   
    #set tokens [split $wname .]   
    #set var_name [ lrange $tokens end end] 
 
    global tclVal$category_name$var_name

    set option_info [get_option_info_by_category_name $category_name $var_name]
   
    set legal_val 1

    set new_val [set tclVal$category_name$var_name]
    set old_val [lindex $option_info $option_orig_val_idx] 
    set option_label [lindex $option_info $option_label_idx]
    set option_label [string trim $option_label ":"] 
    set option_group [lindex $option_info $option_group_idx]

    if { $new_val != $old_val } {

	set txt "Value of "
	append txt  "option '" $option_label "' "
	if { $option_group != ""} {
	    append txt "in group '" $option_group "' "
	}
	
	set type [lindex $option_info $option_type_idx]
	
	if { $type == "NUMERIC" || $type == "UNUMERIC" } {

	    if {[string compare $new_val $max_int_string] == 0 } {
		set new_val $max_int_val
	    } 
	    if {[string compare $new_val $min_int_string] == 0 } {
		set new_val $min_int_val
	    }

	    if { $type == "NUMERIC" } {
		if {[string compare $new_val $undef_string]== 0} {
		    set new_val $undef_val
		}
	    }
	    if {$type == "UNUMERIC" } {
		if {[string compare $new_val $max_uint_string]== 0} {
		    set new_val $max_uint_val
		}
	    }
	    
	    # check that new_val is NUMERIC
	    set legal_val [numeric_validation $new_val $category_name $var_name $old_val $txt]
	    
	    if { $legal_val == 1 } {
		# check that new val is in range
		set min [lindex $option_info $option_min_idx]
		set max [lindex $option_info $option_max_idx]
		set new_val [special_string_to_value $new_val]
		set legal_val [numeric_range_validation $min $max $new_val $type]
		if { $legal_val == 0 } {
		    append txt "is out of range\n"
		    append txt "The valid range is: " $min ".." $max 
		    config_validation_error_message $txt $category_name $var_name $old_val
		}
	    }
	}
	
	if { $type == "QUATED"} {

	    # check that new val is string
	    set re [lindex $option_info $option_re_idx]
	    set legal_val [string_re_validate $new_val $re]
	    if { $legal_val == 0 } {
		append txt "does not match the regular expression: '"
		append txt $re "'."
		config_validation_error_message $txt $category_name $var_name $old_val
	    }
	}
    }

    return $legal_val
}


proc on_page_change { } {

    global curr_page
    global config_book_name

    set raised_page  [$config_book_name raised]

    if { [string compare $raised_page $curr_page ] == 0 } {
	# click on the same page
	return
    } else {
	set status [validate_page $curr_page]
	if { $status == 0 } {
	    #validation error in page
	    new_sn_raise_page $curr_page
	} else {
	    new_sn_update_raised_page_params $raised_page
	}
    }

}

proc validate_page { page } {
 
    return 1
}

proc validate_tab {} {

    return 1
}


#-----------------------------------------------------------------------
# get category information from the_global_config_categories list
# according to category name ( the first list argument for each category )
#-------------------------------------------------------------------------
proc get_category_info_by_name { category} {
        
    global the_global_config_categories

    set category_idx  [get_category_global_idx $category]
    set category_info [get_category_info_by_idx $category_idx]
    return $category_info
}

#-----------------------------------------------------------------------
# get category information from the_global_config_categories list
# according to category idx ( the index in the_global_config_categories)
#-------------------------------------------------------------------------
proc get_category_info_by_idx { category_idx} {
        
    global the_global_config_categories
    
    set categories_num [llength $the_global_config_categories]
     if { $category_idx < 0 || $category_idx >= $categories_num } {
	config_error_message "Internal error - invalid category"
	return ""
    }

    set category_info [lindex $the_global_config_categories $category_idx]
    return $category_info
}

#-----------------------------------------------------------------------
# get the optinos list of category from the_global_config_categories list
# according to category idx ( the index in the_global_config_categories)
#-------------------------------------------------------------------------
proc get_category_options_info_by_name { category} {
        
    global category_options_list_idx

    set category_idx [get_category_global_idx $category]

    set category_info [get_category_info_by_idx $category_idx]
    set options_info [lindex $category_info $category_options_list_idx]

    
    return $options_info
}

#-----------------------------------------------------------------------
# get the optinos list of category from the_global_config_categories list
# according to category idx ( the index in the_global_config_categories)
#-------------------------------------------------------------------------
proc get_category_options_info_by_idx { category_idx} {
        
    global category_options_list_idx
    
    set category_info [get_category_info_by_idx $category_idx]
    set options_info [lindex $category_info $category_options_list_idx]

    return $options_info
}

#-----------------------------------------------------------------------
# get option information from a specific category in 
# the_global_config_categories list according var_name 
#-------------------------------------------------------------------------
proc get_option_info_by_category_name { category_name var_name } {

    global category_options_list_idx
    global option_e_var_name_idx
     
    set ret_val ""

    set category_options [get_category_options_info_by_name $category_name]

    for {set i 0} {$i < [llength $category_options]} {incr i} {
	set option_info [lindex $category_options $i]
	set option_var_name [lindex $option_info $option_e_var_name_idx ]
	if { [string compare $var_name $option_var_name] == 0 } {
	    set ret_val $option_info
	    break
	}
    }    

    if { [string compare $ret_val ""] == 0 } {
	config_error_message "Internal error - invalid var_name"
    }

    return $ret_val
}

#---------------------------------------------------------------
# get the index of category in the_global_config_categories list
#---------------------------------------------------------------
proc get_category_global_idx {category} {

    global the_global_config_categories
    global category_name_idx

    set ret_idx -1

    for {set i 0} {$i < [llength $the_global_config_categories]} {incr i} {
	set category_info [lindex $the_global_config_categories $i]
	set category_name [lindex $category_info  $category_name_idx]
	if { [string compare $category_name $category] == 0 } {
	    set ret_idx $i
	    break
	}
    }
    
    if { $ret_idx == -1} {
	config_error_message "Internal error - invalid category"
    }

    return $ret_idx
}

#------------------------------------------------------------------
# get the index of options in options_list according to var_name
#------------------------------------------------------------------
proc get_option_idx_by_var_name { options_list var_name } {

    global option_e_var_name_idx

    set ret_idx -1

    set loop_len [llength $options_list]
    for {set i 0} {$i < $loop_len } {incr i} {
	set option [lindex $options_list $i]
	set tmp_var [lindex $option  $option_e_var_name_idx]
	if { [string compare $tmp_var $var_name] == 0 } {
	    set ret_idx $i
	    break
	}
    }
    
    if { $ret_idx == -1} {
	config_error_message "Internal error - invalid var_name"
    }

    return $ret_idx
}



#---------------------------------------------------------------------------
#-- Build the global Ok Apply  and Close buttons
# this proc should remian unchenged. due to ontop-dev the name of internal
# procs is set to new....
#------------------------------------------------------------------------------
proc new_sn_build_config_buttons {wname} {
    tixButtonBox $wname.buttons -orientation horizontal
    $wname.buttons add ok -text "OK" \
	-command "new_sn_config_ok $wname" -width 6
    $wname.buttons add apply -text "Apply" \
	-command "new_sn_config_apply $wname" -width 6
    $wname.buttons add close -text "Close" \
	-command "sn_config_close $wname 1" -width 6

    pack $wname.buttons -expand 0 -fill both -side top

    sn_bind_balloon [$wname.buttons subwidget ok] \
    "Close this window, and set the new values" $wname

    sn_bind_balloon [$wname.buttons subwidget apply] \
    "Set the new values" $wname

    sn_bind_balloon [$wname.buttons subwidget close] \
    "Close this window, without setting the new values" $wname

}



#-------------------------------------------------
#------   UTILITIES  ROUTINES 
#-------------------------------------------------
proc sn_build_config_button {parent name label tool_tip kind} {
    global config_wname
    global unique_name_counter
    global items_vertical_specing

    set the_bt_name ""
    append the_bt_name $parent ".button" $unique_name_counter

    set btn_cmd ""
    append btn_cmd "sn_button_clicked " $config_wname " " $name " 1 1"

    if {$kind == "HELP"} {
	button $the_bt_name  -text $label -command $btn_cmd -fg blue
    } else {
	button $the_bt_name  -text $label -command $btn_cmd
    }
    pack $the_bt_name -side top -anchor center -pady $items_vertical_specing   
    
    
    sn_bind_balloon $the_bt_name $tool_tip $config_wname 

    incr unique_name_counter
}

proc sn_build_config_static_text {parent text_body} {

     global config_wname
     global unique_name_counter
     global group_item_hspecing
     global newline

     set the_name ""
     append the_name $parent ".static_text" $unique_name_counter
     regsub -all $newline $text_body \n text_body
     label $the_name -text $text_body -font slant -justify left
     pack $the_name -pady $group_item_hspecing -side top -anchor w

     incr unique_name_counter
}

proc sn_build_config_entry {parent label widget_name var_name all_values \
				balloon_text padx \
				width {build_special_for_seed 0} \
				{pady 5} {side top}} {
    global config_wname

    # build entry with label
    set the_name ""
    append the_name $parent "." $widget_name

    if { [llength $all_values] == 1 } {
	set status disabled
    } else {
	set status normal
    }

    tixLabelEntry $the_name -label $label -labelside top -state $status

   
    # config entry
    set the_entry [$the_name subwidget entry]
    $the_entry config -width $width -textvariable $var_name
    
    # config label
    set the_label [$the_name subwidget label]
    $the_label config -anchor w -pady 4

    if { $build_special_for_seed == 0 } {
	# pack entry with label
	pack $the_name -anchor w -padx $padx -pady $pady -side $side
    }
    
    # bind a balloon
    sn_bind_balloon $the_entry $balloon_text $config_wname
    

    # set colors
    sn_set_text_field_colors [$the_name subwidget entry]

    #bind $the_name <FocusOut> { on_leave_entry %W }
    

    if { $build_special_for_seed == 1} {
	set the_bt_name ""
	append the_bt_name $parent ".random"  
	button $the_bt_name  -text "Random" \
	    -command "sn_config_request_random_seed" \
	    -width 6
	pack $the_name -side left -padx $padx -pady $pady
	pack $the_bt_name -side left -anchor s   
	sn_bind_balloon $the_bt_name "Choose a new random initial seed" \
	    $config_wname
    }

}


proc sn_build_config_checkbox {parent label widget_name var_name all_values balloon_text padx} {

    global config_wname
    global items_vertical_specing
    
    # build space label
    set the_name ""
    append the_name $parent "." $widget_name
    set label_name ""
    append label_name $the_name "_label"
    label $label_name
    pack $label_name 

    if { [llength $all_values] == 1 } {
	set status disabled
    } else {
	set status normal
    }

    checkbutton $the_name -text $label -state $status\
	-variable $var_name \
	-anchor w -onvalue "TRUE" -offvalue "FALSE"  \
	-selectcolor navy -justify left

    # pack entry with label
    pack $the_name -anchor w -padx $padx -pady 1 -side top
    
    bind $the_name <ButtonRelease> {focus %W}
   
    # bind a balloon
    sn_bind_balloon $the_name $balloon_text $config_wname
   
    return
}

proc sn_build_config_dropdown {parent label widget_name var_name all_values tooltip padx\
				   {width 25}} {
    global config_wname
    global items_vertical_specing

    # build space label
    set the_name ""
    append the_name $parent "." $widget_name

    if { [llength $all_values] <= 1 } {
	set status disabled
    } else {
	set status normal
    }

    # build combo box
    tixComboBox $the_name -label $label -editable False -dropdown True \
	-history False  -variable $var_name \
	-labelside top -state $status

    set the_entry [$the_name subwidget entry]	
    $the_entry config -width $width

    set the_listbox [$the_name subwidget listbox]	
    $the_listbox config -height 7
    foreach val $all_values {
	$the_name insert end $val
    }

    set the_label [$the_name subwidget label]
    $the_label config -justify left -anchor w -pady 3

    sn_set_text_field_colors $the_listbox 0
    sn_set_text_field_colors [$the_name subwidget entry]

    # pack entry with label
    pack $the_name -anchor w -padx $padx -pady $items_vertical_specing
    
    # bind a balloon
    sn_bind_balloon $the_name $tooltip $config_wname

    return
}

proc sn_build_config_radiobutton {parent label widget_name var_name all_values 
				  balloon_text {width 25}} {
    global config_wname

    # build radio group
    set the_group_name ""
    append the_group_name $parent "." $widget_name

    set idx 0
    foreach val $all_values {
	radiobutton $the_group_name.$idx -variable $var_name -text $val  -value $val
	pack $the_group_name.$idx -side top
    }

    set the_label [$the_name subwidget label]
    $the_label config -justify left -anchor w -pady 3

    # pack entry with label
    pack $the_name -anchor w -padx 30 -pady 7
    
    # bind a balloon
    sn_bind_balloon $the_name $balloon_text $config_wname

    return
}



proc sn_config_request_random_seed {} {
    set cmd "display.choose_seed()"
    sn_command $cmd specman 0 0 0
    return
}


#-------------------------------------------------
#------   HANDLE SPECMAN COMMANDS
#-------------------------------------------------
proc sn_config_set_new_seed {new_value} {

    global tclValgenseed
    set tclValgenseed $new_value
    return
}



