
<'
//import ~/flow_tool/flow_header_2;
define CONSTANT_PREFIX "SFT_CONST";
define SFT_VERBOSE FALSE;
type context_kind:[NONE,CONSTANTS,DEF_FLOW,DEF_STEP,DO_FLOW,DO_STEP,INLINE_FLOW,INLINE_STEP,PARAMS,VARS,BLOCK,SELECT,RESULTS];
extend global {
   ntf_parser_total_lines: int;
   all_constants: sft_ntf_constants;
   current_target_file_name: string;
   //-----------------------------
   // the genrated name of a struct is sft_struct__<file name>_<struct title>
   //-----------------------------
   create_struct_name(name:string): string is {
      if name in all_sft_predefined_step_names {
         result = name;
      } else {
//         print current_target_file_name;
         result = appendf("sft_struct__%s__%s",current_target_file_name,name);
      };
   };
   remove_head(str: string):string is {
      result = str_sub(str,1,str_len(str));
   };
   substitute_vars(line:string,def_struct: sft_def_struct,line_num: int) :string is {
      var tokens: list of string = str_split_all(line,"/\\$\\w*/") ;
      
//      print tokens;
      if tokens.size()%2 == 0 {
         error("unmached $. on line ",line);
      };
      var command : string;
      var variables : string;
      for i from 0 to tokens.size()-1 {
         if i%2==0 {  // simple string
            command = append(command,tokens[i]);
         } else {      // variable substitution
            var var_name := remove_head(tokens[i]);
            var var_e_name := var_name;
            if var_name in all_constants.constants.name {
               var_e_name = append(CONSTANT_PREFIX,"_",var_name);
            };
            if SFT_VERBOSE {
               print def_struct.vars;
            };
            case {
               var_name in def_struct.vars.all(.var_type == INT).name {
                  command = append(command,"%d");
                  variables = append(variables,",",var_e_name);
               };
               var_name in def_struct.vars.all(.var_type == STRING).name {
                  command = append(command,"%s");
                  variables = append(variables,",",var_e_name);
               };
               default { error("no such variable ",var_name,", on line ",line_num); };
            };
         };
//         print i;
//         print i%2;
//         print command;
      };
      result = append("appendf(",quote(command),variables,")");
   };
};

//--------------------------------------
// the following structures hold the needed data to write the e code of 
// the flow and the different steps.
// a call to to_text() returns the e code.
//--------------------------------------

type var_group_t:[PARAM,VAR,RESULT,CONSTANT];
type var_type_t: [INT,STRING];


   
// this stores the definition of a single flow variable
struct sft_ntf_var {
   var_type: var_type_t;
   var_group: var_group_t;
   name: string;
   add_value(some_value:string) is {
      assert var_group in [PARAM,CONSTANT];
      case {
         me is a PARAM sft_ntf_var (pme)  { pme.default_value = some_value; };
         me is a CONSTANT sft_ntf_var (cme)  { cme.constant_value = some_value; };
      };
   };
   when PARAM sft_ntf_var {
      default_value: string;
   };
   when CONSTANT sft_ntf_var {
      constant_value: string;
   };
};

struct sft_ntf_constants {
   constants: list of CONSTANT sft_ntf_var;
   to_text():list of string is {
      for each (const) in constants {
         var name:= append(CONSTANT_PREFIX,"_",const.name);
         result.add(append("#ifndef ",name," then {"));
         case const.var_type {
            STRING: {result.add(append("define ",name," ",quote(const.constant_value),";")); };
            INT: {result.add(append("define ",name," ",const.constant_value,";")); };
         };
         result.add("};");
      };
   };
};
struct sft_base_struct {
   name: string;    //  the internal name of the struct
   vars: list of sft_ntf_var;
   //   flow_file: string;
   to_text():list of string is undefined;
};


// this struct and it's decentants store the data to create a single struct in the generated e code. 
struct sft_def_struct like sft_base_struct {
   title: string;   //  what the user see
   desc: string;
   get_linked_files():list of string is {
      result = {};
   };
   print_vars_definition():list of string is {
      for each (v) in vars.all(.var_group != CONSTANT) {
         result.add(append(v.name," : ",str_lower(append(v.var_type)),";"));
      };
      if vars.all(.var_group == PARAM).size() > 0  {
         result.add("// set default values:");
         result.add("init() is also {");
         for each (p) in vars.all(.var_group == PARAM) {
            case {
               p is a PARAM INT sft_ntf_var (pi) {
                  result.add(append("\t  ",pi.name," = ",pi.default_value.as_a(int),";"));
               };
               p is a PARAM STRING sft_ntf_var (ps) {
                  result.add(append("\t  ",ps.name," = ",quote(ps.default_value),";"));
               };
               default { assert FALSE; };
            };
         };      
         result.add("};");
      };
   };
};

struct sft_def_flow like sft_def_struct {
   do_steps: list of sft_do_struct;
   linked_files: list of string;
   current_client_name: string;
   get_linked_files():list of string is only {
//      print do_steps;
      result = linked_files;
      for each in do_steps {
         if it.source != NULL {
            result.add(it.source.get_linked_files());
         };
      };
   };
   to_text(): list of string is {
      result.resize(0);
      if desc == "" {
         desc = quote("");
      };
      result.add(<<#:
#ifndef <(create_struct_name(name))> then {
#define <(create_struct_name(name))>;         
struct <(create_struct_name(name))> like step_sequence_template {   
   <(print_vars_definition())>
   next_step(param : list of string):bool is {
      case counter {         
      end #);
      for i from 1 to do_steps.size() {
         if i>1 { do_steps[i-1].pre_calc = do_steps[i-2].post_calc; };
         result.add(append("\t\t",i,": {"));
         result.add(do_steps[i-1].to_text());
         result.add("\t\t};");
         
      };
      result.add(<<#:
      };
      new_step = current_subflow.next_step(param);
      if new_step {
          counter = counter+1;
      };
      if counter > <(do_steps.size())> {
         done = TRUE;
         result =  TRUE;
      };
   };   
};
};  // endif
      end #);        
      };
};

struct sft_def_select like sft_def_flow {
};   
   
struct sft_def_step like sft_def_struct {
   execution_kind : execution_kind_t;
   action: list of string;
   ntf_line: int;
   feedback: string;
   to_text():list of string is{
      result.resize(0);
      if desc == "" {
         desc = quote("");
      };
      result.add(<<#:
#ifndef <(create_struct_name(name))> then {         
#define <(create_struct_name(name))>;
struct <(create_struct_name(name))> like atomic_step_template {
   <(print_vars_definition())>
   init() is also {
      execution_kind = <(execution_kind)>;
   };
   description():string is {
      result = <(desc)>;
   };
   get_name():string is {
      result = "<(title)>";
   };
   execute():list of string is {
       <(action)>
   };
   execute_line():int is {   
      result = <(ntf_line)>;
   };
   feed_back():string is only{
      result = append(<(feedback)>);
   };
};
}; // endif
      end #);
   };

};



// this stores the data to generate the top_flow e struct
struct sft_top_flow like sft_base_struct{
   root_flow: sft_def_flow;
   desc: string;
//   linked_files: list of string;
   has_root():bool is {
      result = root_flow!=NULL;
   };
   to_text():list of string is {
      result.resize(0);
      var linked_files_string := str_join(root_flow.get_linked_files().unique(it),";");
      result.add(<<#:
#ifndef top_flow__<(name)> then {
#define top_flow__<(name)>;                  
struct top_flow__<(name)> like flow_template {         
   set_root_flow_step() is {
      root_flow_step = new <(create_struct_name(root_flow.name))>;
   };
   get_name():string is {
      result = "<(root_flow.title)>";
   };
   description():string is {
      result = <(root_flow.desc)>;
   };
   linked_files():list of string is {
      result = {<(linked_files_string)>};
   };
   preconditions():bool is {
      result = TRUE;
   };
   set_default_parameters() is {
   };
};
}; // endif
      end #);
   };
};


// this struct and it's descentants store ans generate the internal step statements in the generated
// e code
struct sft_do_struct like sft_base_struct {
   client_name: string;
   source : sft_def_struct;
   post_calc: list of string;
//   atomic: bool;
   param_values: list of string;
   pre_calc: list of string;
   set_client_name(client: string) is {
      if client_name == "" {           // only if there is no prior name (defined lower)
         if source is a sft_def_flow (flow) {
            for each (s) in flow.do_steps {
               s.set_client_name(client);
            };
         } else {
            client_name = client;
         };

      };
         
   };
   
   add_param_value(name: string, value:string, parent:sft_def_struct, line:int) is {
      var current_var:= source.vars.first(.name == name);
      var command := substitute_vars(append("sft_temp = (",quote(value),").as_a(string)"),parent,line);
      case current_var.var_type {
         INT: {  
            param_values.add( {
               append("specman(",command,");");
               appendf(".%s = sft_temp.as_a(int);",name) } );
         };
         STRING: {  
            param_values.add( {
               append("specman(",command,");");
               appendf(".%s = sft_temp;",name) } );
         };
      };
   };

   to_text():list of string is {
      result.resize(0);
      result.add(<<#:
           <(pre_calc)>
           if new_step {
              current_subflow = new <(create_struct_name(name))> with {
                 .client_name = "<(client_name)>";
                 <(param_values)>
              };
           };
      end #);
   };
};
struct sft_do_step like sft_do_struct {};
struct sft_do_flow like sft_do_struct {};

struct sft_do_select like sft_do_struct {};   


//-------------------------------
// this is the code generator.
// is fills the flow structures defined above while parsing the NTF.
// some syntax errors in the NTF are checked.
// after parsing is finished, all structures write their code to an
// e file with the same name. an ecom file is also created to launch
// the flow.
//-------------------------------
define current_step current_def.as_a(sft_def_step);
define current_flow current_def.as_a(sft_def_flow);

struct flow_parser_t like sn_text_input_parser {
   !target_file_name: string;
   !target_file_full_name: string;
   !current_context: context_kind;
   !step_def_list: list of sft_def_step;
   !flow_def_list: list of sft_def_flow;
   !flow_stack: list of sft_def_flow;
   !current_def: sft_def_struct;
   !top_flow: sft_top_flow;
   !current_do: sft_do_struct;
   !do_stack: list of sft_do_struct;
   !code_list : list of string;
   !top_flow_name: string;
   !top_flow_default_name: string;
   init() is also {
      current_context = NONE;
   };
   //------------------
   // on every call to (a legal) element, the last element is pushed to the stack.
   //------------------
   start_element(key_word : string,name : string,def_f_name:string,line : int) is {
      if SFT_VERBOSE{  out(" <<  start element ",name, " of type ",key_word); };
//      print current_context;
      var last_flow:= current_flow;  // may be NULL
      case key_word {
         "step": {

            current_def = new sft_def_step with {
               .title = name;  // default name
               .execution_kind = ONCE; // default execution_kind
            };
            current_step.vars.add(all_constants.constants);
            case current_context {
               // create a standalone step
               NONE: { 
                  current_context = DEF_STEP;
                  current_step.name = name;

               };
               // create an inline step
               [DEF_FLOW,INLINE_FLOW]: {       
                  current_context = INLINE_STEP;
                  name = append(last_flow.name,"__step_",last_flow.do_steps.size());
                  current_step.ntf_line = line;
//                  print name;
                  flow_stack.push(last_flow);
                  do_stack.push(current_do);
                  current_do = new sft_do_step with {
                     .name = name;
                     .source = current_step;
                     .set_client_name(last_flow.current_client_name);
                  };
                  // inline steps / flows know the variables of their parents:
                  current_step.vars.add(last_flow.vars);
                  for each (param) in last_flow.vars.all(.var_group!=CONSTANT).name {
                     current_do.param_values.add(append(".",param," = ",param,";"));
                  };
                  
               };
               default: {
                  error("step can not be used in the context of another step");
               };
            };
            current_step.name = name;
//            print current_step;
         };
         "flow": {
            current_def = new sft_def_flow;
            current_flow.vars.add(all_constants.constants);
            case current_context {
               NONE: {
                  current_context = DEF_FLOW;
               };
               [DEF_FLOW,INLINE_FLOW]: {           // nested flow
                  current_context = INLINE_FLOW;
                  name = append(last_flow.name,"__subflow_",last_flow.do_steps.size());
                  flow_stack.push(last_flow);
                  do_stack.push(current_do);
                  current_do = new sft_do_flow with {
                     .name = name;
                     .set_client_name(last_flow.current_client_name);
                     current_flow.current_client_name = last_flow.current_client_name;
                  };
                  // inline steps / flows know the variables of their parents:
                  current_flow.vars.add(last_flow.vars);
                  for each (param) in last_flow.vars.all(.var_group!=CONSTANT).name {
                     current_do.param_values.add(append(".",param," = ",param,";"));
                  };
               };
               default: {
                  error("step can not be used in the context of another step");
               };
            };

            current_flow.name = name;
//            print current_flow;
         }; // flow
         
         "select": {};
         "do": {
            if current_context not in [DEF_FLOW,INLINE_FLOW] {
               error("do can only be used in the context of a flow");
            };
            flow_stack.push(current_flow);
            case {
               name in all_sft_predefined_step_names {
                  current_def = NULL;
                  current_do = new sft_do_step with {
                     .name = name;
                     .set_client_name(last_flow.current_client_name);
                  };
                  current_context = DO_STEP;
               };
               name in step_def_list.name {
                  current_def = step_def_list.first(.name == name);
                  //                  print current_def;
                  current_do = new sft_do_step with {
                     .name = name;
                     .source = current_def;
                     .set_client_name(last_flow.current_client_name);
                  };
                  current_context = DO_STEP;
               };
               name in flow_def_list.name {
                  current_def =  flow_def_list.first(.name == name);
                  current_do = new sft_do_flow with {
                     .name = name;
                     .source = current_def;
                     .set_client_name(last_flow.current_client_name);
                  };
                  current_context = DO_FLOW;
               };
               
               default {
                  print step_def_list.name;
                  error("no such step/flow name ",quote(name));
               };
            };

         };  // do_step          

         "params": {
            if current_context not in [DEF_STEP,DEF_FLOW] {
               error("parameters can only be defined in the context of a new step or flow. on line ",line);
            };
            current_context = PARAMS;
         };
         "private": {
            if current_context not in [DEF_STEP,DEF_FLOW,INLINE_STEP,INLINE_FLOW] {
               error("private vars can only be defined in the context of a step or flow. on line ",line);
            };
            current_context = VARS;
         }; // private
         "results": {
            if current_context not in [DO_STEP,DO_FLOW] {
               error("results can only be defined in the context of a new step or flow. on line ",line);
            };
            current_context = RESULTS;
         }; 
         "constants": {
           if current_context != NONE {
               error("constants cannot be declared in another context. on line ",line);
            };
            current_context = CONSTANTS;
         }; 
         default: { current_context = BLOCK; };
      }; // case key_word
   }; // start_element()
   
   
      
   // need to check syntax
   add_attribute(name : string,value : string,def_f_name:string,line : int) is {
      if SFT_VERBOSE{   out("*** new attribute of type ",name," in context ",current_context);};
      case current_context {
         [DEF_STEP,INLINE_STEP]: {
            case name {
               "title": {
                  current_step.title = value; 
               };
               "description": { current_step.desc = substitute_vars(value,current_step,line); };
               // this command is executed on the SFT 
               "local": { current_step.action.add(append("specman(",quote(value),");")); };
               // this command will be sent and executed on the client tool
               "command": {
                  var command := substitute_vars(value,current_step,line);
                  current_step.action.add(append("send_command(",command,");"));
               };
               // this command will be transformed to the relevant GUI syntax (eg. GT), and executed on the
               // client tool.
               "gui_command": {
                  var command := substitute_vars(append("--GT GUI ",value),current_step,line);
                  current_step.action.add(append("send_command(",command,");"));
               };
               "show_file": {
                  var command := append("show_source_window(",quote(value),");");
                  current_step.action.add(command);
               };
               // this command will be executed by the UNIX system.  ( e.g. spawn a new application )
               "system": {current_step.action.add(append("compute system(",quote(value),");")); };
               "execution": { try {
                  current_step.execution_kind = str_upper(value).as_a(execution_kind_t);
               } else {
                  error("attribute 'execution' must be in [free,must,once,no_exec]. on line ",line);
               };};
               // this is the component to be highligthed in utorial mode.
               "highlight_component" : {};
               // syntax:
               //           eval: some_var = some_expression;
               // 'some_expression' is a value-returning expression executed on the client.
               // 'some_var' is a legal name of step variable.
               // some_var will get the value returned from the expression.
               // e.g:
               //   vars { row: int; s:string; }; 
               //   eval: s = sn_gt.get_table_value($row,4);
               //
               // if the returned value is not of the same type as the variable, the behavior is undefined.
               "eval": {
                  var tokens:list of string = str_split(value,"/\ *=\ */");
//                  print tokens;
                  if tokens.size()<2 {
                     error("syntax error in line ",line,"\ncorrect syntax: \neval: var_name = value_returning_expression;\ne.g. eval: sqrt_x = isqrt(sys.p.x)");
                  };
                  var lhs := tokens[0];
                  tokens.delete(0);
                  var rhs := substitute_vars(str_join(tokens,"="),current_step,line);
//                  print rhs;
                  var command:string;
                  case {
                     lhs in current_step.vars.all(.var_type == INT).name{
                        command = append(lhs," = get_int_value(",rhs,");");
                     };
                     lhs in current_step.vars.all(.var_type == STRING).name {
                        command = append(lhs," = get_exp_value(",rhs,");");
                     };
                     default: { 
                        error("no such variable ",lhs,", line ",line); 
                     };
                  }; // case
                  current_step.action.add(command);
               }; // eval
               "feedback": { current_step.feedback = substitute_vars(value,current_step,line); };
               default: { error("no such attribute ",quote(name)," on line ",line); };
            };  // case name
         }; // STEP

         [DEF_FLOW,INLINE_FLOW]: {
            case name {
               "title": { current_flow.title = value; };
               "description": { current_flow.desc = substitute_vars(value,current_flow,line); };
               "linked_files": { current_flow.linked_files.add(quote(value)); };
               "tool_name" : {current_flow.current_client_name = value; };
               default: { error("no such attribute ",quote(name)," on line ",line); };
            };
         }; // FLOW
         [PARAMS,VARS,CONSTANTS]: {
            var new_var:= new sft_ntf_var with {
               .name = name;
               if current_context == PARAMS {.var_group = PARAM; };
               if current_context == VARS {.var_group = VAR; };
               if current_context == CONSTANTS {.var_group = CONSTANT; };

            };
//            print "------------------------------------------------";
//            print new_var;
            var tokens:= str_split(value,"/\ *=\ */");
//            print tokens;
            var some_value := "";
            if tokens.size() > 1 {
               some_value = tokens[1];
            };
            if current_context not in [PARAMS,CONSTANTS] and some_value != "" {
               error("only parameters can get default values. on line ",line);
            };
            if current_context == CONSTANTS and some_value == "" {
               error("constant without a value. on line ",line);
            };
            // need to check legal values
            // need to handle * case (by reference) ?
            case tokens[0] {
               "int" : { 
                  new_var.var_type = INT; 
                  if some_value != "" {
                     new_var.add_value(some_value.as_a(int).as_a(string));
                  };
               };
               "string":  { 
                  new_var.var_type = STRING; 
                  if some_value != "" {
                     new_var.add_value(some_value);
                  };
               };
               default: { error("unfamiliar variable type : ",value,". on line ",line); };
            };
            var name_exist := FALSE;
            if current_context not in [CONSTANTS] {
               name_exist =  name in {current_def.vars.name};
               current_def.vars.add( new_var );
            } else {
               // check for already exist constants
               name_exist = name in {all_constants.constants.name};
               //               try {
               //                  var const_name: = append(CONSTANT_PREFIX,"_",name);
               //                  specman(append("compute ",const_name));
               //                  name_exist = TRUE;
               //             };
               all_constants.constants.add( new_var.as_a(CONSTANT sft_ntf_var));
            };
            if name_exist {
               error("the constant / variable  ",quote(name)," was already declared. on line ",line);
            };


         }; // PARAM
         
         RESULTS: {  // temporary feature: no syntax check
            var result_source : string = value;
            var result_target : string = name;
            current_do.post_calc.add(appendf("%s = current_subflow.as_a(%s).%s;",result_target,create_struct_name(current_do.name),result_source));
         };
            
            
         [DO_STEP,DO_FLOW]: {
            //            check that the parameter name exists:
            if current_def!= NULL and name not in current_def.vars.all(.var_group == PARAM).name {
               error("this step/flow does not have a parameter with this name, on line ",line);
            };
            current_do.add_param_value(name,value,flow_stack.last(TRUE),line);
//            var current_var:= current_def.vars.first(.name == name);
//            var command := substitute_vars(append("sft_temp = (",quote(value),").as_a(string)"),flow_stack.last(TRUE),line);
//            case current_var.var_type {
//               INT: {  
//                  current_do.param_values.add( {
//                     append("specman(",command,");");
//                     appendf(".%s = sft_temp.as_a(int);",name) } );
//               };
//               STRING: {  
//                  current_do.param_values.add( {
//                     append("specman(",command,");");
//                     appendf(".%s = sft_temp;",name) } );
//               };
                  
//            };
         }; // DO
      }; // case      
      
   }; // add_attribute()
   
   end_element(key_word : string,name : string,def_f_name:string,line : int) is {
      if SFT_VERBOSE {   out(" >> end element ",name, " of type ",key_word); };
       case current_context {
          [PARAMS,VARS,BLOCK]: {
             case {
                current_def is a sft_def_step {
                   current_context = DEF_STEP;
                };
                current_def is a sft_def_flow {
                   current_context = DEF_FLOW;
                };
                default {
                   assert FALSE;
                };
             };
          };
          RESULTS: {
             case {
                current_do is a sft_do_step {
                   current_context = DO_STEP;
                };
                current_do is a sft_do_flow {
                   current_context = DO_FLOW;
                };
                default {
                   assert FALSE;
                };
             };
          };
          CONSTANTS: {
             current_context = NONE;
          };
          [INLINE_FLOW,INLINE_STEP]: {
             if current_context == INLINE_STEP { 
                step_def_list.add(current_step); 
             } else { 
                flow_def_list.add(current_flow); 
             };
             current_def = flow_stack.pop();  // pop last flow

             current_flow.do_steps.add(current_do);
             if do_stack.size()>0 {current_do = do_stack.pop(); };
             current_context = (flow_stack.is_empty())?DEF_FLOW:INLINE_FLOW;
          };
          [DO_STEP,DO_FLOW] : {
             current_def = flow_stack.pop();  // pop last flow
             current_flow.do_steps.add(current_do);
             if do_stack.size()>0 {current_do = do_stack.pop(); };
             current_context = (flow_stack.is_empty())?DEF_FLOW:INLINE_FLOW;
          };
          DEF_FLOW: {
             if current_flow.title == "" {
                current_def.title = current_flow.name;
             };
             // name priorities:
             //    1) user choise
             //    2) the flow with name=filename
             //    3) the last flow in the file 
             flow_def_list.add(current_flow);
             if (top_flow_default_name == name) or 
                (top_flow_default_name == "" and (top_flow==NULL or top_flow.name!= target_file_name)) {
                top_flow = new sft_top_flow with {
                   top_flow_name = name;
                   .name = name;
                   .root_flow = current_flow;
                };
             };

             current_context = NONE;
          };
          DEF_STEP: {
             if current_step.title == "" {
                current_def.title = current_step.name;
             };
             step_def_list.add(current_step);
             current_context = NONE;
          };
        

       }; // case
    }; // end_element()
    
    // this is the activation
    parse_flow(file_name:string, top_flow:string) is {
       top_flow_default_name = top_flow;
       print top_flow_name;
       parse_file(file_name);
    };
    parse_file(file_name:string) is first {
       code_list.resize(0);
       if top_flow_name == "" {
//          error("no top flow name was given, use parse_flow(file_name,flow_name)");
       };
       current_context = NONE;
       current_def = NULL;
       
       // create target file name:
       target_file_full_name = str_remove_ext(file_name);
       target_file_name = str_remove_path(target_file_full_name);
       global.current_target_file_name = target_file_name;
       all_constants = new;
       print target_file_name;
       
    };
    
    parse_file(file_name:string) is also {
       if current_context != NONE {
          error("open context ",current_context);
       };
    }; // parse_file()
    
    write_e_file() is {
       code_list.add("<'");
       code_list.add(append("import ",SFT_HOME,"/flow_header_2;"));
       code_list.add(all_constants.to_text());
       for each (step) in step_def_list {
          code_list.add(step.to_text());
       };
       for each (flow) in flow_def_list {
          code_list.add(flow.to_text());
       };
       code_list.add(top_flow.to_text());
       code_list.add("'>");
       files.write_string_list(append(target_file_full_name,".e"),code_list);
       out("wrote file: ",target_file_full_name,".e");
       files.write_string_list(append(target_file_full_name,".ecom"),{
          append("load -if ",target_file_full_name,".e");
          append("sft_main.loaded_flow = new top_flow__",top_flow_name)});  
    }; // parse_file
    get_main_title():string is {
       result = top_flow.root_flow.title;
    };
}; // struct flow_parser_t



'>

     DRIVER
     
<'
extend global {
   flow_parser2: flow_parser_t;
   ntf() is {
      parse("/home/reshefm/flow_tool/ntf/nested.fd");
    //  print flow_parser_t.code_list;
   }; 
   parse(filename:string) is {
      flow_parser2 = new;
      flow_parser2.parse_file(filename);
      flow_parser2.write_e_file();
   };
}; 
   
'>



atomic_step step1 {
   name:  "name"
   description: "description of step 1";
