//---------------------------------------
// this part reads the flow trees from the disk (.ft files) to displayable vtx trees.
// the files are in simple NTF format
//---------------------------------------

<'
import sft_flow_tree;
define current_flow_node current_node.as_a(FLOW sft_flow_tree_node);
type tree_context_t:[NONE,GROUP,LESSON,FLOW];
extend global {
   current_ntf_line: int;
   ntf_error(err: string) is {
      error(err,". on line ",current_ntf_line);
   };
};
struct flow_tree_parser_t like sn_text_input_parser {
   node_stack: list of sft_flow_tree_node;
   top_groups: list of sft_flow_tree_node;
   tree_context: tree_context_t;
   current_node: sft_flow_tree_node;
   start_element(key_word : string,name : string,def_f_name:string,line : int) is {
      if SFT_VERBOSE {out(" <<  start element ",name, " of type ",key_word);};
      current_ntf_line = line;
      var last_node := current_node;
      if tree_context != NONE {
         node_stack.push(last_node);
      };
      case key_word {
         "group": {
            tree_context = GROUP;
            current_node = new GROUP sft_flow_tree_node with {
               .title = name;
            };
         };
         "lesson": {
            tree_context = LESSON;
            current_node = new LESSON sft_flow_tree_node with {
               .title = name;
            };
         };
         "flow": {
            tree_context = FLOW;
            current_node = new FLOW sft_flow_tree_node ;
         };
      }; // case key_word
      if last_node!=NULL {
         current_node.homedir = last_node.homedir;
      };
   }; // start_element()
   
   add_attribute(name : string,value : string,def_f_name:string,line : int) is {
      current_ntf_line = line;
      case tree_context {
         FLOW: {
            case name {
               "flow_file": {
                  current_flow_node.flow_filename = append(current_flow_node.homedir,"/",value);
               };
            };
         };
         default: {
            case name {
               "home" : { current_node.homedir = value; };
               default: { ntf_error("no such attribute"); };
            };
         };
      }; // case tree_context
      if name == "title"  {  current_node.title = value; };
   }; // add_attribute()
   
   end_element(key_word : string,name : string,def_f_name:string,line : int) is {
      if SFT_VERBOSE{ out(" >> end element ",name, " of type ",key_word); };
      current_ntf_line = line;
      var finished_node := current_node;
      if SFT_VERBOSE {    print node_stack; };
      if node_stack.size()>0 {
         current_node = node_stack.pop();
         tree_context = current_node.kind.as_a(string).as_a(tree_context_t);
      } else {
         tree_context = NONE;
      };
    
      case tree_context {  // parent level
         NONE: {
            top_groups.add(finished_node);
         };
         GROUP: {
            current_node.as_a(GROUP sft_flow_tree_node).subgroups.add(finished_node);
         };
         LESSON: {
            current_node.as_a(LESSON sft_flow_tree_node).flows.add(finished_node.as_a(FLOW sft_flow_tree_node));
         };
      };
   }; // end_element()
   parse_file(filename: string) is first {
      top_groups.resize(0);
   };
   parse_file(filename: string) is also {
      for each in top_groups {
         it.self_check(FALSE);
      };
   };
   // return the top-most groups 
   get_top_groups():list of sft_flow_tree_node is {
      result = top_groups;
   };
};

'>

   DRIVER
   
<'
extend global {
   tree_parser2: flow_tree_parser_t;
   tree() is {
      tree_parser2 = new;
      tree_parser2.parse_file("/home/reshefm/flow_tool/flows/flow_tree_ex.ft");
      specman("tr global.tree_parser.top_groups");
   };
};

'>