<'
'>

-----------------------------------------------------------------------------------------
--
-- created: 28/9/06
-- author:  Reshef Meir
--
-- this file use sockets to pass SN commands to SN based clients (specview, vmanager, etc.)
--
-- one way to use it is load directly the file 'sft_client.e' from the client side, establish 
-- the connection manually (with establish_client_2() and server_accept_2() methods ),
-- and then run wait_for_commands() from the client side.
--
-- a simpler way is to spawn a client program automatically.  e.g:
-- spawn_new_client("vmanager","VM_CLIENT_1");
--
-- * flags can be added to the command (e.g. "vmanager -gt -c 'load something.e'")
-- * the name can be any string, and is used for later reference by send_command() and 
--   get_exp_value()  methods
--
-------------------------------------------------------------------------------------------

<'
import sft_sockets_h;

define CLIENT_PATH append(SFT_HOME,"/sockets");


extend global {
   sft_server: sft_server_t;
};

struct sft_server_t {
   !sock_listener : sft_socket_interactor;
   !sock_client: list of sft_socket_interactor;
   !current_client_control: client_control_t;
   listen() is {
      sock_listener = new;
      try {
         sock_listener.establish_server();
      } else {
         sn_error("port is probably taken. check if any application remained open.");
      };
   };
   ------------------
   -- this is the main method used to raise an SN client, and connect to it.
   -- the client will then wait for commands.
   ------------------
   spawn_new_client(sys_command:string, client_name:string, script:bool): sft_socket_interactor is {
      sock_listener.server_listen();
      out("Specman is waiting to accept connection ...");
      var host := files.get_host_name();
      var port := sock_listener.port;
      print sys_command;
      print host;
      print port;
      if script {
         var script_name:= "run_script.sh";
         files.write_string_list(script_name,{
            "#! /bin/csh";
            "unsetenv SPECMAN_DISPLAY";
            appendf("%s -p 'load -if %s/sft_client;save sft_client; sft_client.connect(%s,%d,%s)' & ",sys_command,CLIENT_PATH,quote(host),port,quote(client_name))
         } );
         compute system(append("chmod 777 ",script_name));
         compute system(script_name);
      } else {
         compute system(appendf("%s -p 'load -if %s/sft_client;save sft_client; sft_client.connect(%s,%d,%s)' & ",sys_command,CLIENT_PATH,quote(host),port,quote(client_name)));
      };
      result = sock_listener.server_accept_2(); 
      outf("connection established. client name = %s\n",result.client_name);
      if result in sock_client {
         error("client ",result.client_name," already exists");
      };
      sock_client.add(result);
   };
   ------------
   -- the sent command will be executed on the client.
   ------------
   send_command(command:string, client_name:string) is {
      var ind := sock_client.first_index(it != NULL and it.client_name==client_name);
      if ind != UNDEF {
            sock_client[ind].send_command(command);
      } else {
         error("client ",quote(client_name)," does not exist");
      };
   };
   ------------
   -- the expression will be evaluated on the client, and the method
   -- returns it's value.
   -- e.g:
   -- send_command("sys.x = 5","CLIENT_4");
   -- send_command("sys.p.y = 3","CLIENT_4");
   -- get_exp_value("sys.x + sys.p.y",NUM,"CLIENT_4")  will return "8"
   --
   -- is it possible to check that exp is a value-returning
   -- expression?
   ------------
   get_exp_value(exp: string,exp_t: expression_type, client_name: string):string is {
      var ind := sock_client.first_index(it != NULL and it.client_name==client_name);
      if ind != UNDEF {
         result = sock_client[ind].get_exp_value(exp,exp_t);
         print result;
      } else {
         error("client ",quote(client_name)," does not exist");
      };
   };
   ----------
   -- close the connection with this client
   ----------
   close(client_name:string) is {
      var ind := sock_client.first_index(it != NULL and it.client_name==client_name);
      if ind != UNDEF {
            sock_client[ind].send_close();
            sock_client[ind].close();
            sock_client[ind] = NULL;

      } else {
         error("client ",quote(client_name)," does not exist");
      };
   };
   pass_control_to_all_clients() is {
      for each (client) in sock_client {
         if (client!=NULL) {
            client.pass_control();
         };
      };
   };
   // it will be good if this method will be activated 
   // all the time (or when the mouse go over the SFT main)
   get_control_state():client_control_t is {
      result = SERVER;
      for each (client) in sock_client {
         if (client!=NULL) {
            result = client.get_control_state();
         };
      };
   };
   close_all() is {
      out("closing all connections...");
      files.write_string_list("client_control",{"NO SERVER"});  -- for manual clients
      for each (client) in sock_client {
         if (client!=NULL && client.state == CONNECTED) {
            client.send_close();
            client.close();
         };
         sock_client = {};
      };
   };
   close_listener() is {
      close_all();
      if sock_listener.state != CLOSED {
         sock_listener.close();
      };
      print sock_listener;
   };
   kill_all_clients() is {
      for each (client) in sock_client {
         if (client!=NULL && client.state == CONNECTED) {
            client.send_command(SOCK_QUIT_COMMAND);
            client.close();
         };
         sock_client = {};
      };
   };


};
'>