<'
define SOCK_VERSION "0.1";
define SOCK_COMMAND_PREFIX "$$$";
define SOCK_EXP_PREFIX "###";
define SOCK_NUM_PREFIX "###";
define SOCK_CLOSE_COMMAND "CLOSE_CONNECTION";
define SOCK_QUIT_COMMAND "CLOSE_CONNECTION_AND_QUIT";
define SOCK_PASS_CONTROL_COMMAND "PASS_CONTROL";
define DEFAULT_TIMEOUT 20;
define DANGER_COMMANDS {"res";"rel";"load"};

type client_control_t:[SERVER,CLIENT];
type expression_type:[NUM,EXP];

struct sft_socket_interactor like sn_socket_interactor {
   !client_name: string;
   !client_control: client_control_t;
   !server_wait: bool;
   //client side   
   establish_client_2(host: string, port: int) is {
      establish_client(host,port,20);
      var server_version := receive();
      print server_version;
      if server_version != SOCK_VERSION {
         outf("Error: client using version %s, while server is using version %s",SOCK_VERSION,server_version);
         return;
      };
      send(client_name);
      client_control = SERVER;
      files.write_string_list("client_control",{"SERVER"});
   };

   send_command(command:string) is {
      if state == CONNECTED {
         send(append(SOCK_COMMAND_PREFIX,command));
      } else {
         error("SFT error: can't sent command. connection is ",state);
      };
   };
   get_exp_value(exp:string,exp_t: expression_type):string is {
      if state == CONNECTED {
         send(append(SOCK_EXP_PREFIX,exp_t.as_a(string),exp));
         result = receive();
      } else {
         error("SFT error: can't get value. connection is ",state);
         result = "";
      };

   };
   send_close() is {
      if state == CONNECTED {
         send(SOCK_CLOSE_COMMAND);
      } else {
         warning("SFT error: can't close. connection is ",state);
      };
   };
   
   // pass the control to the client (i.e. the client will stop listening)
   pass_control() is {
      client_control = CLIENT;
      files.write_string_list("client_control",{"CLIENT"});
      send_command(SOCK_PASS_CONTROL_COMMAND);
   };
   get_control_state():client_control_t is {
      var s: string;
      for each line in file "client_control" {
         s = it;
      };
      client_control = (s=="CLIENT")?CLIENT:SERVER;
      result = client_control;
   };
   
   sft_convert(si: sn_socket_interactor,name: string):sft_socket_interactor is {
      result = new sft_socket_interactor with {
         .socket_id = si.socket_id;
         .buffer = si.buffer;
         .port = si.port;
         .state = si.state;
         .client_name = name;
      };
   };
   server_accept_2(): sft_socket_interactor is {
      result = sft_convert(server_accept(),"");
      result.send(SOCK_VERSION);
      result.client_name = result.receive();
      set_state(BOUND);  // prepare for another connection
   };

   receive_command():string is {
      var command := receive();
      if str_sub(command,0,str_len(SOCK_COMMAND_PREFIX)) == SOCK_COMMAND_PREFIX {
         command = str_replace(command,SOCK_COMMAND_PREFIX,"");
         out("received command: ",command);
      };
      result = command;
      server_wait = FALSE;
      if str_sub(command,0,str_len(SOCK_EXP_PREFIX)) == SOCK_EXP_PREFIX {
         server_wait = TRUE;
         command = str_replace(command,SOCK_EXP_PREFIX,"");
         var exp_t := str_sub(command,0,3).as_a(expression_type);
         command = str_replace(command,"NUM","");
         command = str_replace(command,"EXP","");
         var exp_value :string;
         case exp_t {
            NUM: { exp_value = rf_manager.eval_exp_string(command).get_value().as_a(string); };
            EXP: { exp_value = rf_manager.eval_exp_string(command).get_value().unsafe(); };
         };
         send(exp_value);
         result = "";
      };
   };
   
};
'>